#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include "struct.h"
#include "defines.h"
#include "errors.h"
#include "stats.h"
#include "mem.h"
#include "log.h"
#include <syslog.h>
#include "gc.h"
#include "irc_printf.h"

extern int match(char *, char *);

extern struct msg_use msgs;
extern char *tmp, *t;

/* int my_strchr(const char *s, char c)
   returns the index of c in the string s or -1 if the character not found
 */

int my_strchr(const char *s, char c)
{
     int i = 0;

     while (*s) {
	  if (*s == c) {
	       return i;
	  }
	  i++;
	  s++;
     }
     return -1;
}

/* char *my_substr(const char *s, int begin, int end)
   returns a string representing the characters from indexes begin to end in the string s - memory returned
   by this function must be freed when the function using it is finished - otherwise a memory leak will occur!
 */

char *my_substr(const char *s, int begin, int end)
{
/*  char *buf = (char *) malloc(end - begin + 2); */
     char *buf = (char *) gcAlloc(end - begin + 2);
     int i;

     bzero(buf, end - begin + 2);
     if (begin == end) {
	  buf[0] = s[begin];
     } else {
	  for (i = begin; i <= end; i++) {
	       buf[i - begin] = s[i];
	  }
     }
     return buf;
}

/* int make_pretty_mask(char *my_mask, char *retbuf, int len)
   makes a userhost mask of the form "<nick>!<userid>@<hostname>" out of any string passed into it, and
   stores it in retbuf, with a maximum length of len.

   returns 0
 */

int make_pretty_mask(char *my_mask, char *retbuf, int len)
{
     char *buf = retbuf;
     char *nick = (char *) malloc(15);
     char *userid = (char *) malloc(15);
     char *hm = (char *) malloc(100);
     char *nh = nick;
     char *uh = userid;
     char *hh = hm;
     int exc, at;

     exc = at = -1;

     bzero(nick, 15);
     bzero(userid, 15);
     bzero(hm, 100);

     exc = my_strchr(my_mask, '!');
     at = my_strchr(my_mask, '@');
     if (exc == -1 && at == -1) {
	  snprintf(nh, 15, "*");
	  snprintf(uh, 15, "*");
	  snprintf(hm, 100, my_mask);
     } else if (exc == -1 && at >= 0) {
	  char *q = my_substr(my_mask, 0, at - 1);
	  char *r = my_substr(my_mask, at + 1, strlen(my_mask));

	  snprintf(nh, 15, "*");
	  snprintf(uh, 15, q);
	  snprintf(hm, 100, r);
	  gcDecRefcnt(q);
	  gcDecRefcnt(r);
     } else if (exc >= 0 && at == -1) {
	  char *q = my_substr(my_mask, 0, exc - 1);
	  char *r = my_substr(my_mask, exc + 1, strlen(my_mask));

	  snprintf(nh, 15, q);
	  snprintf(uh, 15, "*");
	  snprintf(hm, 100, r);
	  gcDecRefcnt(q);
	  gcDecRefcnt(r);
     } else {
	  if (exc < at) {
	       char *q = my_substr(my_mask, 0, exc - 1);
	       char *r = my_substr(my_mask, exc + 1, at - 1);
	       char *s = my_substr(my_mask, at + 1, strlen(my_mask));

	       snprintf(nh, 15, q);
	       snprintf(uh, 15, r);
	       snprintf(hm, 100, s);
	       gcDecRefcnt(q);
	       gcDecRefcnt(r);
	       gcDecRefcnt(s);
	  } else {
	       char *q = my_substr(my_mask, 0, at - 1);
	       char *r = my_substr(my_mask, at + 1, exc - 1);
	       char *s = my_substr(my_mask, exc + 1, strlen(my_mask));

	       snprintf(nh, 15, r);
	       snprintf(uh, 15, q);
	       snprintf(hm, 100, s);
	       gcDecRefcnt(q);
	       gcDecRefcnt(r);
	       gcDecRefcnt(s);
	  }
     }

     snprintf(retbuf, len, "%s!%s@%s", nh, uh, hh);
     free(nh);
     free(uh);
     free(hh);
     return 0;
}

/* void __send_cprivmsg(struct fd_list *fd, char *dest, char *msg)
   sends a message to the channel in dest containing the text of msg from the user in fd
 */

void internal_send_cprivmsg(struct fd_list *, char *, char *, int, int, int);

void __send_cprivmsg(struct fd_list *fd, char *dest, char *msg)
{
     internal_send_cprivmsg(fd, dest, msg, 0, 0, 1);
     return;
}

void __send_tprivmsg(struct fd_list *fd, char *dest, char *msg)
{
     int voice = 0;
     int oper = 0;
     char *d = dest;

     while (*d) {
	  switch (*d) {
	  case '+':
	       voice = 1;
	       break;
	  case '@':
	       oper = 1;
	       break;
	  case '#':
	       internal_send_cprivmsg(fd, d, msg, voice, oper, 0);
	       return;
	       break;
	  default:
	       break;
	  }
	  d++;
     }
     return;
}

void internal_send_cprivmsg(struct fd_list *fd, char *dest, char *msg, int voice, int oper, int all)
{
     struct chan_list *list;
     char *my_hostname;

     if (fd->hide_hostname) {
	  my_hostname = "hidden.hurricane.net";
     } else {
	  my_hostname = fd->hostname;
     }

     list = __clist_head->next;

     fd->last_msg = __time(NULL);

     while (list != NULL) {
	  if (!strcasecmp(list->name, dest)) {
	       int ok = 0;
	       struct user_list *user;

	       if (list->no_outside_msg || list->mute) {
		    struct user_list *t;

		    t = list->users;
		    while (t != NULL) {
			 if (t->user == fd) {
			      if (list->mute) {
				   if (t->voice || t->oper)
					ok = 1;
			      } else {
				   ok = 1;
			      }
			      t = NULL;
			 }
			 if (t != NULL)
			      t = t->next;
		    }
	       } else {
		    ok = 1;
	       }

	       {
		    struct ban_list *b = list->bans;
		    char *hm = alloca(200);

		    bzero(hm, 200);
		    snprintf(hm, 200, "%s!%s@%s", fd->nick, fd->userid, my_hostname);
		    while (b) {
			 if (!match(b->banid, hm)) {
			      ok = 0;
			      b = NULL;
			 }
			 if (b)
			      b = b->next;
		    }
	       }

	       if (oper || voice) {
		    struct user_list *u = list->users;

		    while (u) {
			 if (u->user == fd) {
				if (oper && voice) {
					if (!u->oper && !u->voice) {
						ok = 0;
					}
					if (!u->oper) {
						oper = 0;
					}
				} else if (oper) {
					if (!u->oper) {
						ok = 0;
					}
				} else if (voice) {
					if (!u->oper && !u->voice) {
						ok = 0;
					}
				}
				u = NULL;
			 }
			 if (u)
			      u = u->next;
		    }
	       }
	       if (ok) {
		    user = list->users;
		    while (user != NULL) {
			 if (user->user != fd) {
			      if (!user->user->deaf) {
				   if (oper || voice) {
					int sent = 0;

					if (oper) {
					     if (user->oper) {
						  char mydest[strlen(dest) + 3];
						  int i = 0;

						  bzero(mydest, strlen(dest) + 3);
						  sent = 1;
						  if (voice)
						       mydest[i++] = '+';
						  if (oper)
						       mydest[i++] = '@';
						  strcat(&(mydest[i]), dest);
						  irc_fprintf(user->user->sok, PRIVMSG, fd->nick, fd->userid, my_hostname, mydest, msg);
					     }
					}
					if (voice && !sent) {
					     if (user->voice || user->oper) {
						  char mydest[strlen(dest) + 3];
						  int i = 0;

						  bzero(mydest, strlen(dest) + 3);

						  if (voice)
						       mydest[i++] = '+';
						  if (oper)
						       mydest[i++] = '@';
						  strcat(&(mydest[i]), dest);
						  irc_fprintf(user->user->sok, PRIVMSG, fd->nick, fd->userid, my_hostname, mydest, msg);
					     }
					}
				   } else {
					irc_fprintf(user->user->sok, PRIVMSG, fd->nick, fd->userid, my_hostname, dest, msg);
				   }

			      }
			 }
			 user = user->next;
		    }
	       } else {
		    irc_fprintf(fd->sok, ERR_CANNOTSENDTOCHAN, NAME, fd->nick, list->name);
		    irc_fflush(fd->sok);
	       }
	       return;
	  }
	  list = list->next;
     }

     __no_such_channel(fd, dest);
}

/* void __send_cnotice(struct fd_list *fd, char *dest, char *msg)
   same as __send_cprivmsg(), except it sends a notice
 */

void internal_send_cnotice(struct fd_list *, char *, char *, int, int, int);

void __send_cnotice(struct fd_list *fd, char *dest, char *msg)
{
     internal_send_cnotice(fd, dest, msg, 0, 0, 1);
     return;
}

void __send_tnotice(struct fd_list *fd, char *dest, char *msg)
{
     int oper;
     int voice;
     char *d;

     d = dest;
     oper = 0;
     voice = 0;
     while (*d) {
	  switch (*d) {
	  case '+':
	       voice = 1;
	       break;
	  case '@':
	       oper = 1;
	       break;
	  case '#':
	       internal_send_cnotice(fd, d, msg, oper, voice, 0);
	       return;
	       break;
	  default:
	       break;
	  }
	  d++;
     }
     return;
}

void internal_send_cnotice(struct fd_list *fd, char *dest, char *msg, int oper, int voice, int all)
{
     struct chan_list *list;
     char *my_hostname;

     if (fd->hide_hostname) {
	  my_hostname = "hidden.hurricane.net";
     } else {
	  my_hostname = fd->hostname;
     }

     list = __clist_head->next;

     while (list != NULL) {
	  if (!strcasecmp(list->name, dest)) {
	       int ok = 0;
	       struct user_list *user;

	       if (list->no_outside_msg) {
		    struct user_list *t;

		    t = list->users;
		    while (t != NULL) {
			 if (t->user == fd) {
			      ok = 1;
			      t = NULL;
			 }
			 if (t != NULL)
			      t = t->next;
		    }
	       } else {
		    ok = 1;
	       }

	       if (ok) {
		    user = list->users;
		    while (user != NULL) {
			 if (user->user != fd) {
			      if (oper || voice) {
				   int sent = 0;
				   int i = 0;
				   char mydest[strlen(dest) + 3];

				   bzero(mydest, strlen(dest) + 3);
				   if (voice)
					mydest[i++] = '+';
				   if (oper)
					mydest[i++] = '@';
				   strcat(&(mydest[i]), dest);
				   if (oper && user->oper) {
					irc_fprintf(user->user->sok, NOTICE, fd->nick, fd->userid, my_hostname, mydest, msg);
					sent = 1;
				   }
				   if (voice && user->voice && !sent) {
					irc_fprintf(user->user->sok, NOTICE, fd->nick, fd->userid, my_hostname, mydest, msg);
				   }
			      } else {
				   irc_fprintf(user->user->sok, NOTICE, fd->nick, fd->userid, my_hostname, dest, msg);
			      }
			 }
			 user = user->next;
		    }
	       } else {
		    irc_fprintf(fd->sok, ERR_CANNOTSENDTOCHAN, NAME, fd->nick, list->name);
		    irc_fflush(fd->sok);
	       }
	       return;
	  }
	  list = list->next;
     }

     __no_such_channel(fd, dest);
}

/* void __part_channel(struct fd_list *fd, char *chan, int quit, char *msg, int kick, struct fd_list *kfd)
   if quit is set: displays a signoff message to the channel in chan containing the text of msg
   if kick is set: displays a kick message to the channel in chan, stating that kfd kicked the user, with the message msg
   otherwise, sends a part message to the channel
 */

void __part_channel(struct fd_list *fd, char *chan, int quit, char *msg, int kick, struct fd_list *kfd)
{
     struct user_chan_list *clist;
     char *my_hostname, *k_hostname;

     my_hostname = (fd->hide_hostname ? "hidden.hurricane.net" : fd->hostname);
     if (kick) {
	  k_hostname = (kfd->hide_hostname ? "hidden.hurricane.net" : fd->hostname);
     }
     log("entering __part_channel");
     clist = (struct user_chan_list *) (fd->chan_list);

     while (clist != NULL) {
	  {
	       char tmp[255];

	       snprintf(tmp, 255, "clist pointer = %p", (void *) clist);
	       log(tmp);
	       snprintf(tmp, 255, "chan name = %s", clist->chan->name);
	       log(tmp);
	  }
	  if (!strcasecmp(clist->chan->name, chan)) {
	       struct chan_list *t = clist->chan;
	       struct user_list *u = t->users;

	       while (u != NULL) {
		    if (u->user == fd) {
			 if (u->next == NULL && u->prev == NULL) {	/* only user in chan */
			      __free(u);
			      u = NULL;
			      if (!quit && !kick) {
				   irc_fprintf(fd->sok, ":%s!%s@%s PART %s\r\n", fd->nick, fd->userid, my_hostname,
					       t->name);
			      } else if (kick) {
				   irc_fprintf(fd->sok, ":%s!%s@%s KICK %s %s :%s\r\n", kfd->nick, kfd->userid,
					       k_hostname, t->name, fd->nick, msg);
			      }
			      if (t->invites) {
				   struct invite_list *q = t->invites;

				   while (q) {
					struct invite_list *r = q->next;
					struct user_invite_list *m;

					m = fd->invite_list;
					while (m) {
					     if (m->chan == t) {
						  if (m->prev == NULL && m->next == NULL) {
						       fd->invite_list = NULL;
						       gcDecRefcnt(m);
						       m = NULL;
						  } else if (m->prev == NULL && m->next != NULL) {
						       fd->invite_list = m->next;
						       (m->next)->prev = NULL;
						       gcDecRefcnt(m);
						       m = NULL;
						  } else if (m->prev != NULL && m->next == NULL) {
						       (m->prev)->next = NULL;
						       gcDecRefcnt(m);
						       m = NULL;
						  } else {
						       (m->prev)->next = m->next;
						       (m->next)->prev = m->prev;
						       gcDecRefcnt(m);
						       m = NULL;
						  }
					     }
					     if (m)
						  m = m->next;
					}
					gcDecRefcnt(q);
					q = r;
				   }
			      }
			      if (t->bans) {
				   struct ban_list *q = t->bans;

				   while (q) {
					struct ban_list *r = q->next;

					gcDecRefcnt(q);
					q = r;
				   }
			      }
			      if (t->exempts) {
				   struct ban_list *q = t->exempts;

				   while (q) {
					struct ban_list *r = q->next;

					gcDecRefcnt(q);
					q = r;
				   }
			      }
			      if (t->prev == NULL && t->next == NULL) {
				   gcDecRefcnt(t);
				   t = NULL;
			      } else if (t->prev != NULL && t->next == NULL) {
				   (t->prev)->next = NULL;
				   gcDecRefcnt(t);
				   t = NULL;
			      } else {
				   (t->prev)->next = t->next;
				   (t->next)->prev = t->prev;
				   gcDecRefcnt(t);
				   t = NULL;
			      }

			      if (clist->next == NULL && clist->prev == NULL) {		/* only channel we're in */
				   gcDecRefcnt(clist);
				   clist = NULL;
				   fd->chan_list = NULL;
			      } else if (clist->next == NULL && clist->prev != NULL) {
				   (clist->prev)->next = NULL;
				   gcDecRefcnt(clist);
				   clist = NULL;
			      } else if (clist->next != NULL && clist->prev == NULL) {
				   (clist->next)->prev = NULL;
				   fd->chan_list = (void *) (clist->next);
				   gcDecRefcnt(clist);
				   clist = NULL;
			      } else {
				   (clist->prev)->next = clist->next;
				   (clist->next)->prev = clist->prev;
				   gcDecRefcnt(clist);
				   clist = NULL;
			      }
			 } else if (u->next == NULL) {
			      (u->prev)->next = NULL;
			      gcDecRefcnt(u);
			      u = NULL;

			      if (!quit && !kick) {
				   struct user_list *u;

				   u = t->users;
				   while (u != NULL) {
					irc_fprintf(u->user->sok, ":%s!%s@%s PART %s\r\n", fd->nick,
						    fd->userid, my_hostname, chan);
					irc_fflush(u->user->sok);
					u = u->next;
				   }
				   irc_fprintf(fd->sok, ":%s!%s@%s PART %s\r\n", fd->nick, fd->userid,
					       my_hostname, chan);
			      } else if (quit) {
				   struct user_list *u;

				   u = t->users;
				   while (u != NULL) {
					irc_fprintf(u->user->sok, ":%s!%s@%s QUIT :%s\r\n", fd->nick,
						    fd->userid, my_hostname, msg);
					irc_fflush(u->user->sok);
					u = u->next;
				   }
			      } else {
				   struct user_list *u;

				   u = t->users;
				   while (u != NULL) {
					irc_fprintf(u->user->sok, ":%s!%s@%s KICK %s %s :%s\r\n",
						    kfd->nick, kfd->userid, k_hostname, t->name,
						    fd->nick, msg);
					irc_fflush(u->user->sok);
					u = u->next;
				   }
				   irc_fprintf(fd->sok, ":%s!%s@%s KICK %s %s :%s\r\n", kfd->nick, kfd->userid,
					       k_hostname, t->name, fd->nick, msg);
			      }
			      if (clist->next == NULL && clist->prev == NULL) {		/* only channel we're in */
				   gcDecRefcnt(clist);
				   clist = NULL;
				   fd->chan_list = NULL;
			      } else if (clist->next == NULL && clist->prev != NULL) {
				   (clist->prev)->next = NULL;
				   gcDecRefcnt(clist);
				   clist = NULL;
			      } else if (clist->next != NULL && clist->prev == NULL) {
				   (clist->next)->prev = NULL;
				   fd->chan_list = (void *) (clist->next);
				   gcDecRefcnt(clist);
				   clist = NULL;
			      } else {
				   (clist->prev)->next = clist->next;
				   (clist->next)->prev = clist->prev;
				   gcDecRefcnt(clist);
				   clist = NULL;
			      }

			 } else if (u->prev == NULL) {
			      (u->next)->prev = NULL;
			      t->users = u->next;
			      gcDecRefcnt(u);
			      u = NULL;

			      if (!quit && !kick) {
				   struct user_list *u;

				   u = t->users;
				   while (u != NULL) {
					irc_fprintf(u->user->sok, ":%s!%s@%s PART %s\r\n", fd->nick,
						    fd->userid, my_hostname, chan);
					irc_fflush(u->user->sok);
					u = u->next;
				   }
				   irc_fprintf(fd->sok, ":%s!%s@%s PART %s\r\n", fd->nick, fd->userid,
					       my_hostname, chan);
			      } else if (quit) {
				   struct user_list *u;

				   u = t->users;
				   while (u != NULL) {
					irc_fprintf(u->user->sok, ":%s!%s@%s QUIT :%s\r\n", fd->nick,
						    fd->userid, my_hostname, msg);
					irc_fflush(u->user->sok);
					u = u->next;
				   }
			      } else {
				   struct user_list *u;

				   u = t->users;
				   while (u != NULL) {
					irc_fprintf(u->user->sok, ":%s!%s@%s KICK %s %s :%s\r\n",
						    kfd->nick, kfd->userid, k_hostname, t->name,
						    fd->nick, msg);
					irc_fflush(u->user->sok);
					u = u->next;
				   }
				   irc_fprintf(fd->sok, ":%s!%s@%s KICK %s %s :%s\r\n", kfd->nick, kfd->userid,
					       k_hostname, t->name, fd->nick, msg);

			      }
			      if (clist->next == NULL && clist->prev == NULL) {		/* only channel we're in */
				   gcDecRefcnt(clist);
				   clist = NULL;
				   fd->chan_list = NULL;
			      } else if (clist->next == NULL && clist->prev != NULL) {
				   (clist->prev)->next = NULL;
				   gcDecRefcnt(clist);
				   clist = NULL;
			      } else if (clist->next != NULL && clist->prev == NULL) {
				   (clist->next)->prev = NULL;
				   fd->chan_list = (void *) (clist->next);
				   gcDecRefcnt(clist);
				   clist = NULL;
			      } else {
				   (clist->prev)->next = clist->next;
				   (clist->next)->prev = clist->prev;
				   gcDecRefcnt(clist);
				   clist = NULL;
			      }

			 } else {
			      (u->prev)->next = u->next;
			      (u->next)->prev = u->prev;
			      gcDecRefcnt(u);
			      u = NULL;

			      if (!quit && !kick) {
				   struct user_list *u;

				   u = t->users;
				   while (u != NULL) {
					irc_fprintf(u->user->sok, ":%s!%s@%s PART %s\r\n", fd->nick,
						    fd->userid, my_hostname, chan);
					irc_fflush(u->user->sok);
					u = u->next;
				   }
				   irc_fprintf(fd->sok, ":%s!%s@%s PART %s\r\n", fd->nick, fd->userid,
					       my_hostname, chan);
			      } else if (quit) {
				   struct user_list *u;

				   u = t->users;
				   while (u != NULL) {
					irc_fprintf(u->user->sok, ":%s!%s@%s QUIT :%s\r\n", fd->nick,
						    fd->userid, my_hostname, msg);
					irc_fflush(u->user->sok);
					u = u->next;
				   }
			      } else {
				   struct user_list *u;

				   u = t->users;
				   while (u != NULL) {
					irc_fprintf(u->user->sok, ":%s!%s@%s KICK %s %s :%s\r\n",
						    kfd->nick, kfd->userid, k_hostname, t->name,
						    fd->nick, msg);
					irc_fflush(u->user->sok);
					u = u->next;
				   }
				   irc_fprintf(fd->sok, ":%s!%s@%s KICK %s %s :%s\r\n", kfd->nick, kfd->userid,
					       k_hostname, t->name, fd->nick, msg);

			      }

			      if (clist->next == NULL && clist->prev == NULL) {		/* only channel we're in */
				   gcDecRefcnt(clist);
				   clist = NULL;
				   fd->chan_list = NULL;
			      } else if (clist->next == NULL && clist->prev != NULL) {
				   (clist->prev)->next = NULL;
				   gcDecRefcnt(clist);
				   clist = NULL;
			      } else if (clist->next != NULL && clist->prev == NULL) {
				   (clist->next)->prev = NULL;
				   fd->chan_list = (void *) (clist->next);
				   gcDecRefcnt(clist);
				   clist = NULL;
			      } else {
				   (clist->prev)->next = clist->next;
				   (clist->next)->prev = clist->prev;
				   gcDecRefcnt(clist);
				   clist = NULL;
			      }

			 }
		    }
		    if (u != NULL)
			 u = u->next;
	       }
	  }
	  if (clist != NULL)
	       clist = clist->next;
     }
}

/* void __names_reply(struct fd_list *fd, struct chan_list *chan)
   sends a reply for the NAMES command to the user in fd for the channel chan
 */

void __names_reply(struct fd_list *fd, struct chan_list *chan)
{
     struct user_list *user;
     char *buf;

     user = chan->users;
     if (user->oper || user->voice) {
	  buf = (char *) malloc(strlen(user->user->nick) + 3);
	  sprintf(buf, "%c%s", (user->oper ? '@' : '+'), user->user->nick);
     } else {
	  buf = (char *) malloc(strlen(user->user->nick) + 2);
	  sprintf(buf, "%s", user->user->nick);
     }

     user = user->next;
     while (user != NULL) {
	  if (user->oper || user->voice) {
	       buf = (char *) realloc((void *) buf, strlen(buf) + strlen(user->user->nick) + 3);
	       sprintf(buf, "%s %c%s", buf, (user->oper ? '@' : '+'), user->user->nick);
	  } else {
	       buf = (char *) realloc((void *) buf, strlen(buf) + strlen(user->user->nick) + 2);
	       sprintf(buf, "%s %s", buf, user->user->nick);
	  }
	  user = user->next;
     }

     irc_fprintf(fd->sok, ":%s 353 %s = %s :%s\r\n", NAME, fd->nick, chan->name, buf);
     irc_fprintf(fd->sok, ":%s 366 %s %s :End of /NAMES list.\r\n", NAME, fd->nick, chan->name);
     irc_fflush(fd->sok);
     free(buf);
     return;
}

/* void __join_existing_channel(struct fd_list *fd, struct chan_list *chan, char **av, int cp)
   joins the channel in chan, checking the av array for a match to the key to the channel if it is keyed,
   and determines whether to allow the user to join the channel or not
 */

void __join_existing_channel(struct fd_list *fd, struct chan_list *chan, char **av, int cp)
{
     struct user_list *user, *cuser;
     int ok;
     char *my_hostname;

     if (fd->hide_hostname) {
	  my_hostname = "hidden.hurricane.net";
     } else {
	  my_hostname = fd->hostname;
     }

     log("entering __join_existing_channel");
     if (chan->key) {
	  char *key_val;
	  int m;

	  if (cp == 0) {
	       key_val = av[2];
	  } else {
	       char **kv, *kptr, *ktmp;
	       int kc;

	       kc = 0;
	       if (av[2] == NULL) {
		    key_val = NULL;
	       } else {
		    ktmp = (char *) malloc(strlen(av[2]));
		    bzero(ktmp, strlen(av[2]));
		    strcpy(ktmp, av[2]);

		    kv = NULL;
		    kptr = ktmp;
		    while (*ktmp != 0) {
			 if (*ktmp == ',') {
			      *ktmp = 0;
			      kv = (char **) realloc(kv, sizeof(char **) * ++kc);

			      kv[kc - 1] = kptr;
			      ktmp++;
			      kptr = ktmp;
			 } else {
			      ktmp++;
			 }
		    }

		    if (kc < cp) {
			 key_val = NULL;
			 free(kv[0]);
			 free(kv);
		    } else {
			 m = 1;
			 key_val = (char *) calloc(strlen(kv[cp]));
			 strcpy(key_val, kv[cp]);
			 free(kv[0]);
			 free(kv);
		    }
	       }
	  }

	  log("channel key set");

	  if (key_val != NULL) {
	       if (!strcasecmp(key_val, chan->key_val)) {
		    log("passed key valid");
		    if (m)
			 __free(key_val);
		    ok = 1;
	       } else {
		    log("bad key");
		    if (m)
			 __free(key_val);
		    ok = 0;
	       }
	  } else {
	       log("no key passed");
	       ok = 0;
	  }
     } else {
	  log("no key set");
	  ok = 1;
     }

     if (chan->limit) {
	  struct user_list *u;
	  int nusers = 0;

	  u = chan->users;

	  while (u != NULL) {
	       nusers++;
	       u = u->next;
	  }

	  if (nusers >= chan->limit_val) {
	       ok = 0;
	  } else {
	       if (ok) {
		    ok = 1;
	       }
	  }
     } else {
	  if (ok)
	       ok = 1;
     }

     if (chan->invite_only) {
	  struct invite_list *invites = chan->invites;

	  log("channel is invite only..");
	  if (invites == NULL) {
	       log("invite list is null");
	       ok = 0;
	  } else {
	       if (ok == 1) {
		    while (invites != NULL) {
			 if (invites->user == fd) {
			      ok = 1;
			      if (invites->prev == NULL && invites->next == NULL) {
				   struct user_invite_list *p;

				   p = (struct user_invite_list *) invites->user->invite_list;
				   while (p != NULL) {
					if (p->chan == chan) {
					     if (p->prev == NULL && p->next == NULL) {
						  gcDecRefcnt(p);
						  p = NULL;
						  invites->user->invite_list = NULL;
					     } else if (p->prev == NULL && p->next != NULL) {
						  (p->next)->prev = NULL;
						  invites->user->invite_list = p->next;
						  gcDecRefcnt(p);
						  p = NULL;
					     } else if (p->prev != NULL && p->next == NULL) {
						  (p->prev)->next = NULL;
						  gcDecRefcnt(p);
						  p = NULL;
					     } else {
						  (p->prev)->next = p->next;
						  (p->next)->prev = p->prev;
						  gcDecRefcnt(p);
						  p = NULL;
					     }
					}
					if (p != NULL)
					     p = p->next;
				   }
				   chan->invites = NULL;
				   gcDecRefcnt(invites);
			      } else if (invites->prev == NULL && invites->next != NULL) {
				   struct user_invite_list *p;

				   p = (struct user_invite_list *) invites->user->invite_list;
				   while (p != NULL) {
					if (p->chan == chan) {
					     if (p->prev == NULL && p->next == NULL) {
						  gcDecRefcnt(p);
						  p = NULL;
						  invites->user->invite_list = NULL;
					     } else if (p->prev == NULL && p->next != NULL) {
						  (p->next)->prev = NULL;
						  invites->user->invite_list = p->next;
						  gcDecRefcnt(p);
						  p = NULL;
					     } else if (p->prev != NULL && p->next == NULL) {
						  (p->prev)->next = NULL;
						  gcDecRefcnt(p);
						  p = NULL;
					     } else {
						  (p->prev)->next = p->next;
						  (p->next)->prev = p->prev;
						  gcDecRefcnt(p);
						  p = NULL;
					     }
					}
					if (p != NULL)
					     p = p->next;
				   }

				   chan->invites = invites->next;
				   (invites->next)->prev = NULL;
				   gcDecRefcnt(invites);
			      } else if (invites->prev != NULL && invites->next == NULL) {
				   struct user_invite_list *p;

				   p = (struct user_invite_list *) invites->user->invite_list;
				   while (p != NULL) {
					if (p->chan == chan) {
					     if (p->prev == NULL && p->next == NULL) {
						  gcDecRefcnt(p);
						  p = NULL;
						  invites->user->invite_list = NULL;
					     } else if (p->prev == NULL && p->next != NULL) {
						  (p->next)->prev = NULL;
						  invites->user->invite_list = p->next;
						  gcDecRefcnt(p);
						  p = NULL;
					     } else if (p->prev != NULL && p->next == NULL) {
						  (p->prev)->next = NULL;
						  gcDecRefcnt(p);
						  p = NULL;
					     } else {
						  (p->prev)->next = p->next;
						  (p->next)->prev = p->prev;
						  gcDecRefcnt(p);
						  p = NULL;
					     }
					}
					if (p != NULL)
					     p = p->next;
				   }

				   (invites->prev)->next = NULL;
				   gcDecRefcnt(invites);
			      } else {
				   struct user_invite_list *p;

				   p = (struct user_invite_list *) invites->user->invite_list;
				   while (p != NULL) {
					if (p->chan == chan) {
					     if (p->prev == NULL && p->next == NULL) {
						  gcDecRefcnt(p);
						  p = NULL;
						  invites->user->invite_list = NULL;
					     } else if (p->prev == NULL && p->next != NULL) {
						  (p->next)->prev = NULL;
						  invites->user->invite_list = p->next;
						  gcDecRefcnt(p);
						  p = NULL;
					     } else if (p->prev != NULL && p->next == NULL) {
						  (p->prev)->next = NULL;
						  gcDecRefcnt(p);
						  p = NULL;
					     } else {
						  (p->prev)->next = p->next;
						  (p->next)->prev = p->prev;
						  gcDecRefcnt(p);
						  p = NULL;
					     }
					}
					if (p != NULL)
					     p = p->next;
				   }

				   (invites->prev)->next = invites->next;
				   (invites->next)->prev = invites->prev;
				   gcDecRefcnt(invites);
			      }
			      invites = NULL;
			 }
			 if (invites != NULL)
			      invites = invites->next;
		    }
	       }
	  }
	  if (ok == 0) {
	       struct ban_list *perminvites = chan->perminvites;
	       char *userhost = alloca(200);

	       bzero(userhost, 200);
	       snprintf(userhost, 200, "%s!%s@%s", fd->nick, fd->userid, my_hostname);
	       while (perminvites) {
		    if (!match(perminvites->banid, userhost)) {
			 ok = 1;
			 perminvites = NULL;
		    }
		    if (perminvites)
			 perminvites = perminvites->next;
	       }
	  }
     } else {
	  if (ok == 1)
	       ok = 1;
     }

     if (ok) {
	  struct ban_list *bans;
	  char *userhost = alloca(200);

	  bzero(userhost, 200);

	  snprintf(userhost, 200, "%s!%s@%s", fd->nick, fd->userid, my_hostname);

	  bans = chan->bans;
	  log("checking banlist");

	  while (bans != NULL) {
	       if (!match(bans->banid, userhost)) {
		    ok = 0;	/* match, don't let user in */
		    bans = NULL;
	       }
	       if (bans)
		    bans = bans->next;
	  }
     }
     if (!ok) {
	  struct ban_list *exempts;
	  char *userhost = alloca(200);

	  bzero(userhost, 200);

	  snprintf(userhost, 200, "%s!%s@%s", fd->nick, fd->userid, my_hostname);
	  exempts = chan->exempts;
	  while (exempts) {
	       if (!match(exempts->banid, userhost)) {
		    ok = 1;
		    exempts = NULL;
	       }
	       if (exempts)
		    exempts = exempts->next;
	  }
     } {
	  struct user_list *u = chan->users;

	  while (u) {
	       if (u->user == fd) {
		    return;
	       }
	       u = u->next;
	  }
     }

     if (ok) {
	  user = (struct user_list *) gcAlloc(sizeof(struct user_list));

	  bzero(user, sizeof(struct user_list));

	  user->magic = USER_LIST_MAGIC;

	  gcIncRefcnt(fd);
	  user->user = fd;
	  cuser = chan->users;
	  while (cuser->next != NULL)
	       cuser = cuser->next;

	  cuser->next = user;
	  user->prev = cuser;

	  cuser = chan->users;
	  while (cuser != NULL) {
	       irc_fprintf(cuser->user->sok, RPL_JOIN, fd->nick, fd->userid, my_hostname, chan->name);
	       irc_fflush(cuser->user->sok);
	       cuser = cuser->next;
	  }

	  if (chan->topic[0] == 0) {
	       irc_fprintf(fd->sok, RPL_NOTOPIC, NAME, fd->nick, chan->name);
	  } else {
	       irc_fprintf(fd->sok, RPL_TOPIC, NAME, fd->nick, chan->name, chan->topic);
	       irc_fprintf(fd->sok, RPL_TOPICSET, NAME, fd->nick, chan->name, chan->topic_set_by, chan->topic_set_at);
	  }

	  cuser = chan->users;

	  __names_reply(fd, chan);

	  irc_fflush(fd->sok);

	  {
	       struct user_chan_list *t, *new_ent;
	       new_ent = (struct user_chan_list *) gcAlloc(sizeof(struct user_chan_list));

	       bzero(new_ent, sizeof(struct user_chan_list));

	       new_ent->magic = USER_CHAN_LIST_MAGIC;
	       gcIncRefcnt(chan);
	       new_ent->chan = chan;
	       if ((struct user_chan_list *) (fd->chan_list) == NULL) {
		    fd->chan_list = (void *) new_ent;
	       } else {
		    t = (struct user_chan_list *) (fd->chan_list);
		    while (t->next != NULL)
			 t = t->next;
		    t->next = new_ent;
		    new_ent->prev = t;
	       }
	  }
     } else {
	  if (chan->key) {
	       irc_fprintf(fd->sok, ERR_BADCHANNELKEY, NAME, fd->nick, chan->name);
	  } else if (chan->limit) {
	       irc_fprintf(fd->sok, ERR_CHANNELISFULL, NAME, fd->nick, chan->name);
	  } else if (chan->invite_only) {
	       irc_fprintf(fd->sok, ERR_INVITEONLYCHAN, NAME, fd->nick, chan->name);
	  } else {
	       irc_fprintf(fd->sok, ERR_BANNEDFROMCHAN, NAME, fd->nick, chan->name);
	  }
	  irc_fflush(fd->sok);
     }

     return;
}

/* void __join_new_channel(struct fd_list *fd, char *name)
   creates a new channel, putting the user in fd into the channel
 */

void __join_new_channel(struct fd_list *fd, char *name)
{
     struct chan_list *new_ent;
     struct user_list *user;
     char *my_hostname;

     my_hostname = (fd->hide_hostname ? "hidden.hurricane.net" : fd->hostname);

     new_ent = (struct chan_list *) gcAlloc(sizeof(struct chan_list));

     new_ent->magic = CHAN_LIST_MAGIC;

     strncpy(new_ent->name, name, 200);
     new_ent->chan_created_at = __time(NULL);
     user = (struct user_list *) gcAlloc(sizeof(struct user_list));

     user->magic = USER_LIST_MAGIC;
     gcIncRefcnt(fd);
     user->user = fd;
     user->oper = 1;
     new_ent->users = user;

     irc_fprintf(fd->sok, RPL_JOIN, fd->nick, fd->userid, my_hostname, name);

     __names_reply(fd, new_ent);

     irc_fprintf(fd->sok, RPL_NOTOPIC, NAME, fd->nick, name);
     irc_fflush(fd->sok);

     {
	  struct chan_list *t;

	  t = __clist_head;
	  while (t->next != NULL)
	       t = t->next;
	  t->next = new_ent;
	  new_ent->prev = t;
     }

     {
	  struct user_chan_list *t, *_new_ent;
	  _new_ent = (struct user_chan_list *) gcAlloc(sizeof(struct user_chan_list));

	  _new_ent->magic = USER_CHAN_LIST_MAGIC;
	  gcIncRefcnt(new_ent);
	  _new_ent->chan = new_ent;
	  if ((struct user_chan_list *) (fd->chan_list) == NULL) {
	       fd->chan_list = (void *) _new_ent;
	  } else {
	       t = (struct user_chan_list *) (fd->chan_list);
	       while (t->next != NULL)
		    t = t->next;
	       t->next = _new_ent;
	       _new_ent->prev = t;
	  }
     }

     return;
}

struct cmodelist {
     char mode;
     short int mval;
};

struct cmodelist cmlist[] =
{
     {'i', 0x0001},
     {'k', 0x0002},
     {'l', 0x0004},
     {'m', 0x0008},
     {'n', 0x0010},
     {'p', 0x0020},
     {'s', 0x0040},
     {'t', 0x0080},
     {0, 0},
};

const int numcmodes = 8;

/* char *get_cmode(struct chan_list *chan)
   returns the channel mode for the channel chan, with modes that do not require any parameters (no b/e/I/o/v modes)
 */

char *get_cmode(struct chan_list *chan)
{
     static char mstr[20];
     int is_limit = 0;
     int mode = 0, k = 0, n;

     mstr[k++] = '+';

     if (chan->oper_only_topic)
	  mode |= 0x0080;
     if (chan->no_outside_msg)
	  mode |= 0x0010;
     if (chan->private)
	  mode |= 0x0020;
     if (chan->secret)
	  mode |= 0x0040;
     if (chan->invite_only)
	  mode |= 0x0001;
     if (chan->mute)
	  mode |= 0x0008;
     if (chan->key)
	  mode |= 0x0002;
     if (chan->limit) {
	  mode |= 0x0004;
          is_limit = 1;
     }
     for (n = 0; n < numcmodes; n++) {
	  if (mode & cmlist[n].mval) {
	       mstr[k++] = cmlist[n].mode;
	  }
     }

     if (is_limit) {
	char nbuf[15];
	snprintf(nbuf, 15, "%i", chan->limit_val);
        n = 0;
        mstr[k++] = ' ';
        for(n=0; n<strlen(nbuf); n++) {
	  mstr[k++] = nbuf[n];
        }
     }
     mstr[k] = 0;
     return mstr;
}

#undef RPL_LIST
#define RPL_LIST ":%s 322 %s %s %i :[%s] %s\r\n"

/* void __list_all_channels(struct fd_list *fd)
   returns a listing of all non-private/secret channels to the user in fd, and all channels that the user is
   currently a member of (whether +p/+s or not)
 */

void __list_all_channels(struct fd_list *fd)
{
     struct chan_list *list;

     list = __clist_head->next;

     irc_fprintf(fd->sok, RPL_LISTSTART, NAME, fd->nick);
     while (list != NULL) {
	  struct user_list *l;
	  int nusers;
	  int in_chan = 0;

	  {
	       char tmp[255];

	       snprintf(tmp, 255, "list ptr is %p", (void *) list);
	       log(tmp);
	  }
	  l = list->users;
	  nusers = 0;
	  while (l != NULL) {
	       nusers++;
	       if (l->user == fd) {
		    in_chan = 1;
	       }
	       l = l->next;
	  }
	  if ((list->private || list->secret) && !fd->oper) {
	       if (in_chan) {
		    irc_fprintf(fd->sok, RPL_LIST, NAME, fd->nick, list->name, nusers, get_cmode(list), list->topic);
	       } else {
		    if (list->private) {
			 irc_fprintf(fd->sok, RPL_LIST_PRV, NAME, fd->nick, nusers);
		    }
	       }
	  } else {
	       irc_fprintf(fd->sok, RPL_LIST, NAME, fd->nick, list->name, nusers, get_cmode(list), list->topic);
	  }
	  list = list->next;
     }

     irc_fprintf(fd->sok, RPL_LISTEND, NAME, fd->nick);
     irc_fflush(fd->sok);
     return;
}

/* void __list_channel(struct fd_list *fd, char *chan_name)
   returns a listing of only the channel specified in *chan_name to fd
 */

void __list_channel(struct fd_list *fd, char *chan_name)
{
     struct chan_list *list;
     int list_started = 0;

     list = __clist_head->next;

     while (list != NULL) {
	  if (!strcasecmp(list->name, chan_name)) {
	       struct user_list *l;
	       int nusers;

	       l = list->users;
	       nusers = 0;
	       while (l != NULL) {
		    nusers++;
		    l = l->next;
	       }
	       if (!list_started)
		    irc_fprintf(fd->sok, RPL_LISTSTART, NAME, fd->nick);
	       if (!list_started)
		    list_started = 1;
	       irc_fprintf(fd->sok, RPL_LIST, NAME, fd->nick, list->name, nusers, get_cmode(list), list->topic);
	  }
	  list = list->next;
     }

     if (!list_started)
	  __no_such_channel(fd, chan_name);
     else
	  irc_fprintf(fd->sok, RPL_LISTEND, NAME, fd->nick);
     irc_fflush(fd->sok);
     return;
}


#define RPL_WHOREPLY ":%s 352 %s %s %s %s %s %s %c%s :%i %s\r\n"
#define RPL_ENDOFWHO ":%s 315 %s %s :End of /WHO list\r\n"

/* void __who_channel(struct fd_list *fd, char *chan_name)
   returns a listing of the users in a channel
 */

void __who_channel(struct fd_list *fd, char *chan_name)
{
     struct chan_list *clist;
     char *my_hostname;

     clist = __clist_head;

     while (clist != NULL) {
	  if (!strcasecmp(clist->name, chan_name)) {
	       struct user_list *ulist;

	       ulist = clist->users;
	       while (ulist != NULL) {
		    my_hostname = (ulist->user->hide_hostname ? "hidden.hurricane.net" : ulist->user->hostname);
		    irc_fprintf(fd->sok, RPL_WHOREPLY, NAME, fd->nick, clist->name, ulist->user->userid,
				my_hostname, NAME, ulist->user->nick, 'H',
		       (ulist->oper ? "@" : (ulist->voice ? "+" : "")), 1, ulist->user->realname);
		    ulist = ulist->next;
	       }
	       irc_fprintf(fd->sok, RPL_ENDOFWHO, NAME, fd->nick, clist->name);
	       irc_fflush(fd->sok);
	       return;
	  }
	  clist = clist->next;
     }
     __no_such_channel(fd, chan_name);
     return;
}

/* void __add_channel_ban(struct fd_list *fd, struct chan_list *chan, char *banid)
   sets a +b mode on the channel chan with the parameter banid
 */

void __add_channel_ban(struct fd_list *fd, struct chan_list *chan, char *banid)
{
     char *my_hostname = (fd->hide_hostname ? "hidden.hurricane.net" : fd->hostname);

     if (banid == NULL) {
	  return;
     } else {
	  struct ban_list *new_ent;
	  char usermask[120], pretty_mask[120];
	  new_ent = (struct ban_list *) gcAlloc(sizeof(struct ban_list));

	  new_ent->magic = BAN_LIST_MAGIC;
	  bzero(pretty_mask, 120);
	  make_pretty_mask(banid, pretty_mask, 120);
	  strncpy(new_ent->banid, pretty_mask, 75);
	  snprintf(usermask, 120, "%s!%s@%s", fd->nick, fd->userid, my_hostname);
	  strncpy(new_ent->userid, usermask, 120);
	  new_ent->time = __time(NULL);
	  if (chan->bans == NULL) {
	       chan->bans = new_ent;
	  } else {
	       struct ban_list *bans;

	       bans = chan->bans;
	       while (bans->next != NULL)
		    bans = bans->next;
	       bans->next = new_ent;
	       new_ent->prev = bans;
	  }
	  {
	       struct user_list *u;

	       u = chan->users;
	       while (u != NULL) {
		    irc_fprintf(u->user->sok, ":%s!%s@%s MODE %s +b %s\r\n", fd->nick, fd->userid, my_hostname,
				chan->name, pretty_mask);
		    irc_fflush(u->user->sok);
		    u = u->next;
	       }
	  }
     }
}

/* void __del_channel_ban(struct fd_list *fd, struct chan_list *chan, char *banid)
   deletes a ban from the channel in chan
 */

void __del_channel_ban(struct fd_list *fd, struct chan_list *chan, char *banid)
{
     char *my_hostname = (fd->hide_hostname ? "hidden.hurricane.net" : fd->hostname);

     if (banid == NULL) {
	  return;
     } else {
	  struct ban_list *bans;

	  bans = chan->bans;
	  while (bans != NULL) {
	       if (!strcasecmp(bans->banid, banid)) {
		    if (bans->prev == NULL) {
			 if (bans->next == NULL) {
			      gcDecRefcnt(bans);
			      bans = NULL;
			      chan->bans = NULL;
			 } else {
			      chan->bans = bans->next;
			      chan->bans->prev = NULL;
			      gcDecRefcnt(bans);
			      bans = NULL;
			 }
		    } else {
			 if (bans->next == NULL) {
			      (bans->prev)->next = NULL;
			      gcDecRefcnt(bans);
			      bans = NULL;
			 } else {
			      (bans->prev)->next = bans->next;
			      (bans->next)->prev = bans->prev;
			      gcDecRefcnt(bans);
			      bans = NULL;
			 }
		    }
	       } {
		    struct user_list *u;

		    u = chan->users;
		    while (u != NULL) {
			 irc_fprintf(u->user->sok, ":%s!%s@%s MODE %s -b %s\r\n", fd->nick, fd->userid, my_hostname,
				     chan->name, banid);
			 irc_fflush(u->user->sok);
			 u = u->next;
		    }
	       }

	       if (bans != NULL)
		    bans = bans->next;
	  }
     }
}

/* void __add_channel_exempt()
   void __del_channel_exempt()
   sets/deletes +e modes on a channel
   void __add_channel_perminvite()
   void __del_channel_perminvite()
   sets/deletes +I modes on a channel
 */

void __add_channel_exempt(struct fd_list *fd, struct chan_list *chan, char *banid)
{
     char *my_hostname = (fd->hide_hostname ? "hidden.hurricane.net" : fd->hostname);

     if (banid == NULL) {
	  return;
     } else {
	  struct ban_list *new_ent;
	  char usermask[120], pretty_mask[120];
	  new_ent = (struct ban_list *) gcAlloc(sizeof(struct ban_list));

	  new_ent->magic = BAN_LIST_MAGIC;
	  bzero(pretty_mask, 120);
	  make_pretty_mask(banid, pretty_mask, 120);
	  strncpy(new_ent->banid, pretty_mask, 75);
	  snprintf(usermask, 120, "%s!%s@%s", fd->nick, fd->userid, my_hostname);
	  strncpy(new_ent->userid, usermask, 120);
	  new_ent->time = __time(NULL);
	  if (chan->exempts == NULL) {
	       chan->exempts = new_ent;
	  } else {
	       struct ban_list *bans;

	       bans = chan->exempts;
	       while (bans->next != NULL)
		    bans = bans->next;
	       bans->next = new_ent;
	       new_ent->prev = bans;
	  }
	  {
	       struct user_list *u;

	       u = chan->users;
	       while (u != NULL) {
		    irc_fprintf(u->user->sok, ":%s!%s@%s MODE %s +e %s\r\n", fd->nick,
				fd->userid, my_hostname, chan->name, pretty_mask);
		    irc_fflush(u->user->sok);
		    u = u->next;
	       }
	  }
     }
}

void __del_channel_exempt(struct fd_list *fd, struct chan_list *chan, char *banid)
{
     char *my_hostname = (fd->hide_hostname ? "hidden.hurricane.net" : fd->hostname);

     if (banid == NULL) {
	  return;
     } else {
	  struct ban_list *bans;

	  bans = chan->exempts;
	  while (bans != NULL) {
	       if (!strcasecmp(bans->banid, banid)) {
		    if (bans->prev == NULL) {
			 if (bans->next == NULL) {
			      gcDecRefcnt(bans);
			      bans = NULL;
			      chan->exempts = NULL;
			 } else {
			      chan->exempts = bans->next;
			      chan->exempts->prev = NULL;
			      gcDecRefcnt(bans);
			      bans = NULL;
			 }
		    } else {
			 if (bans->next == NULL) {
			      (bans->prev)->next = NULL;
			      gcDecRefcnt(bans);
			      bans = NULL;
			 } else {
			      (bans->prev)->next = bans->next;
			      (bans->next)->prev = bans->prev;
			      gcDecRefcnt(bans);
			      bans = NULL;
			 }
		    }
	       } {
		    struct user_list *u;

		    u = chan->users;
		    while (u != NULL) {
			 irc_fprintf(u->user->sok, ":%s!%s@%s MODE %s -e %s\r\n", fd->nick,
				     fd->userid, my_hostname, chan->name, banid);
			 irc_fflush(u->user->sok);
			 u = u->next;
		    }
	       }

	       if (bans != NULL)
		    bans = bans->next;
	  }
     }
}

void __add_channel_perminvite(struct fd_list *fd, struct chan_list *chan, char *banid)
{
     char *my_hostname = (fd->hide_hostname ? "hidden.hurricane.net" : fd->hostname);

     if (banid == NULL) {
	  return;
     } else {
	  struct ban_list *new_ent;
	  char usermask[120], pretty_mask[120];
	  new_ent = (struct ban_list *) gcAlloc(sizeof(struct ban_list));

	  new_ent->magic = BAN_LIST_MAGIC;
	  bzero(pretty_mask, 120);
	  make_pretty_mask(banid, pretty_mask, 120);
	  strncpy(new_ent->banid, pretty_mask, 75);
	  snprintf(usermask, 120, "%s!%s@%s", fd->nick, fd->userid, my_hostname);
	  strncpy(new_ent->userid, usermask, 120);
	  new_ent->time = __time(NULL);
	  if (chan->perminvites == NULL) {
	       chan->perminvites = new_ent;
	  } else {
	       struct ban_list *bans;

	       bans = chan->perminvites;
	       while (bans->next != NULL)
		    bans = bans->next;
	       bans->next = new_ent;
	       new_ent->prev = bans;
	  }
	  {
	       struct user_list *u;

	       u = chan->users;
	       while (u != NULL) {
		    irc_fprintf(u->user->sok, ":%s!%s@%s MODE %s +I %s\r\n", fd->nick,
				fd->userid, my_hostname, chan->name, pretty_mask);
		    irc_fflush(u->user->sok);
		    u = u->next;
	       }
	  }
     }
}

void __del_channel_perminvite(struct fd_list *fd, struct chan_list *chan, char *banid)
{
     char *my_hostname = (fd->hide_hostname ? "hidden.hurricane.net" : fd->hostname);

     if (banid == NULL) {
	  return;
     } else {
	  struct ban_list *bans;

	  bans = chan->perminvites;
	  while (bans != NULL) {
	       if (!strcasecmp(bans->banid, banid)) {
		    if (bans->prev == NULL) {
			 if (bans->next == NULL) {
			      gcDecRefcnt(bans);
			      bans = NULL;
			      chan->perminvites = NULL;
			 } else {
			      chan->perminvites = bans->next;
			      chan->perminvites->prev = NULL;
			      gcDecRefcnt(bans);
			      bans = NULL;
			 }
		    } else {
			 if (bans->next == NULL) {
			      (bans->prev)->next = NULL;
			      gcDecRefcnt(bans);
			      bans = NULL;
			 } else {
			      (bans->prev)->next = bans->next;
			      (bans->next)->prev = bans->prev;
			      gcDecRefcnt(bans);
			      bans = NULL;
			 }
		    }
	       } {
		    struct user_list *u;

		    u = chan->users;
		    while (u != NULL) {
			 irc_fprintf(u->user->sok, ":%s!%s@%s MODE %s -I %s\r\n", fd->nick,
				     fd->userid, my_hostname, chan->name, banid);
			 irc_fflush(u->user->sok);
			 u = u->next;
		    }
	       }

	       if (bans != NULL)
		    bans = bans->next;
	  }
     }
}

/* void __set_chan_op(struct fd_list *fd, struct chan_list *chan, char *nick, int add)
   sets or clears the +o flag from the user in nick
 */

void __set_chan_op(struct fd_list *fd, struct chan_list *chan, char *nick, int add)
{
     char *my_hostname = (fd->hide_hostname ? "hidden.hurricane.net" : fd->hostname);

     log("entering __set_chan_op");
     if (add) {
	  struct user_list *t, *u;

	  log("adding operator to channel");

	  t = chan->users;

	  while (t != NULL) {
	       if (!strcasecmp(t->user->nick, nick)) {
		    if (t->oper) {
			 log("user is already oper, returning");
			 return;
		    } else {
			 log("making user an operator");
			 t->oper = 1;
			 log("notifying channel of mode change");
			 u = chan->users;
			 while (u != NULL) {
			      irc_fprintf(u->user->sok, RPL_MODECHANOPADD, fd->nick, fd->userid, my_hostname,
					  chan->name, t->user->nick);
			      irc_fflush(u->user->sok);
			      u = u->next;
			 }

			 log("returning");
			 return;
		    }
	       }
	       t = t->next;
	  }

	  log("user not found on channel");
	  irc_fprintf(fd->sok, ERR_USERNOTONCHANNEL, NAME, fd->nick, chan->name, nick);
	  irc_fflush(fd->sok);
	  return;
     } else {
	  struct user_list *t, *u;

	  log("removing operator from channel");

	  t = chan->users;
	  while (t != NULL) {
	       if (!strcasecmp(t->user->nick, nick)) {
		    if (!t->oper) {
			 log("user isn't an oper, returning");
			 return;
		    } else {
			if (t->user->service) {
				irc_fprintf(fd->sok, ":%s 484 %s %s %s :Cannot kill, kick, or deop channel service\r\n",
					server_name, fd->nick, t->user->nick, chan->name);
				return;
			} else {
				 log("removing operator status from user");
				 t->oper = 0;
				 log("notifying channel of mode change");
				 u = chan->users;
				 while (u != NULL) {
				      irc_fprintf(u->user->sok, RPL_MODECHANOPDEL, fd->nick, fd->userid, my_hostname,
						  chan->name, t->user->nick);
				      irc_fflush(u->user->sok);
				      u = u->next;
				 }
			}
			 log("returning");
			 return;
		    }
	       }
	       t = t->next;
	  }

	  log("user not on channel");
	  irc_fprintf(fd->sok, ERR_USERNOTONCHANNEL, NAME, fd->nick, chan->name, nick);
	  irc_fflush(fd->sok);
	  return;
     }
}

/* void __set_chan_voice() - same as __set_chan_op() except +v modes */

void __set_chan_voice(struct fd_list *fd, struct chan_list *chan, char *nick, int add)
{
     char *my_hostname = (fd->hide_hostname ? "hidden.hurricane.net" : fd->hostname);

     log("entering __set_chan_voice");
     if (add) {
	  struct user_list *t, *u;

	  log("adding voice to channel");

	  t = chan->users;

	  while (t != NULL) {
	       if (!strcasecmp(t->user->nick, nick)) {
		    if (t->voice) {
			 log("user is already voice, returning");
			 return;
		    } else {
			 log("making user voiced");
			 t->voice = 1;
			 log("notifying channel of mode change");
			 u = chan->users;
			 while (u != NULL) {
			      irc_fprintf(u->user->sok, RPL_MODECHANVOICEADD, fd->nick, fd->userid, my_hostname,
					  chan->name, t->user->nick);
			      irc_fflush(u->user->sok);
			      u = u->next;
			 }

			 log("returning");
			 return;
		    }
	       }
	       t = t->next;
	  }

	  log("user not found on channel");
	  irc_fprintf(fd->sok, ERR_USERNOTONCHANNEL, NAME, fd->nick, chan->name, nick);
	  irc_fflush(fd->sok);
	  return;
     } else {
	  struct user_list *t, *u;

	  log("removing voice from channel");

	  t = chan->users;
	  while (t != NULL) {
	       if (!strcasecmp(t->user->nick, nick)) {
		    if (!t->voice) {
			 log("user isn't voice, returning");
			 return;
		    } else {
			 log("removing voice from user");
			 t->voice = 0;
			 log("notifying channel of mode change");
			 u = chan->users;
			 while (u != NULL) {
			      irc_fprintf(u->user->sok, RPL_MODECHANVOICEDEL, fd->nick, fd->userid, my_hostname,
					  chan->name, t->user->nick);
			      irc_fflush(u->user->sok);
			      u = u->next;
			 }

			 log("returning");
			 return;
		    }
	       }
	       t = t->next;
	  }

	  log("user not on channel");
	  irc_fprintf(fd->sok, ERR_USERNOTONCHANNEL, NAME, fd->nick, chan->name, nick);
	  irc_fflush(fd->sok);
	  return;
     }
}

/* struct chan_mode_change *__cmode_parse(struct fd_list *fd, struct chan_list *chan, char *mode_str, char **av, int ac, int arg)
   parses a mode string and returns a structure representing the parsed mode
 */

struct chan_mode_change *__cmode_parse(struct fd_list *fd, struct chan_list *chan, char *mode_str, char **av, int ac, int arg)
{
     struct chan_mode_change *mode;
     int flag = 1;
     mode = (struct chan_mode_change *) calloc(sizeof(struct chan_mode_change));

     mode->private.orig = mode->private.new = chan->private;
     mode->secret.orig = mode->secret.new = chan->secret;
     mode->invite_only.orig = mode->invite_only.new = chan->invite_only;
     mode->oper_only_topic.orig = mode->oper_only_topic.new = chan->oper_only_topic;
     mode->no_outside_msg.orig = mode->no_outside_msg.new = chan->no_outside_msg;
     mode->mute.orig = mode->mute.new = chan->mute;
     mode->key.orig = mode->key.new = chan->key;
     mode->limit.orig = mode->limit.new = chan->limit;
     mode->limit.orig_val = mode->limit.new_val = chan->limit_val;

     while (*mode_str != 0) {
	  switch (*mode_str) {
	  case '+':
	       flag = 1;
	       break;
	  case '-':
	       flag = 0;
	       break;
	  case 'b':
	       {
		    char *banid;
		    FILE2 *log = irc_fopen("ircd.log", "a+");

		    irc_fprintf(log, "flag = %i. ac = %i. arg = %i. ", flag, ac, arg);
		    if (arg == ac) {
			 banid = NULL;
		    } else {
			 banid = av[arg++];
		    }
		    irc_fprintf(log, "banid = %s.\n", banid);
		    irc_fflush(log);
		    irc_fclose(log);
		    if (banid == NULL) {
			 __not_enough_parameters(fd, "MODE");
		    } else {
			 if (flag) {
			      __add_channel_ban(fd, chan, banid);
			 } else {
			      __del_channel_ban(fd, chan, banid);
			 }
		    }
	       }
	       break;
	  case 'e':
	       {
		    char *banid;
		    FILE2 *log = irc_fopen("ircd.log", "a+");

		    irc_fprintf(log, "flag = %i. ac = %i. arg = %i. ", flag, ac, arg);
		    if (arg == ac) {
			 banid = NULL;
		    } else {
			 banid = av[arg++];
		    }
		    irc_fprintf(log, "banid = %s.\n", banid);
		    irc_fflush(log);
		    irc_fclose(log);
		    if (banid == NULL) {
			 __not_enough_parameters(fd, "MODE");
		    } else {
			 if (flag) {
			      __add_channel_exempt(fd, chan, banid);
			 } else {
			      __del_channel_exempt(fd, chan, banid);
			 }
		    }
	       }
	       break;
	  case 'I':
	       {
		    char *banid;
		    FILE2 *log = irc_fopen("ircd.log", "a+");

		    irc_fprintf(log, "flag = %i. ac = %i. arg = %i. ", flag, ac, arg);
		    if (arg == ac) {
			 banid = NULL;
		    } else {
			 banid = av[arg++];
		    }
		    irc_fprintf(log, "banid = %s.\n", banid);
		    irc_fflush(log);
		    irc_fclose(log);
		    if (banid == NULL) {
			 __not_enough_parameters(fd, "MODE");
		    } else {
			 if (flag) {
			      __add_channel_perminvite(fd, chan, banid);
			 } else {
			      __del_channel_perminvite(fd, chan, banid);
			 }
		    }
	       }
	       break;
	  case 'i':
	       mode->invite_only.new = flag;
	       break;
	  case 'k':
	       if (mode->key.orig == 1 && flag) {
		    arg++;
		    irc_fprintf(fd->sok, ERR_KEYSET, NAME, fd->nick, chan->name);
		    irc_fflush(fd->sok);
	       } else if (mode->key.orig == 0 && flag) {
		    if (ac == arg) {
			 __not_enough_parameters(fd, "MODE");
		    } else {
			 mode->key.new = 1;
			 mode->key.val = av[arg++];
		    }
	       } else {
		    mode->key.new = 0;
	       }
	       break;
	  case 'l':
	       mode->limit.new = flag;
	       {
		    char *val;
		    int ival;

		    if (mode->limit.new == 1) {
			 if (arg == ac) {
			      val = NULL;
			 } else {
			      val = av[arg++];
			 }

			 if (val == NULL) {
			      mode->limit.new = 0;
			 } else {
			      ival = atoi(val);
			      if (ival == 0) {
				   mode->limit.new = 0;
			      } else {
				   mode->limit.new_val = ival;
			      }
			 }
		    }
	       }
	       break;
	  case 'm':
	       mode->mute.new = flag;
	       break;
	  case 'n':
	       mode->no_outside_msg.new = flag;
	       break;
	  case 'o':
	       {
		    char *onick;

		    if (ac == arg) {
			 log("not enough parameters for mode");
			 __not_enough_parameters(fd, "MODE");
		    } else {
			 onick = av[arg++];
			 log("calling __set_chan_op");
			 __set_chan_op(fd, chan, onick, flag);
		    }
	       }
	       break;
	  case 'p':
	       mode->private.new = flag;
	       break;
	  case 's':
	       mode->secret.new = flag;
	       break;
	  case 't':
	       mode->oper_only_topic.new = flag;
	       break;
	  case 'v':
	       {
		    char *vnick;

		    if (ac == arg) {
			 __not_enough_parameters(fd, "MODE");
		    } else {
			 vnick = av[arg++];
			 __set_chan_voice(fd, chan, vnick, flag);
		    }
	       }
	       break;
	  default:
	       irc_fprintf(fd->sok, ERR_UNKNOWNMODEFLAG, NAME, fd->nick, *mode_str);
	       irc_fflush(fd->sok);
	  };
	  mode_str++;
     }

     return mode;
}

/* void __set_cmode(struct fd_list *fd, struct chan_list *chan, struct chan_mode_change *mode)
   applies a mode change to chan
 */

void __set_cmode(struct fd_list *fd, struct chan_list *chan, struct chan_mode_change *mode)
{
     char add_mode[10], sub_mode[10];
     char *my_hostname = (fd->hide_hostname ? "hidden.hurricane.net" : fd->hostname);
     int num_add = 0, num_sub = 0;

     memset((void *) add_mode, 0, 10);
     memset((void *) sub_mode, 0, 10);

     if (mode->invite_only.new != mode->invite_only.orig) {
	  if (mode->invite_only.new == 1) {
	       add_mode[num_add++] = 'i';
	  } else {
	       sub_mode[num_sub++] = 'i';
	  }
     }
     if (mode->mute.new != mode->mute.orig) {
	  if (mode->mute.new == 1) {
	       add_mode[num_add++] = 'm';
	  } else {
	       sub_mode[num_sub++] = 'm';
	  }
     }
     if (mode->no_outside_msg.new != mode->no_outside_msg.orig) {
	  if (mode->no_outside_msg.new == 1) {
	       add_mode[num_add++] = 'n';
	  } else {
	       sub_mode[num_sub++] = 'n';
	  }
     }
     if (mode->private.new != mode->private.orig) {
	  if (mode->private.new == 1) {
	       add_mode[num_add++] = 'p';
	  } else {
	       sub_mode[num_sub++] = 'p';
	  }
     }
     if (mode->secret.new != mode->secret.orig) {
	  if (mode->secret.new == 1) {
	       add_mode[num_add++] = 's';
	  } else {
	       sub_mode[num_sub++] = 's';
	  }
     }
     if (mode->oper_only_topic.new != mode->oper_only_topic.orig) {
	  if (mode->oper_only_topic.new == 1) {
	       add_mode[num_add++] = 't';
	  } else {
	       sub_mode[num_sub++] = 't';
	  }
     }
     if (num_add > 0 && num_sub > 0) {
	  struct user_list *t;

	  t = chan->users;
	  while (t != NULL) {
	       irc_fprintf(t->user->sok, ":%s!%s@%s MODE %s +%s-%s\r\n", fd->nick, fd->userid, my_hostname, chan->name,
			   add_mode, sub_mode);
	       irc_fflush(t->user->sok);
	       t = t->next;
	  }
     } else if (num_add > 0) {
	  struct user_list *t;

	  t = chan->users;
	  while (t != NULL) {
	       irc_fprintf(t->user->sok, ":%s!%s@%s MODE %s +%s\r\n", fd->nick, fd->userid, my_hostname, chan->name,
			   add_mode);
	       irc_fflush(t->user->sok);
	       t = t->next;
	  }
     } else if (num_sub > 0) {
	  struct user_list *t;

	  t = chan->users;
	  while (t != NULL) {
	       irc_fprintf(t->user->sok, ":%s!%s@%s MODE %s -%s\r\n", fd->nick, fd->userid, my_hostname, chan->name,
			   sub_mode);
	       irc_fflush(t->user->sok);
	       t = t->next;
	  }
     }
     if (mode->key.new == 1) {
	  if (mode->key.orig != 1) {
	       struct user_list *t;
	       char key_val[15];

	       snprintf(key_val, 15, "%s", mode->key.val);

	       t = chan->users;
	       while (t != NULL) {
		    irc_fprintf(t->user->sok, ":%s!%s@%s MODE %s +k %s\r\n", fd->nick, fd->userid, my_hostname, chan->name,
				key_val);
		    irc_fflush(t->user->sok);
		    t = t->next;
	       }

	       chan->key = 1;
	       snprintf(chan->key_val, 15, "%s", mode->key.val);
	  }
     } else {
	  if (mode->key.orig == 1) {
	       struct user_list *t;

	       t = chan->users;
	       while (t != NULL) {
		    irc_fprintf(t->user->sok, ":%s!%s@%s MODE %s -k\r\n", fd->nick, fd->userid, my_hostname, chan->name);
		    irc_fflush(t->user->sok);
		    t = t->next;
	       }
	  }
	  chan->key = 0;
     }

     if (mode->limit.new == 1) {
	  if (mode->limit.new_val != mode->limit.orig_val) {
	       struct user_list *t;

	       t = chan->users;
	       while (t != NULL) {
		    irc_fprintf(t->user->sok, ":%s!%s@%s MODE %s +l %i\r\n", fd->nick, fd->userid, my_hostname, chan->name,
				mode->limit.new_val);
		    irc_fflush(t->user->sok);
		    t = t->next;
	       }

	       chan->limit = 1;
	       chan->limit_val = mode->limit.new_val;
	  }
     } else {
	  if (mode->limit.orig != 0) {
	       struct user_list *t;

	       t = chan->users;
	       while (t != NULL) {
		    irc_fprintf(t->user->sok, ":%s!%s@%s MODE %s -l\r\n", fd->nick, fd->userid, my_hostname,
				chan->name);
		    irc_fflush(t->user->sok);
		    t = t->next;
	       }

	       chan->limit = 0;
	       chan->limit_val = 0;
	  }
     }

     chan->invite_only = mode->invite_only.new;
     chan->mute = mode->mute.new;
     chan->no_outside_msg = mode->no_outside_msg.new;
     chan->private = mode->private.new;
     chan->secret = mode->secret.new;
     chan->oper_only_topic = mode->oper_only_topic.new;
     return;

}

/*
   void __invite_reply(struct fd_list *fd, struct chan_list *chan, char *nick)
   invites nick to chan
 */

void __invite_reply(struct fd_list *fd, struct chan_list *chan, char *nick)
{
     struct fd_list *users = __head;
     struct user_list *cusers = chan->users;
     struct invite_list *invites = chan->invites;
     char *my_hostname = (fd->hide_hostname ? "hidden.hurricane.net" : fd->hostname);

     while (cusers != NULL) {
	  if (!strcasecmp(cusers->user->nick, nick)) {
	       __user_on_channel(fd, nick, chan);
	       return;
	  }
	  cusers = cusers->next;
     }

     while (users != NULL) {
	  if (!strcasecmp(users->nick, nick)) {
	       int ok = 0;

	       if (chan->invite_only) {
		    struct user_list *t;

		    t = chan->users;
		    while (t != NULL) {
			 if (t->user == fd) {
			      if (t->oper)
				   ok = 1;
			      t = NULL;
			 }
			 if (t != NULL)
			      t = t->next;
		    }
	       } else {
		    ok = 1;
	       }

	       if (ok) {
		    if (chan->invite_only) {
			 struct invite_list *new_ent;
			 new_ent = (struct invite_list *) gcAlloc(sizeof(struct invite_list));

			 new_ent->magic = INVITE_LIST_MAGIC;
			 gcIncRefcnt(users);
			 new_ent->user = users;

			 if (invites == NULL) {
			      chan->invites = new_ent;
			 } else {
			      while (invites->next != NULL)
				   invites = invites->next;
			      invites->next = new_ent;
			      new_ent->prev = invites;
			 }
			 {
			      struct user_invite_list *p, *list;
			      p = (struct user_invite_list *) gcAlloc(sizeof(struct user_invite_list));

			      list = (struct user_invite_list *) (users->invite_list);
			      p->magic = USER_INVITE_LIST_MAGIC;
			      gcIncRefcnt(chan);
			      p->chan = chan;
			      if (list != NULL) {
				   while (list->next != NULL)
					list = list->next;
				   list->next = p;
				   p->prev = list;
			      } else {
				   users->invite_list = p;
			      }
			 }
		    }
		    irc_fprintf(fd->sok, RPL_INVITING, NAME, fd->nick, users->nick, chan->name);
		    irc_fprintf(users->sok, ":%s!%s@%s INVITE %s %s\r\n", fd->nick, fd->userid, my_hostname,
				users->nick, chan->name);
		    irc_fflush(fd->sok);
		    irc_fflush(users->sok);
	       } else {
		    irc_fprintf(fd->sok, ERR_CHANOPRIVSNEEDED, NAME, fd->nick, chan->name);
		    irc_fflush(fd->sok);
	       }
	       return;
	  }
	  users = users->next;
     }
     irc_fprintf(fd->sok, ERR_NOSUCHNICK, NAME, fd->nick, nick);
     irc_fflush(fd->sok);
     return;
}

extern int *quit, *die;

/* m_join
   av[0] = "JOIN"
   av[1] = channel(s)
   av[2] = key(s)
   adds a user to a channel
 */

void m_join(struct fd_list *fd, char **av, int ac)
{
     struct chan_list *chan = __clist_head->next;
     char *chan_name;

     msgs.join.use++;
     msgs.join.bytes += strlen(tmp);
     if (ac < 2) {
	  __not_enough_parameters(fd, t);
	  return;
     }
     chan_name = av[1];
     if (!strcasecmp(chan_name, "0")) {		/* leave all channels */
	  char *buf;
	  struct user_chan_list *q = fd->chan_list;
	  int bufptr = 0;

	  buf = (char *) malloc(strlen(q->chan->name) + 6);
	  sprintf(buf, "PART %s", q->chan->name);
	  bufptr = strlen(q->chan->name) + 5;
	  q = q->next;

	  while (q) {
	       buf = (char *) realloc(buf, bufptr + strlen(q->chan->name) + 6);
	       sprintf(&(buf[bufptr]), ",%s", q->chan->name);
	       bufptr += strlen(q->chan->name) + 1;
	       q = q->next;
	  }

	  {
	       int r = __time(NULL);
	       FILE2 *q = irc_fopen("channels.log", "a+");

	       irc_fprintf(q, "[%s] %s\n", ctime((time_t *) & r), buf);
	       irc_fclose(q);
	  }

	  buf[bufptr] = '\r';

	  {
	       int nuh;
	       int *_quit, *_die;

	       _quit = quit;
	       _die = die;
	       parse(buf, fd, &nuh, &nuh);
	       quit = _quit;
	       die = _die;
	  }
	  free(buf);
	  return;
     }
     if (strchr(chan_name, ',')) {
	  char **cv;
	  int cc, i;
	  char *cptr;

	  cv = NULL;
	  cptr = chan_name;
	  while (*chan_name != 0) {
	       if (*chan_name == ',') {
		    *chan_name = 0;
		    cv = (char **) realloc(cv, sizeof(char **) * ++cc);

		    cv[cc - 1] = cptr;
		    chan_name++;
		    cptr = chan_name;
	       } else {
		    chan_name++;
	       }
	  }
	  cv = (char **) realloc(cv, sizeof(char **) * ++cc);

	  cv[cc - 1] = cptr;
	  cv = (char **) realloc(cv, sizeof(char **) * ++cc);

	  cv[cc - 1] = NULL;

	  for (i = 0; i < cc; i++) {
	       int joined = 0;

	       if (!strcasecmp(cv[i], "0")) {
		    char *buf;
		    struct user_chan_list *q = fd->chan_list;
		    int bufptr = 0;

		    buf = (char *) malloc(strlen(q->chan->name) + 6);
		    sprintf(buf, "PART %s", q->chan->name);
		    bufptr = strlen(q->chan->name) + 5;
		    q = q->next;

		    while (q) {
			 buf = (char *) realloc(buf, bufptr + strlen(q->chan->name) + 6);
			 sprintf(&(buf[bufptr]), ",%s", q->chan->name);
			 bufptr += strlen(q->chan->name) + 1;
			 q = q->next;
		    }

		    buf[bufptr] = '\r';

		    {
			 int nuh;
			 int *_quit, *_die;

			 _quit = quit;
			 _die = die;
			 parse(buf, fd, &nuh, &nuh);
			 quit = _quit;
			 die = _die;
		    }
		    free(buf);
		    free(cv);
		    return;
	       }
	       chan = __clist_head->next;
	       while (chan != NULL) {
		    if (!strcasecmp(chan->name, cv[i])) {
			 __join_existing_channel(fd, chan, av, i);
			 joined = 1;
			 chan = NULL;
		    }
		    if (chan != NULL)
			 chan = chan->next;
	       }
	       if (!joined)
		    __join_new_channel(fd, cv[i]);
	  }

	  free(cv);
	  return;
     } else {
	  while (chan != NULL) {
	       if (!strcasecmp(chan->name, chan_name)) {
		    __join_existing_channel(fd, chan, av, 0);
		    return;
	       }
	       chan = chan->next;
	  }
	  __join_new_channel(fd, chan_name);
	  return;
     }
}

/* m_part
   av[0] = "PART"
   av[1] = channel(s)
   removes a user from a channel
 */

void m_part(struct fd_list *fd, char **av, int ac)
{
     char *chan_name;

     msgs.part.use++;
     msgs.part.bytes += strlen(tmp);
     if (ac < 2) {
	  __not_enough_parameters(fd, t);
	  return;
     } else {
	  chan_name = av[1];
	  if (strchr(chan_name, ',')) {
	       char **cv;
	       int cc, i;
	       char *cptr;

	       cv = NULL;
	       cptr = chan_name;
	       while (*chan_name != 0) {
		    if (*chan_name == ',') {
			 *chan_name = 0;
			 cv = (char **) realloc(cv, sizeof(char **) * ++cc);

			 cv[cc - 1] = cptr;
			 chan_name++;
			 cptr = chan_name;
		    } else {
			 chan_name++;
		    }
	       }
	       cv = (char **) realloc(cv, sizeof(char **) * ++cc);

	       cv[cc - 1] = cptr;
	       cv = (char **) realloc(cv, sizeof(char **) * ++cc);

	       cv[cc - 1] = NULL;

	       for (i = 0; i < cc; i++) {
		    __part_channel(fd, cv[i], 0, NULL, 0, NULL);
	       }
	       free(cv);
	       return;
	  } else {
	       __part_channel(fd, chan_name, 0, NULL, 0, NULL);
	       return;
	  }
     }
}

/* m_list
   av[0] = "LIST"
   av[1] = channel
   returns a listing of all channels, or just channel if specified
 */

void m_list(struct fd_list *fd, char **av, int ac)
{
     msgs.list.use++;
     msgs.list.bytes += strlen(tmp);
     if (ac > 1 && av[1][0] != 0) {
	  __list_channel(fd, av[1]);
     } else {
	  __list_all_channels(fd);
     }
     return;
}

/* m_topic
   av[0] = "TOPIC"
   av[1] = channel
   av[2] = new topic
   displays or sets the topic on a channel
 */

void m_topic(struct fd_list *fd, char **av, int ac)
{
     char *my_hostname = (fd->hide_hostname ? "hidden.hurricane.net" : fd->hostname);

     msgs.topic.use++;
     msgs.topic.bytes += strlen(tmp);
     if (ac < 2) {
	  __not_enough_parameters(fd, t);
	  return;
     } else {
	  char *chan_name, *new_topic;
	  struct chan_list *chan = __clist_head->next;

	  chan_name = av[1];
	  if (ac == 2) {
	       new_topic = NULL;
	  } else {
	       new_topic = av[2];
	  }
	  while (chan != NULL) {
	       if (!strcasecmp(chan->name, chan_name)) {
		    if (new_topic == NULL) {
			 if (chan->topic[0] == 0) {
			      irc_fprintf(fd->sok, RPL_NOTOPIC, NAME, fd->nick, chan->name);
			 } else {
			      irc_fprintf(fd->sok, RPL_TOPIC, NAME, fd->nick, chan->name,
					  chan->topic);
			      irc_fprintf(fd->sok, RPL_TOPICSET, NAME, fd->nick, chan->name,
					  chan->topic_set_by, chan->topic_set_at);
			 }
			 return;
		    } else {
			 int ok = 0;

			 if (new_topic[strlen(new_topic) - 1] == '\r' ||
			     new_topic[strlen(new_topic) - 1] == '\n')
			      new_topic[strlen(new_topic) - 1] = 0;
			 if (new_topic[strlen(new_topic) - 1] == '\r' ||
			     new_topic[strlen(new_topic) - 1] == '\n')
			      new_topic[strlen(new_topic) - 1] = 0;

			 if (chan->oper_only_topic) {
			      struct user_list *u;

			      log("mode +t set, checking to see if user is chanop");
			      u = chan->users;
			      while (u != NULL) {
				   if (u->user == fd) {
					if (u->oper) {
					     ok = 1;
					     u = NULL;
					}
				   }
				   if (u != NULL)
					u = u->next;
			      }
			 } else {
			      ok = 1;
			 }
			 if (ok) {
			      if (new_topic[0] == 0) {
				   memset((void *) (chan->topic), 0, 1024);
			      } else {
				   strncpy(chan->topic, new_topic, 1023);
			      }
			      snprintf(chan->topic_set_by, 15, "%s", fd->nick);
			      chan->topic_set_at = __time(NULL);

			      {
				   struct user_list *u;

				   u = chan->users;
				   while (u != NULL) {
					irc_fprintf(u->user->sok, ":%s!%s@%s TOPIC %s :%s\r\n",
						    fd->nick, fd->userid, my_hostname,
						    chan->name, chan->topic);
					irc_fflush(u->user->sok);
					u = u->next;
				   }
			      }
			 } else {
			      irc_fprintf(fd->sok, ERR_CHANOPRIVSNEEDED, NAME, fd->nick,
					  chan->name);
			      irc_fflush(fd->sok);
			 }
			 return;
		    }
	       }
	       chan = chan->next;
	  }
	  __no_such_channel(fd, chan_name);
	  return;
     }
}

/* m_names
   av[0] = "NAMES"
   av[1] = channel
   displays a listing of the users on channel
 */

void m_names(struct fd_list *fd, char **av, int ac)
{
     char *chan_name;

     msgs.names.use++;
     msgs.names.bytes += strlen(tmp);
     if (ac < 2) {
	  __not_enough_parameters(fd, t);
	  return;
     } else {
	  struct chan_list *list;

	  chan_name = av[1];
	  list = __clist_head->next;
	  while (list != NULL) {
	       if (!strcasecmp(list->name, chan_name)) {
		    __names_reply(fd, list);
		    return;
	       }
	       list = list->next;
	  }
	  __no_such_channel(fd, chan_name);
	  return;
     }
}

/* m_invite
   av[0] = "INVITE"
   av[1] = nickname
   av[2] = channel
   invites a user to a channel
 */

void m_invite(struct fd_list *fd, char **av, int ac)
{
     msgs.invite.use++;
     msgs.invite.bytes += strlen(tmp);
     if (ac < 3) {
	  __not_enough_parameters(fd, t);
	  return;
     } else {
	  char *nick = av[1];
	  char *chan_name = av[2];
	  struct chan_list *chan = __clist_head->next;

	  while (chan != NULL) {
	       if (!strcasecmp(chan->name, chan_name)) {
		    __invite_reply(fd, chan, nick);
		    return;
	       }
	       chan = chan->next;
	  }
	  __no_such_channel(fd, chan_name);
	  return;
     }
}

/* m_kick
   av[0] = "KICK"
   av[1] = channel
   av[2] = user
   av[3] = reason
   kicks a user from a channel
 */

void m_kick(struct fd_list *fd, char **av, int ac)
{
     char *chan, *nick;

     msgs.kick.use++;
     msgs.kick.bytes += strlen(tmp);
     if (ac < 3) {
	  __not_enough_parameters(fd, t);
     } else {
	  char *msg;
	  struct fd_list *t;

	  chan = av[1];
	  nick = av[2];
	  if (ac < 4) {
	       msg = fd->nick;
	  } else {
	       msg = av[3];
	  }
	  if (msg[strlen(msg) - 1] == '\r' || msg[strlen(msg) - 1] == '\n') {
	       msg[strlen(msg) - 1] = 0;
	  }
	  if (msg[strlen(msg) - 1] == '\r' || msg[strlen(msg) - 1] == '\n') {
	       msg[strlen(msg) - 1] = 0;
	  }
	  if (msg[0] == '\0')
	       msg = fd->nick;
	  t = __head;
	  while (t != NULL) {
	       if (!strcasecmp(t->nick, nick)) {
			if (t->service) {
				irc_fprintf(fd->sok, ":%s 484 %s %s %s :Cannot kill, kick, or deop channel service\r\n",
					server_name, fd->nick, t->nick, chan);
				return;
			} else {
			    __part_channel(t, chan, 0, msg, 1, fd);
			    return;
			}
	       }
	       if (t != NULL)
		    t = t->next;
	  }
	  irc_fprintf(fd->sok, ERR_NOSUCHNICK, NAME, fd->nick, nick);
     }
     return;
}

/* m_invites
   av[0] = "INVITES"
   av[1] = channel name

   returns a list of invited people
 */

void m_invites(struct fd_list *fd, char **av, int ac)
{
     if (ac < 2 || ac >= 2 && av[1][0] == 0) {
	  __not_enough_parameters(fd, "INVITES");
	  return;
     } {
	  struct chan_list *clist = __clist_head->next;

	  while (clist) {
	       if (!strcasecmp(av[1], clist->name)) {
		    struct invite_list *i = clist->invites;

		    irc_fprintf(fd->sok, ":%s 619 %s %s :Nickname\r\n", server_name,
				fd->nick, clist->name);
		    while (i) {
			 irc_fprintf(fd->sok, ":%s 620 %s %s %s\r\n", server_name,
				     fd->nick, clist->name, i->user->nick);
			 i = i->next;
		    }
		    irc_fprintf(fd->sok, ":%s 621 %s %s :End of /INVITES.\r\n",
				server_name, fd->nick, clist->name);
		    return;
	       }
	       clist = clist->next;
	  }
	  __no_such_channel(fd, av[1]);
	  return;
     }
}
