#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <regex.h>
#include <signal.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/time.h>
#include "struct.h"
#include "dl.h"
#include "msg.h"
#include "user.h"
#include "channels.h"
#include "errors.h"
#include "defines.h"
#include "irc_printf.h"
#include <pthread.h>

extern time_t __time(time_t *);

const char *months[] =
{
     "January",
     "February",
     "March",
     "April",
     "May",
     "June",
     "July",
     "August",
     "September",
     "October",
     "November",
     "December"
};

const char *days[] =
{
     "Sunday",
     "Monday",
     "Tuesday",
     "Wednesday",
     "Thursday",
     "Friday",
     "Saturday"
};

time_t started;

extern char *tzname[2];
extern long int timezone;
extern int daylight;

static char *
  my_ctime(time_t q)
{
     struct tm *t = localtime(&q);
     char is_past_gmt;
     long int mytz;
     int tzhr;
     int tzmn;
     static char buf[255];

     bzero(buf, 255);
     is_past_gmt = (timezone >= 0 ? 1 : 0);
     mytz = (timezone >= 0 ? timezone : -timezone);
     tzhr = (mytz / 60) / 60;
     tzmn = (mytz - (tzhr * 60 * 60)) / 60;

     snprintf(buf, 255, "%s %s %i %i -- %02i:%02i %c%02i:%02i", days[t->tm_wday], months[t->tm_mon], t->tm_mday,
	      t->tm_year + 1900, t->tm_hour, t->tm_min, (is_past_gmt ? '-' : '+'), tzhr, tzmn);
     return buf;
}

Module_Info services_module_info[] =
{
     {"services",
      "service module for ircd 0.3.6",
      "0.4"
     }
};

int privmsg(struct fd_list *, char **, int);
int nick(struct fd_list *, char **, int);
int nickserv(struct fd_list *, char **, int);
int whois(struct fd_list *, char **, int);
int links(struct fd_list *, char **, int);
int who(struct fd_list *, char **, int);
int version(struct fd_list *, char **, int);

const static char *s_NickServ = "NickServ";
const static char *ns_userid = "nickserv";
const static char *ns_hostname = "services.net";
const static char *ns_realname = "Nickname Service";
const static char *ns_server = "services.net";
const static char *ns_server_desc = "IRC Services";

const static char *s_OperServ = "OperServ";
const static char *os_userid = "operserv";
const static char *os_hostname = "services.net";
const static char *os_realname = "Operator Service";
const static char *os_server = "services.net";
const static char *os_server_desc = "IRC Services";

const static char *s_MemoServ = "MemoServ";
const static char *ms_userid = "memoserv";
const static char *ms_hostname = "services.net";
const static char *ms_realname = "Memo Service";
const static char *ms_server = "services.net";
const static char *ms_server_desc = "IRC Services";

const static char *s_ChanServ = "ChanServ";
const static char *cs_userid = "chanserv";
const static char *cs_hostname = "services.net";
const static char *cs_realname = "Channel Service";
const static char *cs_server = "services.net";
const static char *cs_server_desc = "IRC Services";

Functions nickserv_function_list[] =
{
     {"PRIVMSG", privmsg, 1, 0, 0},
     {"NICK", nick, 1, 1, 0},
     {"NICKSERV", nickserv, 1, 0, 0},
     {"WHOIS", whois, 1, 0, 0},
     {"WHO", who, 1, 0, 0},
     {"LINKS", links, 1, 0, 0},
     {"NOTICE", privmsg, 1, 0, 0},
     {"VERSION", version, 1, 0, 0},
     {NULL, NULL, 0, 0, 0}
};

Module_Info *
  __module_get_info()
{
     return services_module_info;
}

Functions *
  __module_get_functions()
{
     return nickserv_function_list;
}

struct ns_func {
     char *cmd;
     void (*function) (struct fd_list *, char **, int);
     int need_identify;
};

typedef struct ns_func NS_Functions;

int id = 0;

struct ns_user {
     struct ns_user *prev;
     struct ns_user *next;
     char nick[15];
     char pass[15];
     int language;
     char url[100];
     char email[100];
     int kill;
     int secure;
     int private;
     struct {
	  char usermask:1;
	  char email:1;
	  char quit:1;
	  char reserved:5;
     } hide;
     int registered_at;
     int last_seen_at;
     char last_mask[100];
     char last_quit[200];
     char last_realname[100];
     int user_num;
};

typedef struct ns_user NS_User;

struct ns_access {
     struct ns_access *prev;
     struct ns_access *next;
     NS_User *user;
     int user_num;
     char hostmask[100];
};

typedef struct ns_access NS_AccessList;

struct ns_links {
     struct ns_links *prev;
     struct ns_links *next;
     NS_User *user;
     int user_num;
     char nick[15];
};

typedef struct ns_links NS_LinkList;

struct ms_msg {
     NS_User *to;
     int to_user_num;
     NS_User *from;
     int from_user_num;
     char memo[255];
     time_t sent_at;
     int memo_id;
     struct ms_msg *prev, *next;
};

int last_memo_id = 0;

typedef struct ms_msg MS_Memo;

MS_Memo *memo_list = NULL;

struct cs_accesslist {
     struct cs_accesslist *prev, *next;
     NS_User *user;
     int user_num;
     unsigned char level:4;
     unsigned char reserved:4;
};

typedef struct cs_accesslist CS_AccessList;

struct cs_chan {
     struct cs_chan *prev;
     struct cs_chan *next;
     NS_User *founder;
     int founder_user_num;
     NS_User *successor;
     int successor_user_num;
     char name[20];
     char password[15];
     char description[100];
     char url[100];
     char email[100];
     char entrymsg[100];
     char topic[100];
     struct {
	  char invite_only:1;
	  char key:1;
	  char limit:1;
	  char mute:1;
	  char no_outside_msg:1;
	  char private:1;
	  char secret:1;
	  char oper_only_topic:1;
	  char keyval[15];
	  short int limit_val;
     } modes;
     short int keep_topic:1;
     short int topic_lock:1;
     short int mode_lock:1;
     short int private:1;
     short int restricted:1;
     short int secure:1;
     short int secureops:1;
     short int leaveops:1;
     short int opnotice:1;
     short int reserved:7;
     CS_AccessList *acl;
};

typedef struct cs_chan CS_Channel;

CS_Channel *registered_channels = NULL;

NS_User *registered_nicks = NULL;
NS_AccessList *access_list = NULL;
NS_LinkList *link_list = NULL;

int valid_pass(char *pass, char *nick, char *userid)
{
     if (!strcasecmp(pass, nick) || !strcasecmp(pass, userid)) {
	  return -1;
     }
     if (strlen(pass) < 5) {
	  return -1;
     } {
	  int i, j, k;

	  i = 0;
	  j = 0;
	  k = 0;
	  j = pass[0];
	  for (i = 1; i < strlen(pass); i++) {
	       if (j == pass[i] || j == pass[i] + 1 || j == pass[i] - 1) {
		    k++;
	       }
	  }

	  if (k > strlen(pass) / 2) {
	       return -1;
	  }
     }

     {
	  int num_caps, num_lwr, num_symbol, num_numeric, i, j;
	  char *caps = "ABCDEFGHIJKLMNOPQRSTUVWXYZ", *lwr = "abcdefghijklmnopqrstuvwxyz", *sym = "`~!@#$%^&*()-_=+[{]}|';:/?.>,<",
	   *num = "1234567890";

	  num_caps = 0;
	  num_lwr = 0;
	  num_symbol = 0;
	  num_numeric = 0;
	  i = 0;
	  j = 0;
	  for (i = 0; i < strlen(pass); i++) {
	       if (strchr(caps, pass[i])) {
		    num_caps++;
	       } else if (strchr(lwr, pass[i])) {
		    num_lwr++;
	       } else if (strchr(sym, pass[i])) {
		    num_symbol++;
	       } else if (strchr(num, pass[i])) {
		    num_numeric++;
	       }
	  }

	  if ((num_caps + num_symbol) < 1 || num_numeric < 2) {
	       return -1;
	  }
     }

     return 0;
}

void substr(char *dest, char *str, int begin, int len)
{
     int i;

     for (i = begin; i < (begin + len); i++) {
	  dest[i - begin] = str[i];
     }
}

NS_User *
  find_user(char *nick)
{
     NS_User *list;
     NS_LinkList *list2;

     list = registered_nicks;
     list2 = link_list;

     while (list2 != NULL) {
	  if (!strcasecmp(list2->nick, nick)) {
	       return list2->user;
	  }
	  list2 = list2->next;
     }

     while (list != NULL) {
	  if (!strcasecmp(list->nick, nick)) {
	       return list;
	  }
	  list = list->next;
     }

     return NULL;
}

void ns_dummy(struct fd_list *fd, char **args, int nargs)
{
     irc_fprintf(fd->sok, ":%s!%s@%s NOTICE %s :%s: %s\r\n", s_NickServ, ns_userid, ns_hostname, fd->nick, args[1],
		 "I don't know how to do that yet.");
     return;
}

void ns_register(struct fd_list *fd, char **args, int nargs)
{
     if (nargs >= 2) {
	  {
	       NS_User *list;

	       list = registered_nicks;
	       while (list != NULL) {
		    if (!strcasecmp(list->nick, fd->nick)) {
			 char tmp[255];

			 snprintf(tmp, 255, "Nickname \002%s\002 is already registered!", fd->nick);
			 irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick,
				     tmp);
			 return;
		    }
		    list = list->next;
	       }
	  }
	  {
	       NS_User *new_user = (NS_User *) malloc(sizeof(NS_User));
	       NS_AccessList *new_acl = (NS_AccessList *) malloc(sizeof(NS_AccessList));
	       char userid[15], hostmask[100];
	       int i, j = 0;

	       bzero(new_user, sizeof(NS_User));
	       bzero(new_acl, sizeof(NS_AccessList));
	       bzero(hostmask, 100);
	       strncpy(new_user->nick, fd->nick, 15);
	       strncpy(new_user->pass, args[1], 15);
	       new_user->registered_at = time(NULL);
	       strncpy(new_user->last_realname, fd->realname, 100);

	       strncpy(userid, fd->userid, 15);

	       for (i = 0; i < strlen(fd->hostname); i++) {
		    if (!j && fd->hostname[i] == '.') {
			 j = i + 1;
		    } else if (j) {
			 hostmask[i - j + 1] = fd->hostname[i];
		    }
	       }
	       hostmask[0] = '*';

	       snprintf(new_acl->hostmask, 100, "%s@%s", userid, hostmask);
	       new_acl->user = new_user;

	       new_user->user_num = ++id;
	       new_acl->user_num = new_user->user_num;

	       if (registered_nicks == NULL) {
		    fprintf(stderr, "registered_nicks = NULL\n");
		    registered_nicks = new_user;
	       } else {
		    NS_User *list;

		    fprintf(stderr, "registered_nicks = %p\n", registered_nicks);
		    list = registered_nicks;
		    while (list->next != NULL) {
			 fprintf(stderr, "list = %p\n", list);
			 list = list->next;
		    }
		    list->next = new_user;
		    new_user->prev = list;
	       }

	       if (access_list == NULL) {
		    access_list = new_acl;
	       } else {
		    NS_AccessList *list;

		    list = access_list;
		    while (list->next != NULL)
			 list = list->next;
		    list->next = new_acl;
		    new_acl->prev = list;
	       }

	       {
		    char tmp[255];

		    snprintf(tmp, 255, "Nickname \002%s\002 registered under your account: %s", fd->nick, new_acl->hostmask);
		    irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, tmp);
		    snprintf(tmp, 255, "Your password is \002%s\002 - remember this for later use.", new_user->pass);
		    irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, tmp);
	       }
	  }
     } else {
	  irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick,
		      "Syntax: \002REGISTER\002 password");
	  irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick,
		      "\002/msg NickServ HELP REGISTER\002 for more information.");
     }
}

#define NS_IS_IDENTIFIED 0x10000000

void ns_identify(struct fd_list *fd, char **args, int nargs)
{
     if (nargs > 1) {
	  NS_User *user = find_user(fd->nick);

	  if (user == NULL) {
	       irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, "Your nick isn't registered.");
	       return;
	  } else {
	       if (!strcmp(args[1], user->pass)) {
		    fd->registered |= NS_IS_IDENTIFIED;
		    irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick,
				"Password accepted - you are now recognized.");
		    snprintf(user->last_realname, 100, "%s", fd->realname);
		    snprintf(user->last_mask, 200, "%s@%s", fd->userid, fd->hostname);
		    user->last_seen_at = time(NULL);
		    {
			 MS_Memo *ml = memo_list;
			 int n_memos = 0;

			 while (ml) {
			      if (ml->to == user) {
				   n_memos++;
			      }
			      ml = ml->next;
			 }
			 if (n_memos) {
			      irc_fprintf(fd->sok, ":%s!%s@%s NOTICE %s :You have %i memos waiting.\r\n",
					  s_MemoServ, ms_userid, ms_hostname, fd->nick, n_memos);
			      irc_fprintf(fd->sok,
					  ":%s!%s@%s NOTICE %s :Type \x02/msg MemoServ LIST\x02 for a list.\r\n",
					  s_MemoServ, ms_userid, ms_hostname, fd->nick);
			 }
		    }
	       } else {
		    irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick,
				"Password incorrect.");
	       }
	  }
     } else {
	  irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick,
		      "Syntax: \002IDENTIFY\002 password");
	  irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick,
		      "\002/msg NickServ HELP IDENTIFY\002 for more information.");
     }
}

void ns_access(struct fd_list *fd, char **args, int nargs)
{
     if (nargs == 1) {
	  irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick,
		      "Syntax: \002ACCESS { ADD | DEL | LIST } [\002mask\002]\002");
	  irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick,
		      "\002/msg NickServ HELP ACCESS\002 for more information.");
     } else if (nargs == 2) {
	  if (fd->registered & NS_IS_IDENTIFIED) {
	       if (!strcasecmp(args[1], "LIST")) {
		    NS_AccessList *list;
		    NS_User *user = find_user(fd->nick);

		    list = access_list;
		    irc_fprintf(fd->sok, ":%s!%s@%s NOTICE %s :Access list:\r\n", s_NickServ, ns_userid, ns_hostname,
				fd->nick);
		    while (list != NULL) {
			 if (list->user == user) {
			      irc_fprintf(fd->sok, ":%s!%s@%s NOTICE %s :    %s\r\n", s_NickServ, ns_userid,
					  ns_hostname, fd->nick, list->hostmask);
			 }
			 list = list->next;
		    }
	       } else {
		    irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick,
				"Syntax: \002ACCESS { ADD | DEL | LIST } [\002mask\002]\002");
		    irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick,
				"\002/msg NickServ HELP ACCESS\002 for more information.");
	       }
	  } else {
	       if (!strcasecmp(args[1], "LIST")) {
		    if (find_user(fd->nick)) {
			 irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick,
				     "Password authentication required for that command.");
			 irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick,
				   "Retry after typing \002/msg NickServ IDENTIFY\002 password.");
		    } else {
			 irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick,
				     "Your nick isn't registered.");
		    }
	       } else {
		    irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick,
				"Syntax: \002ACCESS { ADD | DEL | LIST } [\002mask\002]\002");
		    irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick,
				"\002/msg NickServ HELP ACCESS\002 for more information.");
	       }
	  }
     } else if (nargs >= 3) {
	  if (!strcasecmp(args[1], "ADD")) {
	       if (fd->registered & NS_IS_IDENTIFIED) {
		    NS_AccessList *new_acl;
		    NS_User *user;

		    user = find_user(fd->nick);
		    if (user == NULL) {
			 fd->registered |= ~NS_IS_IDENTIFIED;
			 irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick,
				     "Your nick isn't registered.");
		    } else {
			 NS_AccessList *list = access_list;

			 new_acl = (NS_AccessList *) malloc(sizeof(NS_AccessList));
			 bzero(new_acl, sizeof(NS_AccessList));
			 strncpy(new_acl->hostmask, args[2], 100);
			 new_acl->user = user;
			 new_acl->user_num = user->user_num;
			 while (list->next != NULL)
			      list = list->next;
			 list->next = new_acl;
			 new_acl->prev = list;
			 {
			      char tmp[255];

			      snprintf(tmp, 255, "\002%s\002 added to your access list.", new_acl->hostmask);
			      irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, tmp);
			 }
		    }
	       } else {
		    irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick,
				"Password authentication required for that command.");
		    irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick,
				"Retry after typing \002/msg NickServ IDENTIFY\002 password.");
	       }
	  } else if (!strcasecmp(args[1], "DEL")) {
	       if (fd->registered & NS_IS_IDENTIFIED) {
		    NS_AccessList *acl;
		    NS_User *user;

		    user = find_user(fd->nick);
		    if (user == NULL) {
			 fd->registered |= ~NS_IS_IDENTIFIED;
			 irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick,
				     "Your nick isn't registered.");
		    } else {
			 int removed = 0;

			 acl = access_list;
			 while (acl != NULL) {
			      if (acl->user == user) {
				   if (!strcasecmp(acl->hostmask, args[2])) {
					if (acl->prev == NULL && acl->next == NULL) {
					     access_list = NULL;
					     free(acl);
					     acl = NULL;
					} else if (acl->prev == NULL && acl->next != NULL) {
					     access_list = acl->next;
					     (acl->next)->prev = NULL;
					     free(acl);
					     acl = NULL;
					} else if (acl->prev != NULL && acl->next == NULL) {
					     (acl->prev)->next = NULL;
					     free(acl);
					     acl = NULL;
					} else {
					     (acl->prev)->next = acl->next;
					     (acl->next)->prev = acl->prev;
					     free(acl);
					     acl = NULL;
					}
					{
					     char tmp[255];

					     snprintf(tmp, 255,
						      "\002%s\002 deleted from your access list.",
						      args[2]);
					     irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid,
							 ns_hostname, fd->nick, tmp);
					}
					removed = 1;
				   }
			      }
			      if (acl)
				   acl = acl->next;
			 }

			 if (!removed) {
			      char tmp[255];

			      snprintf(tmp, 255, "\002%s\002 not found on your access list.", args[2]);
			      irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick,
					  tmp);
			 }
		    }
	       } else {
		    irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick,
				"Password authentication required for that command.");
		    irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick,
				"Retry after typing \002/msg NickServ IDENTIFY\002 password.");
	       }
	  } else if (!strcasecmp(args[0], "LIST")) {
	       NS_AccessList *acl;
	       NS_User *user;

	       if (fd->registered & NS_IS_IDENTIFIED) {
		    user = find_user(fd->nick);
		    if (user == NULL) {
			 fd->registered |= ~NS_IS_IDENTIFIED;
			 irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick,
				     "Your nick isn't registered.");
		    } else {
			 acl = access_list;
			 irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick,
				     "Access list:");
			 while (acl != NULL) {
			      if (acl->user == user) {
				   irc_fprintf(fd->sok, ":%s!%s@%s NOTICE %s :    %s\r\n",
					       s_NickServ, ns_userid, ns_hostname, fd->nick,
					       acl->hostmask);
			      }
			      acl = acl->next;
			 }
		    }
	       } else {
		    irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick,
				"Password authentication required for that command.");
		    irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick,
				"Retry after typing \002/msg NickServ IDENTIFY\002 password.");
	       }
	  } else {
	       irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick,
			   "Syntax: \002ACCESS { ADD | DEL | LIST } [\002mask\002]\002");
	       irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick,
			   "\002/msg NickServ HELP ACCESS\002 for more information.");
	  }
     }
}

void ns_link(struct fd_list *fd, char **args, int nargs)
{
     if (nargs < 3) {
	  irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick,
		      "Syntax: \002LINK\002 nick password");
	  irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick,
		      "\002/msg NickServ HELP LINK\002 for more information.");
     } else {
	  NS_User *user = NULL, *me, *list;
	  NS_LinkList *new_link;

	  me = find_user(fd->nick);
	  if (me == NULL) {
	       irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick,
			   "Your nick isn't registered.");
	       return;
	  }
	  list = registered_nicks;
	  while (list != NULL) {
	       if (!strcasecmp(list->nick, args[1])) {
		    user = list;
		    list = NULL;
	       }
	       if (list)
		    list = list->next;
	  }

	  if (user == NULL) {
	       char tmp[255];

	       snprintf(tmp, 255, "Nick \002%s\002 isn't registered.", args[1]);
	       irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, tmp);
	       return;
	  }
	  if (!strcasecmp(user->nick, me->nick)) {
	       irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick,
			   "You can't link to your current nick!");
	       return;
	  }
	  if (!strcasecmp(user->pass, args[2])) {
	       NS_LinkList *list = link_list;

	       new_link = (NS_LinkList *) malloc(sizeof(NS_LinkList));
	       new_link->user = user;
	       new_link->user_num = user->user_num;
	       strncpy(new_link->nick, fd->nick, 15);
	       if (list == NULL) {
		    link_list = new_link;
	       } else {
		    while (list->next != NULL)
			 list = list->next;
		    new_link->prev = list;
		    list->next = new_link;
	       }
	       {
		    char tmp[255];

		    snprintf(tmp, 255, "Your nick has been linked to \002%s\002.", user->nick);
		    irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, tmp);
	       }
	       return;
	  } else {
	       irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick,
			   "Password incorrect.");
	       return;
	  }
     }
}

void ns_set_password(struct fd_list *fd, char **args, int nargs)
{
     NS_User *user;

     user = find_user(fd->nick);

     if (user == NULL) {
	  irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick,
		      "Your nick isn't registered.");
	  return;
     } else {
	  if (fd->registered & NS_IS_IDENTIFIED) {
	       if (strlen(args[2]) < 5) {
		    irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick,
		    "Please try again with a more obscure password.  Passwords should be at least "
			"five characters long, should not be something easily guessed (e.g. your "
		      "real name or your nick), and cannot contain the space or tab characters.");
		    return;
	       } else if (!strcasecmp(args[2], fd->nick)) {
		    irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick,
		    "Please try again with a more obscure password.  Passwords should be at least "
			"five characters long, should not be something easily guessed (e.g. your "
		      "real name or your nick), and cannot contain the space or tab characters.");
		    return;
	       } else if (!valid_pass(args[2], fd->nick, fd->userid)) {
		    char tmp[255];

		    snprintf(user->pass, 15, "%s", args[2]);
		    snprintf(tmp, 255, "Password changed to \002%s\002.", args[2]);
		    irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, tmp);
		    return;
	       } else {
		    irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick,
		    "Please try again with a more obscure password.  Passwords must be at least 5 "
			     "characters long, should not be something easily guessed (e.g. your "
				"real name or your nick), cannot consist of a string consisting of more than half "
				"repeating or consecutive characters (i.e. no passwords like 'abcdef' or 'abc000'), "
				"must contain at least either 1 uppercase or symbol character, or 2 numeric characters, "
		    "i.e. 'blah12' or 'abQhty', and cannot contain the space or tab characters.");
		    return;
	       }
	  } else {
	       irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick,
			   "Password authentication required for that command.");
	       irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick,
			   "\002/msg NickServ HELP SET\002 for more information.");
	       return;
	  }
     }
}

struct ns_lang {
     int num;
     char *name;
};

typedef struct ns_lang NS_Languages;

NS_Languages language[] =
{
     {1, "English"},
     {0, NULL}
};

int max_lang = 1;

void ns_set_language(struct fd_list *fd, char **args, int nargs)
{
     NS_User *user = find_user(fd->nick);

     if (user == NULL) {
	  irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick,
		      "Your nick isn't registered.");
	  return;
     } else {
	  int n;

	  n = atoi(args[2]);
	  if (fd->registered & NS_IS_IDENTIFIED) {
	       if (n < 1 || n > max_lang) {
		    char tmp[255];

		    snprintf(tmp, 255, "Unknown language number \002%i\002. "
		      "Type \002/msg NickServ HELP SET LANGUAGE\002 for a list of languages.", n);
		    irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, tmp);
		    return;
	       } else {
		    NS_Languages *lang = language;

		    while (lang->name != NULL) {
			 if (lang->num == n) {
			      char tmp[255];

			      snprintf(tmp, 255, "Language changed to \002%s\002.", lang->name);
			      irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, tmp);
			      user->language = n;
			      return;
			 }
			 lang++;
		    }

		    {
			 char tmp[255];

			 snprintf(tmp, 255, "Unknown language number \002%i\002. "
				  "Type \002/msg NickServ HELP SET LANGUAGE\002 for a list of languages.", n);
			 irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, tmp);
		    }
		    return;
	       }
	  } else {
	       irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick,
			   "Password authentication required for that command.");
	       irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick,
			   "\002/msg NickServ HELP SET\002 for more information.");
	       return;
	  }
     }
}

void ns_set_url(struct fd_list *fd, char **args, int nargs)
{
     NS_User *user = find_user(fd->nick);

     if (user == NULL) {
	  irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick,
		      "Your nick isn't registered.");
	  return;
     } else {
	  if (fd->registered & NS_IS_IDENTIFIED) {
	       if (nargs == 2) {
		    bzero(user->url, 100);
		    irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick,
				"URL unset.");
	       } else {
		    bzero(user->url, 100);
		    strncpy(user->url, args[2], 100);
		    {
			 char tmp[255];

			 snprintf(tmp, 255, "URL changed to \002%s\002.", user->url);
			 irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, tmp);
		    }
	       }
	       return;
	  } else {
	       irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick,
			   "Password authentication required for that command.");
	       irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick,
			   "\002/msg NickServ HELP SET\002 for more information.");
	       return;
	  }
     }
}

void ns_set_email(struct fd_list *fd, char **args, int nargs)
{
     NS_User *user = find_user(fd->nick);

     if (user == NULL) {
	  irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick,
		      "Your nick isn't registered.");
	  return;
     } else {
	  if (fd->registered & NS_IS_IDENTIFIED) {
	       if (nargs == 2) {
		    bzero(user->email, 100);
		    irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick,
				"E-mail unset.");
	       } else {
		    bzero(user->email, 100);
		    strncpy(user->email, args[2], 100);
		    {
			 char tmp[255];

			 snprintf(tmp, 255, "E-mail changed to \002%s\002.", user->email);
			 irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, tmp);
		    }
	       }
	       return;
	  } else {
	       irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick,
			   "Password authentication required for that command.");
	       irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick,
			   "\002/msg NickServ HELP SET\002 for more information.");
	       return;
	  }
     }
}

void ns_set_kill(struct fd_list *fd, char **args, int nargs)
{
     NS_User *user = find_user(fd->nick);

     if (user == NULL) {
	  irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick,
		      "Your nick isn't registered.");
	  return;
     } else {
	  if (fd->registered & NS_IS_IDENTIFIED) {
	       if (!strcasecmp(args[2], "ON")) {
		    user->kill = 1;
		    irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick,
				"Kill protection is now \002ON\002.");
		    return;
	       } else if (!strcasecmp(args[2], "QUICK")) {
		    user->kill = 2;
		    irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick,
				"Kill protection is now \002ON\002, with a reduced kill delay.");
		    return;
	       } else if (!strcasecmp(args[2], "IMMED")) {
		    irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick,
				"The \002IMMED\002 option is not availiable on this network.");
		    return;
	       } else if (!strcasecmp(args[2], "OFF")) {
		    user->kill = 0;
		    irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick,
				"Kill protection is now \002OFF\002.");
		    return;
	       } else {
		    irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick,
				"Syntax: \002SET KILL {ON | QUICK | OFF}\002");
		    irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick,
				"\002/msg NickServ HELP SET KILL\002 for more information.");
		    return;
	       }
	  } else {
	       irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick,
			   "Password authentication required for that command.");
	       irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick,
			   "\002/msg NickServ HELP SET\002 for more information.");
	       return;
	  }
     }
}

void ns_set_secure(struct fd_list *fd, char **args, int nargs)
{
     NS_User *user = find_user(fd->nick);

     if (user == NULL) {
	  irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick,
		      "Your nick isn't registered.");
	  return;
     } else {
	  if (fd->registered & NS_IS_IDENTIFIED) {
	       if (!strcasecmp(args[2], "ON")) {
		    user->secure = 1;
		    irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick,
				"Secure option is now \002ON\002.");
		    return;
	       } else if (!strcasecmp(args[2], "OFF")) {
		    user->secure = 0;
		    irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick,
				"Secure option is now \002OFF\002.");
		    return;
	       } else {
		    irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick,
				"Syntax: \002SET SECURE {ON | OFF}\002");
		    irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick,
				"\002/msg NickServ HELP SET SECURE\002 for more information.");
		    return;
	       }
	  } else {
	       irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick,
			   "Password authentication required for that command.");
	       irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick,
			   "\002/msg NickServ HELP SET\002 for more information.");
	       return;
	  }
     }
}

void ns_set_private(struct fd_list *fd, char **args, int nargs)
{
     NS_User *user = find_user(fd->nick);

     if (user == NULL) {
	  irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick,
		      "Your nick isn't registered.");
	  return;
     } else {
	  if (fd->registered & NS_IS_IDENTIFIED) {
	       if (!strcasecmp(args[2], "ON")) {
		    user->private = 1;
		    irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick,
				"Private option is now \002ON\002.");
		    return;
	       } else if (!strcasecmp(args[2], "OFF")) {
		    user->private = 0;
		    irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick,
				"Private option is now \002OFF\002.");
		    return;
	       } else {
		    irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick,
				"Syntax: \002SET PRIVATE {ON | OFF}\002");
		    irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick,
				"\002/msg NickServ HELP SET PRIVATE\002 for more information.");
		    return;
	       }
	  } else {
	       irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick,
			   "Password authentication required for that command.");
	       irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick,
			   "\002/msg NickServ HELP SET\002 for more information.");
	       return;
	  }
     }
}

void ns_set_hide(struct fd_list *fd, char **args, int nargs)
{
     NS_User *user = find_user(fd->nick);

     if (user == NULL) {
	  irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick,
		      "Your nick isn't registered.");
	  return;
     } else {
	  if (fd->registered & NS_IS_IDENTIFIED) {
	       if (!strcasecmp(args[3], "ON")) {
		    if (!strcasecmp(args[2], "EMAIL")) {
			 user->hide.email = 1;
			 irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick,
			   "Your E-mail address will now be hidden from NickServ INFO displays.");
			 return;
		    } else if (!strcasecmp(args[2], "USERMASK")) {
			 user->hide.usermask = 1;
			 irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick,
				     "Your last seen user@host mask will now be hidden from NickServ INFO displays.");
			 return;
		    } else if (!strcasecmp(args[2], "QUIT")) {
			 user->hide.quit = 1;
			 irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick,
			 "Your last quit message will now be hidden from NickServ INFO displays.");
			 return;
		    } else {
			 irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick,
				     "Syntax: \002SET HIDE {EMAIL | USERMASK | QUIT} {ON | OFF}");
			 irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick,
				     "\002/msg NickServ HELP SET HIDE\002 for more information.");
			 return;
		    }
	       } else if (!strcasecmp(args[3], "OFF")) {
		    if (!strcasecmp(args[2], "EMAIL")) {
			 user->hide.email = 0;
			 irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick,
			      "Your E-mail address will now be shown in NickServ INFO displays.");
			 return;
		    } else if (!strcasecmp(args[2], "USERMASK")) {
			 user->hide.usermask = 0;
			 irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick,
				     "Your last seen user@host mask will now be shown in NickServ INFO displays.");
			 return;
		    } else if (!strcasecmp(args[2], "QUIT")) {
			 user->hide.quit = 0;
			 irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick,
			   "Your last quit message will now be shown in NickServ INFO displays.");
			 return;
		    } else {
			 irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick,
				     "Syntax: \002SET HIDE {EMAIL | USERMASK | QUIT} {ON | OFF}");
			 irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick,
				     "\002/msg NickServ HELP SET HIDE\002 for more information.");
			 return;
		    }
	       } else {
		    irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick,
				"Syntax: \002SET HIDE {EMAIL | USERMASK | QUIT} {ON | OFF}\002");
		    irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick,
				"\002/msg NickServ HELP SET HIDE\002 for more information.");
		    return;
	       }
	  } else {
	       irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick,
			   "Password authentication required for that command.");
	       irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick,
			   "\002/msg NickServ HELP SET\002 for more information.");
	       return;
	  }
     }
}

struct ns_set_options {
     char *option;
     void (*handler) (struct fd_list *, char **, int);
     int nargs;
};

typedef struct ns_set_options NS_SetOptions;

NS_SetOptions opts[] =
{
     {"PASSWORD", ns_set_password, 3},
     {"LANGUAGE", ns_set_language, 3},
     {"URL", ns_set_url, 2},
     {"EMAIL", ns_set_email, 2},
     {"KILL", ns_set_kill, 3},
     {"SECURE", ns_set_secure, 3},
     {"PRIVATE", ns_set_private, 3},
     {"HIDE", ns_set_hide, 4},
     {NULL, NULL, 1}
};

void ns_set(struct fd_list *fd, char **args, int nargs)
{
     NS_SetOptions *p;

     p = opts;

     if (nargs > 1) {
	  while (p->option != NULL) {
	       if (!strcasecmp(p->option, args[1])) {
		    if (nargs >= p->nargs) {
			 p->handler(fd, args, nargs);
			 return;
		    } else {
			 irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick,
				     "Syntax: \002SET\002 option parameters");
			 irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick,
				     "\002/msg NickServ HELP SET\002 for more information.");
			 return;
		    }
	       }
	       p++;
	  }

	  {
	       char tmp[255];

	       snprintf(tmp, 255, "Unknown SET option \002%s\002.", args[1]);
	       irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, tmp);
	  }
	  return;
     } else {
	  irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick,
		      "Syntax: \002SET\002 option parameters");
	  irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick,
		      "\002/msg NickServ HELP SET\002 for more information.");
	  return;
     }
}

void ns_drop(struct fd_list *fd, char **args, int nargs)
{
     NS_User *user = find_user(fd->nick);

     if (user == NULL) {
	  irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick,
		      "Your nick isn't registered.");
	  return;
     }
     if (!(fd->registered & NS_IS_IDENTIFIED)) {
	  irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick,
		      "Password authentication required for that command.");
	  irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick,
		      "\002/msg NickServ HELP DROP\002 for more information.");
	  return;
     }
     if (!strcasecmp(fd->nick, user->nick)) {
	  {
	       NS_AccessList *acl = access_list;

	       while (acl != NULL) {
		    if (acl->user == user) {
			 if (acl->prev == NULL && acl->next == NULL) {
			      access_list = NULL;
			      free(acl);
			      acl = NULL;
			 } else if (acl->prev == NULL && acl->next != NULL) {
			      access_list = acl->next;
			      (acl->next)->prev = NULL;
			      free(acl);
			      acl = access_list;
			 } else if (acl->prev != NULL && acl->next == NULL) {
			      (acl->prev)->next = NULL;
			      free(acl);
			      acl = NULL;
			 } else if (acl->prev != NULL && acl->next != NULL) {
			      NS_AccessList *p = acl->next;

			      (acl->prev)->next = acl->next;
			      (acl->next)->prev = acl->prev;
			      free(acl);
			      acl = p;
			 }
		    } else {
			 acl = acl->next;
		    }
	       }
	  }
	  {
	       NS_LinkList *links;

	       links = link_list;
	       while (links != NULL) {
		    if (links->user == user) {
			 if (links->prev == NULL && links->next == NULL) {
			      link_list = NULL;
			      free(links);
			      links = NULL;
			 } else if (links->prev == NULL && links->next != NULL) {
			      link_list = links->next;
			      (links->next)->prev = NULL;
			      free(links);
			      links = link_list;
			 } else if (links->prev != NULL && links->next == NULL) {
			      (links->prev)->next = NULL;
			      free(links);
			      links = NULL;
			 } else if (links->prev != NULL && links->next != NULL) {
			      NS_LinkList *p = links->next;

			      (links->prev)->next = links->next;
			      (links->next)->prev = links->prev;
			      free(links);
			      links = p;
			 }
		    } else {
			 links = links->next;
		    }
	       }
	  }
	  if (user->prev == NULL && user->next == NULL) {
	       registered_nicks = NULL;
	       free(user);
	  } else if (user->prev == NULL && user->next != NULL) {
	       registered_nicks = user->next;
	       (user->next)->prev = NULL;
	       free(user);
	  } else if (user->prev != NULL && user->next == NULL) {
	       (user->prev)->next = NULL;
	       free(user);
	  } else if (user->prev != NULL && user->next != NULL) {
	       (user->prev)->next = user->next;
	       (user->next)->prev = user->prev;
	       free(user);
	  }
	  fd->registered = 1;
	  irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick,
		      "Your nickname has been dropped.");
	  return;
     } else {
	  NS_LinkList *link = link_list;

	  while (link != NULL) {
	       if (!strcasecmp(link->nick, fd->nick)) {
		    if (!link->prev && !link->next) {
			 link_list = NULL;
			 free(link);
		    } else if (link->prev && !link->next) {
			 (link->prev)->next = NULL;
			 free(link);
		    } else if (!link->prev && link->next) {
			 (link->next)->prev = NULL;
			 link_list = link->next;
			 free(link);
		    } else if (link->prev && link->next) {
			 (link->prev)->next = link->next;
			 (link->next)->prev = link->prev;
			 free(link);
		    }
		    fd->registered = 1;
		    irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick,
				"Your nickname has been dropeed.");
		    return;
	       }
	       link = link->next;
	  }
     }
}

void ns_help(struct fd_list *fd, char **args, int nargs)
{
     if (nargs > 1) {
	  if (!strcasecmp(args[1], "REGISTER")) {
	       char *line1 = "Syntax: \002REGISTER\002 password", *line2 = " ", *line3 = "Registers your nickname in the NickServ database.  Once",
	        *line4 = "your nick is registered, you can use the \002SET\002 and \002ACCESS\002",
	        *line5 = "commands to configure your nick's settings as you like", *line6 = "them.  Make sure you remember the password you use when",
	        *line7 = "registering - you'll need it to make changes to your nick", *line8 = "later. (Note that \002case matters!  FIDO\002, \002Fido\002, and \002fido\002",
	        *line9 = "are all different passwords!)", *line10 = " ", *line11 = "Guidelines on choosing passwords:",
	        *line12 = " ", *line13 = "Passwords should not be easily guessable.  For example,",
	        *line14 = "using your real name as a password is a bad idea.  Using", *line15 = "your nickname as a password is a much worse idea ;) and,",
	        *line16 = "in fact, NickServ will not allow it. Also, short", *line17 = "passwords are vulnerable to trial-and-error searches, so",
	        *line18 = "you should choose a password at least 5 characters long.", *line19 = "Finally, the space character cannot be used in passwords.";

	       irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line1);
	       irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line2);
	       irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line3);
	       irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line4);
	       irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line5);
	       irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line6);
	       irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line7);
	       irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line8);
	       irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line9);
	       irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line10);
	       irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line11);
	       irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line12);
	       irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line13);
	       irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line14);
	       irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line15);
	       irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line16);
	       irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line17);
	       irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line18);
	       irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line19);
	  } else if (!strcasecmp(args[1], "IDENTIFY")) {
	       char *line1 = "Syntax: \002IDENTIFY\002 password", *line2 = " ", *line3 = "Tells NickServ that you really are the owner of this",
	        *line4 = "nick.  Many commands require you to authenticate yourself", *line5 = "with this command before you use them.  The password",
	        *line6 = "should be the same one you sent with the \002REGISTER\002", *line7 = "command.";

	       irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line1);
	       irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line2);
	       irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line3);
	       irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line4);
	       irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line5);
	       irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line6);
	       irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line7);
	  } else if (!strcasecmp(args[1], "ACCESS")) {
	       char *line1 = "Syntax: \002ACCESS ADD\002 mask", *line2 = "        \002ACCESS DEL\002 mask",
	        *line3 = "        \002ACCESS LIST\002", *line4 = " ", *line5 = "Modifies or displays the access list for your nick.  This",
	        *line6 = "is the list of addresses which will be automatically", *line7 = "recognized by NickServ as allowed to use the nick.  If",
	        *line8 = "you want to use the nick from a different address, you", *line9 = "need to send an \002IDENTIFY\002 command to make NickServ",
	        *line10 = "recognize you.", *line11 = " ", *line12 = "Examples:", *line13 = " ",
	        *line14 = "    \002ACCESS ADD achurch@*.dragonfire.net\002", *line15 = "        Allows access to user \002achurch\002 from any machine in",
	        *line16 = "        the \002dragonfire.net\002 domain.", *line17 = " ", *line18 = "    \002ACCESS DEL achurch@*.dragonfire.net\002",
	        *line19 = "        Reverses the previous command.", *line20 = " ", *line21 = "    \002ACCESS LIST\002",
	        *line22 = "        Displays the current access list.";

	       irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line1);
	       irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line2);
	       irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line3);
	       irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line4);
	       irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line5);
	       irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line6);
	       irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line7);
	       irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line8);
	       irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line9);
	       irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line10);
	       irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line11);
	       irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line12);
	       irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line13);
	       irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line14);
	       irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line15);
	       irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line16);
	       irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line17);
	       irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line18);
	       irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line19);
	       irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line20);
	       irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line21);
	       irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line22);
	  } else if (!strcasecmp(args[1], "LINK")) {
	       char *line1 = "Syntax: \002LINK\002 nick password", *line2 = " ", *line3 = "Links your nickname to another, effectively making the nick",
	        *line4 = "you are using an alias for the given nick.  When", *line5 = "you give this command, the access list for the nick you are",
	        *line6 = "using is deleted and replaced by that of the nick you are", *line7 = "linking to; all memos for your current nick are added to",
	        *line8 = "the list of memos for the target nick (this may cause the", *line9 = "nick to exceed it's limit of memos, in which case you will",
	        *line10 = "need to delete some before you can receive more for either", *line11 = "nick).",
	        *line12 = " ", *line13 = "Once the link has been established, your nick will be",
	        *line14 = "transparently converted into the target nick everywhere in", *line15 = "Services, except for the NickServ \002UNLINK\002 and \002DROP\002 commands.  For",
	        *line16 = "example, you can use either nick to read the same set of", *line17 = "memos (and memos sent to either nick will go into the same",
	        *line18 = "list).  If you identify for one nick, you will", *line19 = "automatically be identified for the other as well.  If",
	        *line20 = "either nick is on a channel's access list, then both nicks", *line21 = "will get the same access priveliges.",
	        *line22 = " ", *line23 = "In order to use this command, you must identify for your",
	        *line24 = "current nick (using the \002IDENTIFY\002 command), and you must",
	        *line25 = "supply the password for the nick you wish to link to.";

	       irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line1);
	       irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line2);
	       irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line3);
	       irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line4);
	       irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line5);
	       irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line6);
	       irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line7);
	       irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line8);
	       irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line9);
	       irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line10);
	       irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line11);
	       irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line12);
	       irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line13);
	       irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line14);
	       irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line15);
	       irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line16);
	       irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line17);
	       irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line18);
	       irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line19);
	       irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line20);
	       irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line21);
	       irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line22);
	       irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line23);
	       irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line24);
	       irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line25);
	  } else if (!strcasecmp(args[1], "SET")) {
	       if (nargs > 2) {
		    if (!strcasecmp(args[2], "PASSWORD")) {
			 char *line1 = "Syntax: \002SET PASSWORD\002 new-password", *line2 = " ",
			  *line3 = "Changes the password you use to identify as the nick's",
			  *line4 = "owner.";

			 irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line1);
			 irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line2);
			 irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line3);
			 irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line4);
		    } else if (!strcasecmp(args[2], "LANGUAGE")) {
			 char *line1 = "Syntax: \002SET LANGUAGE\002 number", *line2 = " ", *line3 = "Changes the language Services uses when sending messages to",
			  *line4 = "you (for example, when replying to a command you send).",
			  *line5 = "\002number\002 should be chosen from the following list of",
			  *line6 = "supported languages:", *line7 = "     1) English";

			 irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line1);
			 irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line2);
			 irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line3);
			 irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line4);
			 irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line5);
			 irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line6);
			 irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line7);
		    } else if (!strcasecmp(args[2], "URL")) {
			 char *line1 = "Syntax: \002SET URL\002 url", *line2 = " ", *line3 = "Associates the given URL with your nickname.  This URL",
			  *line4 = "will be displayed whenever someone requests information",
			  *line5 = "on your nick with the \002INFO\002 command.";

			 irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line1);
			 irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line2);
			 irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line3);
			 irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line4);
			 irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line5);
		    } else if (!strcasecmp(args[2], "EMAIL")) {
			 char *line1 = "Syntax: \002SET EMAIL\002 address", *line2 = " ", *line3 = "Associates the given E-mail address with your nickname.",
			  *line4 = "This information will be displayed whenever someone requests",
			  *line5 = "information on the channel with the \002INFO\002 command.";

			 irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line1);
			 irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line2);
			 irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line3);
			 irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line4);
			 irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line5);
		    } else if (!strcasecmp(args[2], "KILL")) {
			 char *line1 = "Syntax: \002SET KILL {ON | QUICK | IMMED | OFF}\002",
			  *line2 = " ", *line3 = "Turns the automatic kill protection option for your nick",
			  *line4 = "on or off.  With kill protection on, if another user", *line5 = "tries to take your nick, they will be given one minute to",
			  *line6 = "change to another nick, after which they will be forcibly",
			  *line7 = "removed from IRC by NickServ.", *line8 = " ", *line9 = "If you select \002QUICK\002, the user will be given only 20 seconds",
			  *line10 = "to change nicks instead of the usual 60.  If you select",
			  *line11 = "\002IMMED\002, the user will be killed immediately \002without\002 being",
			  *line12 = "warned first or given a chance to change their nick; please",
			  *line13 = "do not use this option unless necessary.  Also, your", *line14 = "network's administrators may have disabled this option.";

			 irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line1);
			 irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line2);
			 irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line3);
			 irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line4);
			 irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line5);
			 irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line6);
			 irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line7);
			 irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line8);
			 irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line9);
			 irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line10);
			 irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line11);
			 irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line12);
			 irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line13);
			 irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line14);
		    } else if (!strcasecmp(args[1], "SECURE")) {
			 char *line1 = "Syntax: \002SET SECURE {ON | OFF}\002", *line2 = " ",
			  *line3 = "Turns NickServ's security features on or off for your", *line4 = "nick.  With \002SECURE\002 set, you must enter your password",
			  *line5 = "before you will be recognized as owner of the nick,", *line6 = "regardless of whether your address is on the access",
			  *line7 = "list.  However, if you are on the access list, NickServ",
			  *line8 = "will not auto-kill you regardless of the setting of the",
			  *line9 = "\002KILL\002 option.";

			 irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line1);
			 irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line2);
			 irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line3);
			 irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line4);
			 irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line5);
			 irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line6);
			 irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line7);
			 irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line8);
			 irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line9);
		    }
	       } else {
		    char *line1 = "Syntax: \002SET [function] [params]", *line2 = " ", *line3 = "Sets options for your nick in NickServ.",
		     *line4 = " ", *line5 = "Functions are:", *line6 = "\002PASSWORD\002 change password",
		     *line7 = "\002LANGUAGE\002 change language", *line8 = "\002URL     \002 change url",
		     *line9 = "\002EMAIL   \002 change email", *line10 = "\002KILL    \002 change kill setting",
		     *line11 = "\002SECURE  \002 change security", *line12 = " ", *line13 = "/msg NickServ HELP SET function for more help.";

		    irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line1);
		    irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line2);
		    irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line3);
		    irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line4);
		    irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line5);
		    irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line6);
		    irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line7);
		    irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line8);
		    irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line9);
		    irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line10);
		    irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line11);
		    irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line12);
		    irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line13);
	       }
	  } else {
	       irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, "Oops.");
	  }
     } else {
	  char *line1 = "\002NickServ\002 allows you to \"register\" a nickname and", *line2 = "prevent others from using them. If the nick is not used",
	   *line3 = "for 30 days, the registration will expire.  The following", *line4 = "commands allow registration and maintenance of",
	   *line5 = "nicknames; to use them, use \002/MSG NickServ <command>\002.", *line6 = "For more information on a specific command, type",
	   *line7 = "\002/MSG NickServ HELP <command>\002.", *line8 = " ", *line9 = "   REGISTER   Register a nickname",
	   *line10 = "   IDENTIFY   Identify yourself with your password", *line11 = "   ACCESS     Modify the list of authorized addresses",
	   *line12 = "   LINK       Link another nickname to your primary one", *line13 = "   SET        Set options, including kill protection",
	   *line14 = "   DROP       Cancel the registration of a nickname", *line15 = "   RECOVER    Reclaim your nick after another user has taken it",
	   *line18 = " ", *line19 = "Other commands: UNLINK, GHOST, INFO, LIST, STATUS", *line20 = " ",
	   *line21 = "\002NOTICE:\x02 This service is intended to provide a way for", *line22 = "IRC users to ensure their identity is not compromised.",
	   *line23 = "It is \002NOT\002 intended to facilitate \"stealing\" of", *line24 = "nicknames or other malicious actions. Abuse of NickServ",
	   *line25 = "will result in, at minimum, loss of the abused nickname(s).";

	  irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line1);
	  irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line2);
	  irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line3);
	  irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line4);
	  irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line5);
	  irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line6);
	  irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line7);
	  irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line8);
	  irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line9);
	  irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line10);
	  irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line11);
	  irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line12);
	  irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line13);
	  irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line14);
	  irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line15);
	  irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line18);
	  irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line19);
	  irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line20);
	  irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line21);
	  irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line22);
	  irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line23);
	  irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line24);
	  irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, line25);
     }
};

void ns_info(struct fd_list *fd, char **args, int nargs)
{
     if (nargs < 2) {
	  irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, "Syntax: \002INFO nick [ALL]\002");
	  irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick,
		      "\002/msg NickServ HELP INFO\002 for more information.");
	  return;
     } else {
	  NS_User *user = find_user(args[1]);

	  if (user == NULL) {
	       char tmp[255];

	       snprintf(tmp, 255, "Nick \002%s\002 isn't registered.", args[1]);
	       irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, tmp);
	       return;
	  } else {
	       char tmp[255];
	       char *nick;

	       bzero(tmp, 255);
	       if (!strcasecmp(args[1], user->nick)) {
		    nick = user->nick;
	       } else {
		    nick = args[1];
	       }
	       snprintf(tmp, 255, "%s is %s", nick, user->last_realname);
	       irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, tmp);
	       {
		    struct fd_list *p;
		    int is_online = 0;

		    p = __head->next;
		    while (p != NULL) {
			 if (!strcasecmp(p->nick, nick)) {
			      if (fd->registered & NS_IS_IDENTIFIED) {
				   is_online = 1;
				   p = NULL;
			      } else {
				   p = NULL;
			      }
			 }
			 if (p)
			      p = p->next;
		    }

		    if (user->last_mask[0] != 0) {
			 if (is_online) {
			      snprintf(tmp, 255, "   Is online from: %s", user->last_mask);
			      irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, tmp);
			 } else {
			      snprintf(tmp, 255, "Last seen address: %s", user->last_mask);
			      irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, tmp);
			      snprintf(tmp, 255, "   Last seen time: %s", ctime((time_t *) & (user->last_seen_at)));
			      irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, tmp);
			 }
		    }
	       }
	       snprintf(tmp, 255, "  Time registered: %s", ctime((time_t *) & (user->registered_at)));
	       irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, tmp);
	       if (user->last_quit[0] != 0) {
		    snprintf(tmp, 255, "Last quit message: %s", user->last_quit);
		    irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, tmp);
	       } {
		    char options[255];

		    bzero(options, 255);
		    if (user->kill) {
			 snprintf(options, 255, "Kill protection");
		    }
		    if (user->secure) {
			 if (options[0] == 0) {
			      snprintf(options, 255, "Security");
			 } else {
			      snprintf(options, 255, "%s, Security", options);
			 }
		    }
		    if (user->private) {
			 if (options[0] == 0) {
			      snprintf(options, 255, "Private");
			 } else {
			      snprintf(options, 255, "%s, Private", options);
			 }
		    }
		    if (options[0] == 0) {
			 snprintf(options, 255, "None");
		    }
		    snprintf(tmp, 255, "          Options: %s", options);
		    irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, tmp);
	       }
	       return;
	  }
     }
}

void ns_version(struct fd_list *fd, char **args, int nargs)
{
     irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname,
		 fd->nick, "\001VERSION Services v0.3 for ircd 0.2\001");
     return;
}

int guest_num = 0;

extern int *die, *quit;
extern char *parse(char *, struct fd_list *, int *, int *);

/* recover <nick> <pass> */
void ns_recover(struct fd_list *fd, char **args, int nargs)
{
     if (nargs >= 3) {
	  struct fd_list *u = __head->next;

	  if (!strcasecmp(args[1], fd->nick)) {
	       irc_fprintf(fd->sok, ":%s!%s@%s NOTICE %s :You can't recover yourself!\r\n", s_NickServ, ns_userid,
			   ns_hostname, fd->nick);
	       return;
	  }
	  while (u) {
	       if (!strcasecmp(u->nick, args[1]))
		    goto found_user;
	       u = u->next;
	  }
	  goto no_such_user;

	found_user:{
	       NS_User *ul = find_user(args[1]);

	       if (!ul)
		    goto not_registered;

	       if (!strcmp(args[2], ul->pass))
		    goto pass_ok;
	       goto bad_pass;

	     pass_ok:{
		    /* change u->nick to a guest nick */
		    char tmpnick[21];
		    int nuh, *my_quit, *my_die;

		    snprintf(tmpnick, 21, "NICK Guest%i\r", guest_num++);
		    irc_fprintf(u->sok, ":%s!%s@%s NOTICE %s :This nick was recovered by %s!\r\n", s_NickServ,
				ns_userid, ns_hostname, u->nick, fd->nick);
		    my_quit = quit;
		    my_die = die;
		    parse(tmpnick, u, &nuh, &nuh);
		    /* change fd->nick to the recovered nick */
		    snprintf(tmpnick, 21, "NICK %s\r", args[1]);
		    parse(tmpnick, fd, &nuh, &nuh);
		    my_quit = quit;
		    my_die = die;
		    fd->registered |= NS_IS_IDENTIFIED;
		    u->registered &= ~NS_IS_IDENTIFIED;
		    return;
	       }

	     bad_pass:{
		    irc_fprintf(fd->sok, ":%s!%s@%s NOTICE %s :Bad password!\r\n", s_NickServ, ns_userid,
				ns_hostname, fd->nick);
		    return;
	       }

	     not_registered:{
		    irc_fprintf(fd->sok, ":%s!%s@%s NOTICE %s :The nickname \002%s\002 is not registered!\r\n",
				s_NickServ, ns_userid, ns_hostname, fd->nick, args[1]);
		    return;
	       }
	  }

	no_such_user:{
	       irc_fprintf(fd->sok, ":%s!%s@%s NOTICE %s :The nickname \002%s\002 is not in use!\r\n",
			   s_NickServ, ns_userid, ns_hostname, fd->nick, args[1]);
	       return;
	  }
     } else {
	  irc_fprintf(fd->sok, ":%s!%s@%s NOTICE %s :Syntax: \002RECOVER\002 nickname \002[\002password\002]\002\r\n",
		      s_NickServ, ns_userid, ns_hostname, fd->nick);
	  irc_fprintf(fd->sok, ":%s!%s@%s NOTICE %s :\002/msg NickServ HELP RECOVER\002 for more information.\r\n",
		      s_NickServ, ns_userid, ns_hostname, fd->nick);
	  return;
     }
}

extern void m_kill(struct fd_list *, char **, int);

void ns_ghost(struct fd_list *fd, char **args, int nargs)
{
     if (nargs >= 3) {
	  NS_User *u = find_user(args[1]);

	  if (!u) {
	       irc_fprintf(fd->sok, ":%s!%s@%s NOTICE %s :The nickname \002%s\002 has not been registered!\r\n",
			   s_NickServ, ns_userid, ns_hostname, fd->nick, args[1]);
	       return;
	  }
	  if (!strcasecmp(fd->nick, args[1])) {
	       irc_fprintf(fd->sok, ":%s!%s@%s NOTICE %s :You can't GHOST yourself!\r\n", s_NickServ, ns_userid, ns_hostname, fd->nick);
	       return;
	  }
	  if (!strcmp(u->pass, args[2])) {
	       struct fd_list p;
	       char kmsg[100];
	       char *argv[3];
	       int argc = 3;

	       bzero(&p, sizeof(p));
	       snprintf(p.nick, 15, "%s", s_NickServ);
	       snprintf(p.userid, 15, "%s", ns_userid);
	       snprintf(p.hostname, 15, "%s", ns_hostname);
	       bzero(kmsg, 100);
	       argv[0] = "KILL";
	       argv[1] = args[1];
	       argv[2] = kmsg;
	       snprintf(kmsg, 100, "GHOST command used by %s", fd->nick);
	       m_kill(&p, argv, argc);
	       return;
	  } else {
	       irc_fprintf(fd->sok, ":%s!%s@%s NOTICE %s :Incorrect password!\r\n", s_NickServ, ns_userid, ns_hostname, fd->nick);
	       return;
	  }
     } else {
	  irc_fprintf(fd->sok, ":%s!%s@%s NOTICE %s :Syntax: \002GHOST nick password\002\r\n", s_NickServ, ns_userid, ns_hostname,
		      fd->nick);
	  irc_fprintf(fd->sok, ":%s!%s@%s NOTICE %s :\002/msg NickServ HELP GHOST\002 for more information.\r\n", s_NickServ, ns_userid,
		      ns_hostname, fd->nick);
	  return;
     }
}

NS_Functions fn_list[] =
{
     {"REGISTER", ns_register, 0},
     {"IDENTIFY", ns_identify, 0},
     {"ACCESS", ns_access, 1},
     {"LINK", ns_link, 1},
     {"SET", ns_set, 1},
     {"DROP", ns_drop, 1},
     {"RECOVER", ns_recover, 0},
     {"UNLINK", ns_dummy, 1},
     {"GHOST", ns_ghost, 0},
     {"INFO", ns_info, 0},
     {"LIST", ns_dummy, 0},
     {"STATUS", ns_dummy, 0},
     {"HELP", ns_help, 0},
     {"\001VERSION\001", ns_version, 0},
     {NULL, NULL, 0}
};

int num_funcs = 14;

void nickserv_msg_handle(struct fd_list *fd, char **av, int ac)
{
     char **args, *input, *tmp;
     int i, nargs;

     if (ac > 2) {
	  input = av[2];
	  tmp = input;
	  args = NULL;
	  nargs = 0;
	  while (*input != 0) {
	       if (*input == ' ') {
		    args = (char **) realloc(args, sizeof(char **) * ++nargs);

		    args[nargs - 1] = tmp;
		    *input = 0;
		    input++;
		    tmp = input;
	       } else {
		    input++;
	       }
	  }
	  args = (char **) realloc(args, sizeof(char **) * ++nargs);

	  args[nargs - 1] = tmp;
	  args = (char **) realloc(args, sizeof(char **) * nargs + 1);

	  args[nargs] = NULL;
     } else {
	  irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, "I need a command to parse.");
	  irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick,
		      "\x02/MSG NickServ HELP\x02 for more information.");
	  return;
     }

     for (i = 0; i < num_funcs; i++) {
	  if (!strcasecmp(args[0], fn_list[i].cmd)) {
	       fn_list[i].function(fd, args, nargs);
	       free(args);
	       return;
	  }
     }

     irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, "Unknown command.");
     irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick,
		 "\x02/MSG NickServ HELP\x02 for more information.");
     free(args);
     return;
}

struct cs_func {
     char *cmd;
     void (*function) (struct fd_list *, char **, int);
     int need_ident;
};

typedef struct cs_func CS_Functions;

void cs_register(struct fd_list *fd, char **args, int nargs)
{
     NS_User *user = find_user(fd->nick);

     if (!user) {
	  irc_fprintf(fd->sok, ":%s!%s@%s NOTICE %s :You must register your nickname first. Type \x02/msg NickServ HELP\x02 for more information on registering nicknames.\r\n",
		      s_ChanServ, cs_userid, cs_hostname, fd->nick);
	  return;
     }
     if (nargs >= 4) {
	  CS_Channel *new_chan;
	  CS_AccessList *new_acl;

	  new_chan = malloc(sizeof(CS_Channel));
	  bzero(new_chan, sizeof(CS_Channel));
	  new_acl = malloc(sizeof(CS_AccessList));
	  bzero(new_acl, sizeof(CS_AccessList));

	  new_chan->founder = user;
	  new_chan->founder_user_num = user->user_num;
	  new_chan->successor = NULL;
	  new_chan->successor_user_num = -1;
	  snprintf(new_chan->password, 15, "%s", args[2]);
	  snprintf(new_chan->name, 20, "%s", args[1]);
	  {
	       int i, l = 0;

	       for (i = 3; i < nargs; i++) {
		    snprintf(&(new_chan->description[l]), 100 - l, "%s ", args[i]);
		    l += strlen(args[i]) + 1;
	       }
	  }

	  new_acl->user = user;
	  new_acl->user_num = user->user_num;
	  new_chan->acl = new_acl;
	  new_acl->level = 16;
	  new_chan->modes.no_outside_msg = 1;
	  new_chan->modes.oper_only_topic = 1;
	  {
	       CS_Channel *l = registered_channels;

	       while (l->next)
		    l = l->next;
	       l->next = new_chan;
	       new_chan->prev = l;
	  }

	  irc_fprintf(fd->sok, ":%s!%s@%s NOTICE %s :Channel \x02%s\x02 registered under your nickname: %s\r\n",
		      s_ChanServ, cs_userid, cs_hostname, fd->nick, new_chan->name, fd->nick);
	  irc_fprintf(fd->sok, ":%s!%s@%s NOTICE %s :Your password is \x02%s\x02 - remember it for later use.\r\n",
		      s_ChanServ, cs_userid, cs_hostname, fd->nick, new_chan->password);
	  return;
     } else {
	  irc_fprintf(fd->sok, ":%s!%s@%s NOTICE %s :Syntax: \x02REGISTER\x02 channel password description\r\n",
		      s_ChanServ, cs_userid, cs_hostname, fd->nick);
	  irc_fprintf(fd->sok, ":%s!%s@%s NOTICE %s :\x02/msg ChanServ HELP REGISTER\x02 for more information.\r\n",
		      s_ChanServ, cs_userid, cs_hostname, fd->nick);
	  return;
     }
}

/* use 0x20xxxx00 for channel and channel number -
   0x20000000 to check that a user has identified for a channel,
   0x00xxxx00 for the channel number
 */

#if 0

int cs_is_identified(CS_Channel * chan, struct fd_list *user)
{
     NS_User *u = find_user(user->nick);

     if (!u)
	  return 0;
     if (user->registered & 0x20000000) {
	  int q = user->registered;
	  int r = q << 8;	/* shift off the top byte */
	  int s = r >> 16;	/* shift off lower byte .. and the byte that was shifted up */

	  if (s == chan->chan_num) {
	       return 1;
	  } else {
	       return 0;
	  }
     } else {
	  return 0;
     }

     return 0;
}

CS_Functions cs_func_list[] =
{
     {"REGISTER", cs_register, 1},
     {NULL, NULL, 0}
};

#endif

struct os_func {
     char *cmd;
     void (*function) (struct fd_list *, char **, int);
     int nargs;
     int need_oper;
     int need_admin;
     int need_root;
};

typedef struct os_func OS_Functions;

void os_global(struct fd_list *fd, char **args, int nargs)
{
     char *message = NULL;

     if (nargs > 1) {
	  int i, len = 0;

	  for (i = 1; i < nargs; i++) {
	       len = len + strlen(args[i]) + 1;
	  }

	  message = (char *) malloc(len);
	  len = 0;
	  for (i = 1; i < nargs; i++) {
	       sprintf(&(message[len]), "%s ", args[i]);
	       len = len + strlen(args[i]) + 1;
	  }

	  {
	       struct fd_list *users;

	       users = __head->next;
	       while (users != NULL) {
		    irc_fprintf(users->sok, NOTICE, "Global", "global", "services.net", users->nick, message);
		    users = users->next;
	       }
	  }

	  free(message);
     } else {
	  irc_fprintf(fd->sok, NOTICE, s_OperServ, os_userid, os_hostname, fd->nick, "Syntax: \002GLOBAL message\002");
	  irc_fprintf(fd->sok, NOTICE, s_OperServ, os_userid, os_hostname, fd->nick,
		      "\002/msg OperServ HELP GLOBAL\002 for more information.");
     }
     return;
}

void os_stats(struct fd_list *fd, char **args, int nargs)
{
     if (nargs > 1) {
	  if (!strcasecmp(args[1], "AKILL")) {
	       char tmp[255];

	       snprintf(tmp, 255, "Current number of AKILLs: \002%i\002", 0);
	       irc_fprintf(fd->sok, NOTICE, s_OperServ, os_userid, os_hostname, fd->nick, tmp);
	       snprintf(tmp, 255, "Default AKILL expiry time: \002%i days\002", 30);
	       irc_fprintf(fd->sok, NOTICE, s_OperServ, os_userid, os_hostname, fd->nick, tmp);
	       return;
	  } else if (!strcasecmp(args[1], "ALL")) {
	       irc_fprintf(fd->sok, NOTICE, s_OperServ, os_userid, os_hostname, fd->nick,
			   "OperServ \002STATS ALL\002 is broken right now.");
	       return;
	  } else {
	       irc_fprintf(fd->sok, NOTICE, s_OperServ, os_userid, os_hostname, fd->nick,
			   "Syntax: \002STATS [AKILL | ALL]\002");
	       irc_fprintf(fd->sok, NOTICE, s_OperServ, os_userid, os_hostname, fd->nick,
			   "\002/msg OperServ HELP STATS\002 for more information.");
	       return;
	  }
     } else {
	  char tmp[255];
	  struct fd_list *list = __head->next;
	  int p = 0, ops = 0;

	  while (list != NULL) {
	       p++;
	       if (list->oper)
		    ops++;
	       list = list->next;
	  }
	  snprintf(tmp, 255, "Current users: \002%i\002 (\002%i\002 ops)", p, ops);
	  irc_fprintf(fd->sok, NOTICE, s_OperServ, os_userid, os_hostname, fd->nick, tmp);
	  irc_fprintf(fd->sok, NOTICE, s_OperServ, os_userid, os_hostname, fd->nick, "Maximum users: Unknown");
	  {
	       int d, h, m, s, now, t;

	       now = __time(NULL);
	       t = now - started;

	       d = t / 60 / 60 / 24;
	       t = t - (d * 60 * 60 * 24);
	       h = t / 60 / 60;
	       t = t - (h * 60 * 60);
	       m = t / 60;
	       t = t - (m * 60);
	       s = t;
	       irc_fprintf(fd->sok, ":%s!%s@%s NOTICE %s :Services up ", s_OperServ, os_userid, os_hostname, fd->nick);
	       if (!d) {
		    if (!h) {
			 if (!m) {
			      irc_fprintf(fd->sok, "\x02%i\x02 seconds.\r\n", s);
			 } else {
			      irc_fprintf(fd->sok, "\x02%i\x02 minutes, ", m);
			      irc_fprintf(fd->sok, "\x02%i\x02 seconds.\r\n", s);
			 }
		    } else {
			 irc_fprintf(fd->sok, "\x02%i\x02 hours, ", h);
			 if (!m) {
			      irc_fprintf(fd->sok, "\x02%i\x02 seconds.\r\n", s);
			 } else {
			      irc_fprintf(fd->sok, "\x02%i\x02 minutes, ", m);
			      irc_fprintf(fd->sok, "\x02%i\x02 seconds.\r\n", s);
			 }
		    }
	       } else {
		    irc_fprintf(fd->sok, "\x02%i\x02 days, ", d);
		    if (!h) {
			 if (!m) {
			      irc_fprintf(fd->sok, "\x02%i\x02 seconds.\r\n", s);
			 } else {
			      irc_fprintf(fd->sok, "\x02%i\x02 minutes, ", m);
			      irc_fprintf(fd->sok, "\x02%i\x02 seconds.\r\n", s);
			 }
		    } else {
			 irc_fprintf(fd->sok, "\x02%i\x02 hours, ", h);
			 if (!m) {
			      irc_fprintf(fd->sok, "\x02%i\x02 seconds.\r\n", s);
			 } else {
			      irc_fprintf(fd->sok, "\x02%i\x02 minutes, ", m);
			      irc_fprintf(fd->sok, "\x02%i\x02 seconds.\r\n", s);
			 }
		    }
	       }
	  }
	  return;
     }
}

void os_dummy(struct fd_list *fd, char **args, int nargs)
{
     char tmp[255];

     snprintf(tmp, 255, "OperServ \002%s\002 doesn't work yet..", args[0]);
     irc_fprintf(fd->sok, NOTICE, s_OperServ, os_userid, os_hostname, fd->nick, tmp);
     return;
}

void os_op(struct fd_list *fd, char **args, int nargs)
{
     struct user_chan_list *chan = (struct user_chan_list *) fd->chan_list;

     if (nargs >= 2) {
	  while (chan != NULL) {
	       if (!strcasecmp(chan->chan->name, args[1])) {
		    struct user_list *u = chan->chan->users;

		    while (u) {
			 if (u->user == fd) {
			      u->oper = 1;
			 }
			 u = u->next;
		    }
		    u = chan->chan->users;
		    while (u) {
			 irc_fprintf(u->user->sok, ":%s MODE %s +o %s\r\n", os_server, args[1], fd->nick);
			 u = u->next;
		    }
		    chan = NULL;
	       }
	       if (chan)
		    chan = chan->next;
	  }
     }
}

#define os_oper os_dummy
#define os_admin os_dummy
#define os_mode os_dummy
#define os_kick os_dummy
#define os_akill os_dummy
#define os_session os_dummy
#define os_exception os_dummy
#define os_killclones os_dummy
#define os_logonnews os_dummy
#define os_opernews os_dummy
#define os_jupe os_dummy
#define os_raw os_dummy
#define os_set os_dummy
#define os_update os_dummy
#define os_quit os_dummy
#define os_restart os_dummy

void os_help(struct fd_list *fd, char **av, int ac)
{
     irc_fprintf(fd->sok, NOTICE, s_OperServ, os_userid, os_hostname, fd->nick, "I don't know how to do /operserv HELP yet.");
     return;
}

OS_Functions os_func_list[] =
{
     {"GLOBAL", os_global, 1, 0, 0},
     {"STATS", os_stats, 1, 0, 0},
     {"OPER", os_oper, 1, 0, 0},
     {"ADMIN", os_admin, 1, 0, 0},
     {"MODE", os_mode, 1, 0, 0},
     {"KICK", os_kick, 1, 0, 0},
     {"AKILL", os_akill, 1, 0, 0},
     {"SESSION", os_session, 0, 1, 0},
     {"EXCEPTION", os_exception, 0, 1, 0},
     {"KILLCLONES", os_killclones, 0, 1, 0},
     {"LOGONNEWS", os_logonnews, 0, 1, 0},
     {"OPERNEWS", os_opernews, 0, 1, 0},
     {"JUPE", os_jupe, 0, 1, 0},
     {"RAW", os_raw, 0, 1, 0},
     {"SET", os_set, 0, 1, 0},
     {"UPDATE", os_update, 0, 1, 0},
     {"QUIT", os_quit, 0, 1, 0},
     {"RESTART", os_restart, 0, 1, 0},
     {"OP", os_op, 1, 0, 0},
     {"HELP", os_help, 0, 0, 0},
     {NULL, NULL, 0, 0, 0}
};

int os_num_funcs = 19;

void operserv_msg_handle(struct fd_list *fd, char **av, int ac)
{
     char **args, *input, *tmp;
     int i, nargs;

     if (ac > 2) {
	  input = av[2];
	  tmp = input;
	  args = NULL;
	  nargs = 0;
	  while (*input != 0) {
	       if (*input == ' ') {
		    args = (char **) realloc(args, sizeof(char **) * ++nargs);

		    args[nargs - 1] = tmp;
		    *input = 0;
		    input++;
		    tmp = input;
	       } else {
		    input++;
	       }
	  }
	  args = (char **) realloc(args, sizeof(char **) * ++nargs);

	  args[nargs - 1] = tmp;
	  args = (char **) realloc(args, sizeof(char **) * nargs + 1);

	  args[nargs] = NULL;
     } else {
	  irc_fprintf(fd->sok, NOTICE, s_OperServ, os_userid, os_hostname, fd->nick, "I need a command to parse.");
	  irc_fprintf(fd->sok, NOTICE, s_OperServ, os_userid, os_hostname, fd->nick,
		      "\x02/MSG OperServ HELP\x02 for more information.");
	  return;
     }

     if (!fd->oper) {
	  irc_fprintf(fd->sok, NOTICE, s_OperServ, os_userid, os_hostname, fd->nick, "Access Denied - OperServ is only "
		      "availiable to IRC Operators!");
	  return;
     }
     for (i = 0; i < os_num_funcs; i++) {
	  if (!strcasecmp(args[0], os_func_list[i].cmd)) {
	       os_func_list[i].function(fd, args, nargs);
	       free(args);
	       return;
	  }
     }

     irc_fprintf(fd->sok, NOTICE, s_OperServ, os_userid, os_hostname, fd->nick, "Unknown command.");
     irc_fprintf(fd->sok, NOTICE, s_OperServ, os_userid, os_hostname, fd->nick,
		 "\x02/MSG OperServ HELP\x02 for more information.");
     free(args);
     return;
}

#if 0

struct cs_func {
     char *cmd;
     void (*function) (struct fd_list *, char **, int);
     int need_reg_nick;
};

struct cs_func CS_Functions[] =
{
     {"REGISTER", cs_register, 1},
     {"IDENTIFY", cs_identify, 1},
     {"SET", cs_set, 1},
     {"ACCESS", cs_access, 1},
     {"LEVELS", cs_levels, 1},
     {"AKICK", cs_akick, 1},
     {"DROP", cs_drop, 1},
     {"INFO", cs_info, 0},
     {"LIST", cs_list, 0},
     {"INVITE", cs_invite, 1},
     {"OP", cs_op, 1},
     {"DEOP", cs_deop, 1},
     {"UNBAN", cs_unban, 1},
     {"CLEAR", cs_clear, 1},
     {"HELP", cs_help, 0},
     {NULL, NULL, 0}
};

#endif

struct ms_func {
     char *cmd;
     void (*function) (struct fd_list *, char **, int);
     int need_identify;
};

typedef struct ms_func MS_Functions;

void ms_list(struct fd_list *fd, char **av, int ac)
{
     MS_Memo *l = memo_list;
     NS_User *u = find_user(fd->nick);

     irc_fprintf(fd->sok, ":%s!%s@%s NOTICE %s :Memo list for %s\r\n", s_MemoServ, ms_userid, ms_hostname, fd->nick,
		 fd->nick);
     while (l) {
	  if (l->to == u) {
	       irc_fprintf(fd->sok, ":%s!%s@%s NOTICE %s :%i: From - %s, Sent at - %s\r\n", s_MemoServ, ms_userid,
			   ms_hostname, fd->nick, l->memo_id, l->from->nick, my_ctime(l->sent_at));
	  }
	  l = l->next;
     }
     irc_fprintf(fd->sok, ":%s!%s@%s NOTICE %s :End of memo list\r\n", s_MemoServ, ms_userid, ms_hostname, fd->nick);
     return;
}

extern time_t __time(time_t *);

void ms_send(struct fd_list *fd, char **av, int ac)
{
     if (ac < 3) {
	  irc_fprintf(fd->sok, NOTICE, s_MemoServ, ms_userid, ms_hostname, fd->nick, "\x02SEND\x02 <nick> <text>");
	  irc_fprintf(fd->sok, NOTICE, s_MemoServ, ms_userid, ms_hostname, fd->nick,
		      "\x02/msg MemoServ HELP SEND\x02 for more information.");
	  return;
     } else {
	  MS_Memo *m, *l;
	  NS_User *from, *to;

	  from = find_user(fd->nick);
	  to = find_user(av[1]);
	  if (!to) {
	       irc_fprintf(fd->sok, "%s!%s@%s NOTICE %s :\x02%s\x02 is not a registered nickname!\r\n", s_MemoServ,
			   ms_userid, ms_hostname, fd->nick, av[1]);
	       return;
	  } else {
	       fprintf(stderr, "filling in memo struct..\n");
	       m = malloc(sizeof(MS_Memo));
	       m->from = from;
	       m->from_user_num = from->user_num;
	       m->to = to;
	       m->to_user_num = to->user_num;
	       m->sent_at = __time(NULL);
	       m->memo_id = ++last_memo_id;
	       snprintf(m->memo, 255, "%s", av[2]);
	       fprintf(stderr, "attaching..\n");
	       l = memo_list;
	       fprintf(stderr, "l = %p\n", l);
	       while (l->next) {
		    fprintf(stderr, "l = %p\n", l);
		    l = l->next;
	       }
	       l->next = m;
	       m->prev = l;
	       irc_fprintf(fd->sok, NOTICE, s_MemoServ, ms_userid, ms_hostname, fd->nick, "Memo sent!");
	       {
		    struct fd_list *l = __head->next;

		    while (l) {
			 if (!strcasecmp(l->nick, to->nick)) {
			      irc_fprintf(l->sok, ":%s!%s@%s NOTICE %s :A memo has just arrived for you!\r\n",
					  s_MemoServ, ms_userid, ms_hostname, fd->nick);
			      irc_fprintf(l->sok, ":%s!%s@%s NOTICE %s :Type \x02/msg MemoServ LIST\x02 for a list.\r\n",
					  s_MemoServ, ms_userid, ms_hostname, fd->nick);
			      l = NULL;
			 }
			 if (l)
			      l = l->next;
		    }
	       }
	       return;
	  }
     }
}

void ms_read(struct fd_list *fd, char **av, int ac)
{
     NS_User *to = find_user(fd->nick);
     MS_Memo *l = memo_list->next;
     int memo_number;

     if (ac < 2) {
	  irc_fprintf(fd->sok, NOTICE, s_MemoServ, ms_userid, ms_hostname, fd->nick, "Not enough parameters!");
	  irc_fprintf(fd->sok, NOTICE, s_MemoServ, ms_userid, ms_hostname, fd->nick,
		      "\x02/msg MemoServ HELP READ\x02 for more information.");
	  return;
     } else {
	  memo_number = atoi(av[1]);
	  while (l) {
	       if (l->memo_id == memo_number && l->to == to) {
		    irc_fprintf(fd->sok, ":%s!%s@%s NOTICE %s :Memo for %s\r\n", s_MemoServ, ms_userid, ms_hostname,
				fd->nick, fd->nick);
		    irc_fprintf(fd->sok, ":%s!%s@%s NOTICE %s :From %s\r\n", s_MemoServ, ms_userid, ms_hostname,
				fd->nick, l->from->nick);
		    irc_fprintf(fd->sok, ":%s!%s@%s NOTICE %s :Sent at: %s\r\n", s_MemoServ, ms_userid, ms_hostname,
				fd->nick, my_ctime(l->sent_at));
		    irc_fprintf(fd->sok, ":%s!%s@%s NOTICE %s :%s\r\n", s_MemoServ, ms_userid, ms_hostname,
				fd->nick, l->memo);
	       }
	       l = l->next;
	  }
	  irc_fprintf(fd->sok, ":%s!%s@%s NOTICE %s :End of memos!\r\n", s_MemoServ, ms_userid, ms_hostname, fd->nick);
	  return;
     }
}

void ms_erase(struct fd_list *fd, char **av, int ac)
{
     if (ac < 2) {
	  irc_fprintf(fd->sok, NOTICE, s_MemoServ, ms_userid, ms_hostname, fd->nick, "Not enough parameters!");
	  irc_fprintf(fd->sok, NOTICE, s_MemoServ, ms_userid, ms_hostname, fd->nick,
		      "\x02/msg MemoServ HELP ERASE\x02 for more information.");
     } else {
	  int memo_number = atoi(av[1]);
	  MS_Memo *l = memo_list->next;
	  NS_User *u = find_user(fd->nick);

	  while (l) {
	       if (l->to == u && l->memo_id == memo_number) {
		    if (l->next) {
			 l->prev->next = l->next;
			 l->next->prev = l->prev;
		    } else {
			 l->prev->next = NULL;
		    }
		    free(l);
		    irc_fprintf(fd->sok, NOTICE, s_MemoServ, ms_userid, ms_hostname, fd->nick,
				"Memo deleted!");
		    return;
	       }
	       l = l->next;
	  }
	  irc_fprintf(fd->sok, NOTICE, s_MemoServ, ms_userid, ms_hostname, fd->nick,
		      "No such memo!");
	  return;
     }
}

struct l {
     char *buf;
};

struct m {
     char *buf;
};

struct ms_help {
     char *cmd;
     struct m *helptext;
};

static struct m list_helptext[] =
{
     {"HELP on \x02LIST\x02"},
     {" "},
     {"\x02LIST\x02"},
     {" "},
     {"Returns a listing of messages waiting for you, with"},
     {"ID numbers and the user it is from."},
     {NULL}
};

static struct m send_helptext[] =
{
     {"HELP on \x02SEND\x02"},
     {" "},
     {"\x02SEND\x02 <nick> <message>"},
     {" "},
     {"Sends a memo to the user specified in <nick>."},
     {"The user must have registered their name in NickServ"},
     {"in order to receive messages this way."},
     {NULL}
};

static struct m read_helptext[] =
{
     {"HELP on \x02READ\x02"},
     {" "},
     {"\x02READ\x02 <msgid>"},
     {" "},
     {"Reads the message specified, with the ID from \x02LIST\x02."},
     {NULL}
};

static struct m erase_helptext[] =
{
     {"HELP on \x02ERASE\x02"},
     {" "},
     {"\x02ERASE\x02 <msgid>"},
     {" "},
     {"Erases the message specified, with the ID from \x02LIST\x02."},
     {NULL}
};

struct ms_help ms_help_info[] =
{
     {"LIST", list_helptext},
     {"SEND", send_helptext},
     {"READ", read_helptext},
     {"ERASE", erase_helptext},
     {NULL, NULL}
};

struct l ms_helptext[] =
{
     {"\x02MemoServ\x02 Help"},
     {" "},
     {"MemoServ provides services for sending memos between"},
     {"users on IRC at all times, including when the user"},
     {"is not online."},
     {" "},
     {"Commands:"},
     {"\x02SEND\x02  - Sends a message to a user"},
     {"\x02LIST\x02  - Lists messages waiting for you"},
     {"\x02READ\x02  - Reads a message"},
     {"\x02" "ERASE\x02 - Erases a message"},
     {" "},
     {"/msg MemoServ HELP <cmd> for more help."},
     {NULL}
};

void ms_help(struct fd_list *fd, char **av, int ac)
{
     struct ms_help *h = ms_help_info;

     if (ac < 2) {
	  struct l *q = ms_helptext;

	  while (q->buf) {
	       irc_fprintf(fd->sok, NOTICE, s_MemoServ, ms_userid, ms_hostname, fd->nick, q->buf);
	       q++;
	  }
	  return;
     }
     while (h->cmd) {
	  if (!strcasecmp(h->cmd, av[1])) {
	       struct m *q = h->helptext;

	       while (q->buf) {
		    irc_fprintf(fd->sok, NOTICE, s_MemoServ, ms_userid, ms_hostname, fd->nick, q->buf);
		    q++;
	       }
	       return;
	  }
	  h++;
     }
     irc_fprintf(fd->sok, NOTICE, s_NickServ, ms_userid, ms_hostname, fd->nick, "No such command!");
     return;
}

MS_Functions ms_func_list[] =
{
     {"LIST", ms_list, 1},
     {"SEND", ms_send, 1},
     {"READ", ms_read, 1},
     {"ERASE", ms_erase, 1},
     {"HELP", ms_help, 0},
     {NULL, NULL, 0}
};

int ms_num_funcs = 5;

void memoserv_msg_handle(struct fd_list *fd, char **av, int ac)
{
     char **args, *input, *tmp;
     int i, nargs;

     if (memo_list == NULL) {
	  memo_list = malloc(sizeof(MS_Memo));
	  bzero(memo_list, sizeof(MS_Memo));
     }
     if (ac > 2) {
	  input = av[2];
	  tmp = input;
	  args = NULL;
	  nargs = 0;
	  while (*input != 0) {
	       if (*input == ' ' && nargs < 2) {
		    args = (char **) realloc(args, sizeof(char **) * ++nargs);

		    args[nargs - 1] = tmp;
		    *input = 0;
		    input++;
		    tmp = input;
	       } else {
		    input++;
	       }
	  }
	  args = (char **) realloc(args, sizeof(char **) * ++nargs);

	  args[nargs - 1] = tmp;
	  args = (char **) realloc(args, sizeof(char **) * nargs + 1);

	  args[nargs] = NULL;
     } else {
	  irc_fprintf(fd->sok, NOTICE, s_MemoServ, ms_userid, ms_hostname, fd->nick, "I need a command to parse.");
	  irc_fprintf(fd->sok, NOTICE, s_MemoServ, ms_userid, ms_hostname, fd->nick,
		      "\x02/MSG OperServ HELP\x02 for more information.");
	  return;
     }

     for (i = 0; i < ms_num_funcs; i++) {
	  if (!strcasecmp(args[0], ms_func_list[i].cmd)) {
	       if (ms_func_list[i].need_identify) {
		    if (fd->registered & NS_IS_IDENTIFIED) {
			 ms_func_list[i].function(fd, args, nargs);
			 free(args);
			 return;
		    } else {
			 irc_fprintf(fd->sok, NOTICE, s_MemoServ, ms_userid, ms_hostname, fd->nick,
				     "You must identify in NickServ before using this command!");
			 irc_fprintf(fd->sok, NOTICE, s_MemoServ, ms_userid, ms_hostname, fd->nick,
				     "Try \x02/msg NickServ IDENTIFY <pass>\x02 first!");
			 free(args);
			 return;
		    }
	       } else {
		    ms_func_list[i].function(fd, args, nargs);
		    free(args);
		    return;
	       }
	  }
     }

     irc_fprintf(fd->sok, NOTICE, s_MemoServ, ms_userid, ms_hostname, fd->nick, "Unknown command.");
     irc_fprintf(fd->sok, NOTICE, s_MemoServ, ms_userid, ms_hostname, fd->nick,
		 "\x02/MSG MemoServ HELP\x02 for more information.");
     free(args);
     return;
}

extern int match(char *, char *);

int version(struct fd_list *fd, char **av, int ac)
{
     if (ac > 1) {
	  if (!match(av[1], "services.net")) {
	       irc_fprintf(fd->sok, RPL_VERSION, "services.net", fd->nick, "0.4.", "services.net", "IRC Services 0.4 for ircd 0.3.6");
	       return 0;
	  } else {
	       return 1;
	  }
     } else {
	  return 1;
     }
}

int privmsg(struct fd_list *fd, char **av, int ac)
{
     if (ac > 1) {
	  if (!strcasecmp(av[1], s_NickServ)) {
	       nickserv_msg_handle(fd, av, ac);
	       return 0;
	  } else if (!strcasecmp(av[1], s_OperServ)) {
	       operserv_msg_handle(fd, av, ac);
	       return 0;
	  } else if (!strcasecmp(av[1], s_MemoServ)) {
	       memoserv_msg_handle(fd, av, ac);
	       return 0;
	  } else {
	       return 1;
	  }
     } else {
	  return 1;
     }
     return 0;
}

int nick(struct fd_list *fd, char **av, int ac)
{
     NS_User *user = find_user(av[1]);

     if (user == NULL) {
	  return 1;
     } else {
	  m_nick(fd, av, ac);
	  fd->registered &= ~NS_IS_IDENTIFIED;
	  if (user->secure || user->kill) {
	       irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, "This nickname is owned by someone else.");
	       irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, "If this is your nickname, please identify - \002/msg NickServ IDENTIFY [pass]\002");
	       if (user->kill) {
		    irc_fprintf(fd->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, fd->nick, "If you do not comply within 60 seconds, you will be disconnected.");
	       }
	  }
	  return 0;
     }
}

int nickserv(struct fd_list *fd, char **av, int ac)
{
     char **new_av;
     int new_ac;
     char *arg0 = "PRIVMSG", *arg1 = "NickServ", *arg2;
     int need_free = 0;

     if (ac > 2) {
	  char tmp[255];
	  int n;

	  bzero(tmp, 255);
	  for (n = 1; n < ac; n++) {
	       snprintf(tmp, 255, "%s %s", tmp, av[n]);
	  }
	  arg2 = (char *) malloc(strlen(tmp));
	  strncpy(arg2, tmp, strlen(tmp));
	  need_free = 1;
     } else {
	  arg2 = av[1];
	  need_free = 0;
     }
     new_av = (char **) malloc(sizeof(char **) * 4);

     new_av[0] = arg0;
     new_av[1] = arg1;
     new_av[2] = arg2;
     new_av[3] = NULL;

     new_ac = 3;
     privmsg(fd, new_av, new_ac);

     if (need_free)
	  free(arg2);
     free(new_av);

     return 0;
}

void nickserv_whois_reply(struct fd_list *fd)
{
     irc_fprintf(fd->sok, RPL_WHOISUSER, server_name, fd->nick, s_NickServ, ns_userid, ns_hostname, ns_realname);
     irc_fprintf(fd->sok, RPL_WHOISSERVER, server_name, fd->nick, s_NickServ, ns_server, ns_server_desc);
     irc_fprintf(fd->sok, RPL_WHOISOPERATOR, server_name, fd->nick, s_NickServ);
     irc_fprintf(fd->sok, RPL_ENDOFWHOIS, server_name, fd->nick, s_NickServ);
     return;
}

void operserv_whois_reply(struct fd_list *fd)
{
     irc_fprintf(fd->sok, RPL_WHOISUSER, server_name, fd->nick, s_OperServ, os_userid, os_hostname, os_realname);
     irc_fprintf(fd->sok, RPL_WHOISSERVER, server_name, fd->nick, s_OperServ, os_server, os_server_desc);
     irc_fprintf(fd->sok, RPL_WHOISOPERATOR, server_name, fd->nick, s_OperServ);
     irc_fprintf(fd->sok, RPL_ENDOFWHOIS, server_name, fd->nick, s_OperServ);
}

void memoserv_whois_reply(struct fd_list *fd)
{
     irc_fprintf(fd->sok, RPL_WHOISUSER, server_name, fd->nick, s_MemoServ, ms_userid, ms_hostname, ms_realname);
     irc_fprintf(fd->sok, RPL_WHOISSERVER, server_name, fd->nick, s_MemoServ, ms_server, ms_server_desc);
     irc_fprintf(fd->sok, RPL_WHOISOPERATOR, server_name, fd->nick, s_MemoServ);
     irc_fprintf(fd->sok, RPL_ENDOFWHOIS, server_name, fd->nick, s_MemoServ);
}

int whois(struct fd_list *fd, char **av, int ac)
{
     if (ac < 2) {
	  __not_enough_parameters(fd, av[0]);
     } else {
	  if (ac == 2) {
	       if (!strcasecmp(av[1], s_NickServ)) {
		    nickserv_whois_reply(fd);
	       } else if (!strcasecmp(av[1], s_OperServ)) {
		    operserv_whois_reply(fd);
	       } else if (!strcasecmp(av[1], s_MemoServ)) {
		    memoserv_whois_reply(fd);
	       } else {
		    return 1;
	       }
	  } else {
	       if (!strcasecmp(av[2], s_NickServ)) {
		    nickserv_whois_reply(fd);
	       } else if (!strcasecmp(av[2], s_OperServ)) {
		    operserv_whois_reply(fd);
	       } else if (!strcasecmp(av[2], s_MemoServ)) {
		    memoserv_whois_reply(fd);
	       } else {
		    return 1;
	       }
	  }
     }
     return 0;
}

int links(struct fd_list *fd, char **av, int ac)
{
     irc_fprintf(fd->sok, RPL_LINKS, server_name, fd->nick, server_name, server_name, 0, server_desc);
     irc_fprintf(fd->sok, RPL_LINKS, server_name, fd->nick, ns_server, server_name, 1, ns_server_desc);
     irc_fprintf(fd->sok, RPL_ENDOFLINKS, server_name, fd->nick, "*");
     return 0;
}

void nickserv_who(struct fd_list *fd)
{
     irc_fprintf(fd->sok, RPL_WHOREPLY, server_name, fd->nick, "*", ns_userid, ns_hostname, ns_server, s_NickServ, 'H', "*", 1,
		 ns_realname);
     irc_fprintf(fd->sok, RPL_ENDOFWHO, server_name, fd->nick, s_NickServ);
     return;
}

int who(struct fd_list *fd, char **av, int ac)
{
     if (ac > 1) {
	  if (av[1][0] == '#') {
	       __who_channel(fd, av[1]);
	  } else {
	       if (!strcasecmp(av[1], s_NickServ)) {
		    nickserv_who(fd);
	       } else {
		    __who_user(fd, av[1]);
	       }
	  }
     } else {
	  __not_enough_parameters(fd, av[0]);
     }
     return 0;
}

void sigprof(int signum)
{
     struct fd_list *u = __head->next;
     NS_User *q;

     while (u) {
	  if ((q = find_user(u->nick)) != NULL) {
	       if (!(u->registered & NS_IS_IDENTIFIED)) {
		    if (q->kill) {
			 struct fd_list tmp;
			 char *argv[4];

			 argv[0] = "KILL";
			 argv[1] = u->nick;
			 argv[2] = "Never identified!";
			 argv[3] = NULL;
			 bzero(&tmp, sizeof(tmp));
			 snprintf(tmp.nick, 15, "%s", s_NickServ);
			 snprintf(tmp.userid, 15, "%s", ns_userid);
			 snprintf(tmp.hostname, 50, "%s", ns_hostname);
			 m_kill(&tmp, argv, 3);
		    }
	       }
	  }
	  u = u->next;
     }

     signal(SIGPROF, sigprof);
}

void _services_init_signals()
{
     struct itimerval val;
     struct timeval tv;

     tv.tv_sec = 60;
     tv.tv_usec = 0;
     val.it_interval = tv;
     val.it_value = tv;
     setitimer(ITIMER_PROF, &val, NULL);
     signal(SIGPROF, sigprof);
}

void load_dbs();
void save_dbs();

void *
  ns_kill_thread(void *arg)
{
     pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);
     pthread_setcanceltype(PTHREAD_CANCEL_DEFERRED, NULL);
     while (1) {
	  NS_User *u;
	  struct fd_list *l;

	  pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, NULL);
	  sleep(60);
	  pthread_setcanceltype(PTHREAD_CANCEL_DEFERRED, NULL);
	  l = __head->next;
	  while (l) {
	       if ((u = find_user(l->nick)) != NULL) {
		    if (u->kill) {
			 if (!(l->registered & NS_IS_IDENTIFIED)) {
			      struct fd_list q;
			      char *av[3];

			      snprintf(q.nick, 15, "%s", s_NickServ);
			      snprintf(q.userid, 15, "%s", ns_userid);
			      snprintf(q.hostname, 15, "%s", ns_hostname);
			      av[0] = "KILL";
			      av[1] = l->nick;
			      av[2] = "Never identified!";
			      m_kill(&q, av, 3);
			 }
		    }
	       }
	       l = l->next;
	  }
	  pthread_testcancel();
     }
}

void *
  ms_notify_thread(void *arg)
{
     pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);
     pthread_setcanceltype(PTHREAD_CANCEL_DEFERRED, NULL);
     while (1) {
	  pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, NULL);
	  sleep(600);		/* 10 min */
	  pthread_setcanceltype(PTHREAD_CANCEL_DEFERRED, NULL);
	  {
	       NS_User *u;
	       struct fd_list *l = __head->next;
	       MS_Memo *m;

	       while (l) {
		    if ((u = find_user(l->nick)) != NULL) {
			 int n_memos = 0;

			 m = memo_list;
			 while (m) {
			      if (m->to == u) {
				   n_memos++;
			      }
			      m = m->next;
			 }
			 if (n_memos) {
			      irc_fprintf(l->sok, ":%s!%s@%s NOTICE %s :You have %i memos waiting.\r\n",
					  s_MemoServ, ms_userid, ms_hostname, l->nick, n_memos);
			      irc_fprintf(l->sok, ":%s!%s@%s NOTICE %s :Type \x02/msg MemoServ LIST\x02 for a list.\r\n",
					  s_MemoServ, ms_userid, ms_hostname, l->nick);
			 }
		    }
		    l = l->next;
	       }
	  }
	  pthread_testcancel();
     }
}

void *
  db_save_thread(void *arg)
{
     pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);
     pthread_setcanceltype(PTHREAD_CANCEL_DEFERRED, NULL);
     while (1) {
	  pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, NULL);
	  sleep(3600);		/* 1 hour */
	  pthread_setcanceltype(PTHREAD_CANCEL_DEFERRED, NULL);
	  save_dbs();
	  pthread_testcancel();
     }
}

pthread_t db_thr;
pthread_t ms_thr;
pthread_t ns_thr;

void _init()
{
     if (__head != NULL) {
	  struct fd_list *list = __head->next;

	  while (list != NULL) {
	       if (list->registered)
		    list->registered = 1;
	       list = list->next;
	  }
     }
     memo_list = malloc(sizeof(MS_Memo));
     bzero(memo_list, sizeof(MS_Memo));

     load_dbs();

     started = __time(NULL);

     if (__head != NULL) {
	  struct fd_list *l = __head->next;

	  while (l) {
	       NS_User *u;

	       if ((u = find_user(l->nick)) != NULL) {
		    if (u->secure || u->kill) {
			 irc_fprintf(l->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, l->nick,
				     "This nickname is registered.");
			 irc_fprintf(l->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, l->nick,
				     "If this is your nickname, type \x02/msg NickServ IDENTIFY [pass]\x02 now.");
		    }
		    if (u->kill) {
			 irc_fprintf(l->sok, NOTICE, s_NickServ, ns_userid, ns_hostname, l->nick,
			     "If you do not comply within 60 seconds, you will be disconnected.");
		    }
	       }
	       l = l->next;
	  }
     }
     pthread_create(&db_thr, NULL, db_save_thread, NULL);
     pthread_create(&ms_thr, NULL, ms_notify_thread, NULL);
     pthread_create(&ns_thr, NULL, ns_kill_thread, NULL);
     _services_init_signals();
}

void _fini()
{
     pthread_cancel(db_thr);
     pthread_cancel(ms_thr);
     pthread_cancel(ns_thr);
     save_dbs();
     {
	  NS_User *list, *p;

	  list = registered_nicks;
	  while (list != NULL) {
	       p = list->next;
	       free(list);
	       list = p;
	  }
     }
     {
	  NS_AccessList *list, *p;

	  list = access_list;
	  while (list != NULL) {
	       p = list->next;
	       free(list);
	       list = p;
	  }
     }
     {
	  NS_LinkList *list, *p;

	  list = link_list;
	  while (list != NULL) {
	       p = list->next;
	       free(list);
	       list = p;
	  }
     }

}

void save_dbs()
{
     NS_User u, *ul;
     NS_AccessList a, *al;
     NS_LinkList l, *ll;
     MS_Memo m, *ml;
     FILE *db = fopen("nickserv.users", "w");

     fprintf(stderr, "saving nickserv.users..\n");
     fprintf(stderr, "registered_nicks = %p, registered_nicks->next = %p\n", registered_nicks, registered_nicks->next);
     ul = registered_nicks->next;
     while (ul) {
	  fprintf(stderr, "ul = %p\n", ul);
	  bcopy(ul, &u, sizeof(NS_User));
	  u.prev = u.next = NULL;
	  fprintf(stderr, "writing user id %i\n", u.user_num);
	  fwrite((void *) &u, sizeof(NS_User), 1, db);
	  ul = ul->next;
     }
     fclose(db);
     db = fopen("nickserv.access", "w");

     al = access_list->next;
     while (al) {
	  bcopy(al, &a, sizeof(NS_AccessList));
	  a.prev = a.next = NULL;
	  a.user = (NS_User *) NULL;
	  fwrite((void *) &a, sizeof(NS_AccessList), 1, db);
	  al = al->next;
     }
     fclose(db);

     db = fopen("nickserv.links", "w");
     ll = link_list->next;
     while (ll) {
	  bcopy(ll, &l, sizeof(NS_LinkList));
	  l.prev = l.next = NULL;
	  l.user = (NS_User *) NULL;
	  fwrite((void *) &l, sizeof(NS_LinkList), 1, db);
	  ll = ll->next;
     }
     fclose(db);

     db = fopen("memoserv.memos", "w");
     ml = memo_list->next;
     while (ml) {
	  bcopy(ml, &m, sizeof(MS_Memo));
	  m.to = m.from = NULL;
	  m.prev = m.next = NULL;
	  fwrite((void *) &m, sizeof(MS_Memo), 1, db);
	  ml = ml->next;
     }
     fclose(db);

}

void load_dbs()
{
     NS_User u, *ul;
     NS_AccessList a, *al;
     NS_LinkList l, *ll;
     MS_Memo m, *ml;

     FILE *db = fopen("nickserv.users", "r");

     if (!db) {
	  /* nothing */
	  registered_nicks = malloc(sizeof(NS_User));
	  bzero(registered_nicks, sizeof(NS_User));
     } else {
	  ul = malloc(sizeof(NS_User));
	  bzero(ul, sizeof(NS_User));

	  while (!feof(db)) {
	       int l = fread((void *) &u, sizeof(NS_User), 1, db);

	       if (l == 0) {
		    continue;
	       } else {
		    if (!ul) {
			 ul = malloc(sizeof(NS_User));
			 bzero(ul, sizeof(NS_User));
			 bcopy(&u, ul, sizeof(NS_User));
		    } else {
			 NS_User *q = ul;
			 NS_User *r = malloc(sizeof(NS_User));

			 while (q->next)
			      q = q->next;
			 bzero(r, sizeof(NS_User));
			 bcopy(&u, r, sizeof(NS_User));
			 r->prev = q;
			 q->next = r;
		    }
	       }
	  };
	  fclose(db);
	  registered_nicks = ul;
     }

     db = fopen("nickserv.access", "r");
     if (!db) {
	  /* nothing */
	  access_list = malloc(sizeof(NS_AccessList));
	  bzero(access_list, sizeof(NS_AccessList));
     } else {
	  al = malloc(sizeof(NS_AccessList));
	  bzero(al, sizeof(NS_AccessList));
	  while (!feof(db)) {
	       int l = fread((void *) &a, sizeof(NS_AccessList), 1, db);

	       if (l == 0)
		    continue;
	       else {
		    if (!al) {
			 al = malloc(sizeof(NS_AccessList));
			 bzero(al, sizeof(NS_AccessList));
			 bcopy(&a, al, sizeof(NS_AccessList));
		    } else {
			 NS_AccessList *q = al;
			 NS_AccessList *r = malloc(sizeof(NS_AccessList));

			 while (q->next)
			      q = q->next;
			 bzero(r, sizeof(NS_AccessList));
			 bcopy(&a, r, sizeof(NS_AccessList));
			 r->prev = q;
			 q->next = r;
		    }
	       }
	  }
	  fclose(db);
	  access_list = al;
     }

     db = fopen("nickserv.links", "r");
     if (!db) {
	  /* nothing */
	  link_list = malloc(sizeof(NS_LinkList));
	  bzero(link_list, sizeof(NS_LinkList));
     } else {
	  ll = malloc(sizeof(NS_LinkList));
	  bzero(ll, sizeof(NS_LinkList));
	  while (!feof(db)) {
	       int q = fread((void *) &l, sizeof(NS_LinkList), 1, db);

	       if (q == 0)
		    continue;
	       else {
		    if (!ll) {
			 ll = malloc(sizeof(NS_LinkList));
			 bzero(ll, sizeof(NS_LinkList));
			 bcopy(&l, ll, sizeof(NS_LinkList));
		    } else {
			 NS_LinkList *q = ll;
			 NS_LinkList *r = malloc(sizeof(NS_LinkList));

			 while (q->next)
			      q = q->next;
			 bzero(r, sizeof(NS_LinkList));
			 bcopy(&l, r, sizeof(NS_LinkList));
			 r->prev = q;
			 q->next = r;
		    }
	       }
	  }
	  fclose(db);
	  link_list = ll;
     }

     db = fopen("memoserv.memos", "r");
     if (!db) {
	  /* nothing */
	  ml = malloc(sizeof(MS_Memo));
	  bzero(ml, sizeof(MS_Memo));
	  memo_list = ml;
     } else {
	  ml = malloc(sizeof(MS_Memo));
	  bzero(ml, sizeof(MS_Memo));
	  while (!feof(db)) {
	       int l = fread((void *) &m, sizeof(MS_Memo), 1, db);

	       if (l == 0)
		    continue;
	       else {
		    MS_Memo *q = ml;
		    MS_Memo *r = malloc(sizeof(MS_Memo));

		    while (q->next)
			 q = q->next;
		    bzero(r, sizeof(MS_Memo));
		    bcopy(&m, r, sizeof(MS_Memo));
		    r->prev = q;
		    q->next = r;
	       }
	  }
	  fclose(db);
	  memo_list = ml;
     }

     {
	  NS_AccessList *al = access_list;

	  while (al) {
	       NS_User *ul = registered_nicks;

	       while (ul) {
		    if (ul->user_num == al->user_num) {
			 al->user = ul;
			 ul = NULL;
		    }
		    if (ul)
			 ul = ul->next;
	       }
	       al = al->next;
	  }
     }

     {
	  NS_LinkList *ll = link_list;

	  while (ll) {
	       NS_User *ul = registered_nicks;

	       while (ul) {
		    if (ul->user_num == ll->user_num) {
			 ll->user = ul;
			 ul = NULL;
		    }
		    if (ul)
			 ul = ul->next;
	       }
	       ll = ll->next;
	  }
     }

     {
	  MS_Memo *ml = memo_list;

	  while (ml) {
	       NS_User *ul = registered_nicks;

	       while (ul) {
		    if (ul->user_num == ml->to_user_num) {
			 ml->to = ul;
			 ul = NULL;
		    }
		    if (ul)
			 ul = ul->next;
	       }
	       ul = registered_nicks;
	       while (ul) {
		    if (ul->user_num == ml->from_user_num) {
			 ml->from = ul;
			 ul = NULL;
		    }
		    if (ul)
			 ul = ul->next;
	       }
	       ml = ml->next;
	  }
     }

     {
	  NS_User *ul = registered_nicks;

	  while (ul) {
	       if (ul->user_num > id) {
		    id = ul->user_num;
	       }
	       ul = ul->next;
	  }
     }

     {
	  MS_Memo *ml = memo_list;

	  while (ml) {
	       if (ml->memo_id > last_memo_id) {
		    last_memo_id = ml->memo_id;
	       }
	       ml = ml->next;
	  }
     }
}
