#include "dl.h"
#include "connect.h"
#include <stdarg.h>

int socks_intercept_m_user (struct fd_list *fd, char **av, int ac);

Functions socks_mod_functions[] =
{
  {"USER", socks_intercept_m_user, 0, 1, 0},
  {NULL, NULL, 0, 0, 0}
};

Module_Info socks_mod_info[] =
{
  {
    "socks",
    "socks proxy detection for ircd-0.3.4",
    "0.1"
  }};

Functions *
__module_get_functions ()
{
  return socks_mod_functions;
}

Module_Info *
__module_get_info ()
{
  return socks_mod_info;
}

extern int errno;

static char *
my_inet_ntoa (int addr)
{
  static char buf[16];
  char *tmp = (char *) &addr;
  bzero (buf, 16);
  snprintf (buf, 16, "%3i.%3i.%3i.%3i", tmp[0], tmp[1], tmp[2], tmp[3]);
  return buf;
}

void 
broadcast_to_opers (char *fmt,...)
{
  struct fd_list *fds = __head->next;
  va_list ap;

  va_start (ap, fmt);

  while (fds)
    {
      if (fds->oper)
	{
	  char buf[255];
	  snprintf (buf, 255, ":%s NOTICE %s :Notice -- %s\r\n", server_name, fds->nick, fmt);
	  vfprintf (fds->sok, buf, ap);
	}
      fds = fds->next;
    }

  va_end (ap);
}

int 
socks_intercept_m_user (struct fd_list *fd, char **av, int ac)
{
  struct sockaddr_in sa;
  int salen;
  int myfd;

  if (fd->nick[0])
    {
      fprintf (fd->sok, ":%s NOTICE %s :Notice -- Checking for socks proxy...\r\n", server_name, fd->nick);
    }
  else
    {
      fprintf (fd->sok, ":%s NOTICE * :Notice -- Checking for socks proxy...\r\n", server_name);
    }
  salen = sizeof (sa);
  if (getpeername (fd->fd, &sa, &salen))
    {
      fd->is_killed = 1;
      return 0;
    }

  myfd = get_connection (my_inet_ntoa (sa.sin_addr.s_addr), 1080);	/* 1080 is the standard socks port */
  if (myfd == -1)
    {
      fprintf (fd->sok, ":%s NOTICE %s :Notice -- No socks proxy found (good)..\r\n", server_name, (fd->nick[0] ? fd->nick : "*"));
    }
  else
    {
      fprintf (fd->sok, ":%s NOTICE %s :Notice -- Socks proxy found (bad).. warning active IRCops\r\n", server_name, fd->nick[0] ? fd->nick : "*");
      broadcast_to_opers ("Warning: Open socks proxy found on %s!", my_inet_ntoa (sa.sin_addr.s_addr));
    }
  close (myfd);
  return 1;
}
