#include <stdio.h>
#include <sys/utsname.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <time.h>
#include <syslog.h>
#include "struct.h"
#include "defines.h"
#include "mem.h"
#include "dl.h"
#include "stats.h"
#include "errors.h"
#include "gline.h"
#include "gc.h"
#include "irc_printf.h"
#include "msg_stats.h"
#include <unistd.h>

#ifdef CRYPTEDOPER
extern char *crypt(const char *, const char *);

#endif

extern struct oline *olines;
extern struct pline *plines;
int highest_users;
extern struct msg_use msgs;
extern char *t, *tmp;
extern struct motd *motd_head;
int my_time_offset;

void m_motd(struct fd_list *, char **, int);
extern int match(char *, char *);

struct l {
     char *line;
};

struct l p[] =
{
     {"IRC--"},
     {"Version " VERSION},
     {" "},
     {"Complete re-write of IRCD from scratch by Chris Vandomelen"},
     {" "},
     {"This program is free software; you can redistribute it and/or"},
     {"modify it under the terms of the GNU General Public License"},
     {"as published by the Free Software Foundataion (www.fsf.org);"},
     {"either Version 2 or (at your option) any later version."},
     {" "},
     {"This program is distributed with the hope that it will be useful,"},
     {"but without ANY WARRANTY; without even the implied warranty of"},
     {"MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the"},
     {"GNU General Public License for more details."},
     {" "},
     {"NOTE: No warranties, either express or implied, are hereby given. All"},
     {"software is supplied as is, without guarantee.  The user assumes all"},
     {"responsibility for damages resulting from the use of these features,"},
     {"including, but not limited to, frustration, disgust, system abends, disk"},
     {"head-crashes, general malfeasance, floods, fires, shark attack, nerve"},
     {"gas, locust infestation, cyclones, hurricanes, tsunamis, local"},
     {"electromagnetic disruptions, hydraulic brake system failure, invasion,"},
     {"hashing collisions, normal wear and tear of friction surfaces, comic"},
     {"radiation, inadvertent destruction of sensitive electronic components,"},
     {"windstorms, the Riders of Nazgul, infuriated chickens, malfunctioning"},
     {"mechanical or electrical sexual devices, premature activation of the"},
     {"distant early warning system, peasant uprisings, halitosis, artillery"},
     {"bombardment, explosions, cave-ins, and/or frogs falling from the sky."},
     {" "},
     {"The entire team:"},
     {" "},
     {"chrisv               Chris Vandomelen        chrisv@b0rked.dhs.org"},
     {" "},
     {"Thanks goes to anyone who has helped test this code, or offered"},
     {"any suggestions or ideas to make it better."},
     {NULL}
};

/* time_t __time(time_t *t)
   returns the internal server time - system time adjusted by my_time_offset
 */

time_t
__time(time_t * t)
{
     time_t cur_time = time(NULL);

     cur_time += my_time_offset;
     if (t)
	  *t = cur_time;
     return cur_time;
}

/* void __lusers_reply(struct fd_list *fd)
   returns the /LUSERS reply to fd
 */

void __lusers_reply(struct fd_list *fd)
{
     {
	  struct fd_list *t;
	  int visible = 0, invisible = 0, opers = 0;

	  t = __head->next;
	  while (t != NULL) {
	       if (t->invisible)
		    invisible++;
	       else
		    visible++;
	       if (t->oper)
		    opers++;
	       t = t->next;
	  }
	  irc_fprintf(fd->sok, RPL_LUSERCLIENT, NAME, fd->nick, visible, invisible, 1);
	  irc_fprintf(fd->sok, RPL_LUSEROPER, NAME, fd->nick, opers);
	  {
	       struct chan_list *t;
	       int n = 0;

	       t = __clist_head->next;
	       while (t != NULL) {
		    n++;
		    t = t->next;
	       }
	       irc_fprintf(fd->sok, RPL_LUSERCHANNELS, NAME, fd->nick, n);
	  }
	  irc_fprintf(fd->sok, RPL_LUSERME, NAME, fd->nick, visible + invisible, 0);
	  if (visible + invisible > highest_users) {
	       highest_users = visible + invisible;
	  }
	  irc_fprintf(fd->sok, RPL_HIGHESTUSERS, NAME, fd->nick, highest_users, highest_users);
     }
     irc_fflush(fd->sok);
     return;
}

/* void __registered_reply(struct fd_list *fd)
   returns a registered client reply to *fd, or closes their connection if they are explicitly denied access
 */

void __registered_reply(struct fd_list *fd)
{
     struct utsname *buf;
     struct gline *q;
     char userhost[200];
     struct iline *i;
     char uhost[200];
     char ip_mask[200];

     bzero(uhost, 200);
     bzero(ip_mask, 200);

     log2("entering __registered_reply: fd = %p\n", fd);
     {
	  FILE *out = fopen("connect.log", "a+");

	  fprintf(out, "connection: %s!%s@%s\n", fd->nick, fd->userid, fd->hostname);
	  fclose(out);
     }

     buf = (struct utsname *) calloc(sizeof(struct utsname));

     log("getting system information");
     uname(buf);

     snprintf(userhost, 200, "%s@%s", fd->userid, fd->hostname);
     log("matching glines..");
     if ((q = find_gline(userhost, 0)) != NULL) {
	  log("gline found, disconnecting user.");
	  fd->is_killed = 1;
	  irc_fprintf(fd->sok, ":%s NOTICE %s :You have been glined: %s\r\n", server_name, fd->nick, q->reason);
	  return;
     }
     log("finished..");

     log("matching I: lines..");
     i = iline_list;
     if (fd->userid[0] == '~') {
	  struct sockaddr_in sa;
	  unsigned char *s;
	  int l = sizeof(struct sockaddr_in);

	  snprintf(uhost, 200, "%s", fd->hostname);
	  getpeername(fd->fd, &sa, &l);
	  s = (char *) &(sa.sin_addr.s_addr);
	  snprintf(ip_mask, 200, "%i.%i.%i.%i", s[0], s[1], s[2], s[3]);
     } else {
	  struct sockaddr_in sa;
	  unsigned char *s;
	  int l = sizeof(struct sockaddr_in);

	  snprintf(uhost, 200, "%s@%s", fd->userid, fd->hostname);
	  getpeername(fd->fd, &sa, &l);
	  s = (char *) &(sa.sin_addr.s_addr);
	  snprintf(ip_mask, 200, "%s@%i.%i.%i.%i", fd->userid, s[0], s[1], s[2], s[3]);
     }

     while (i) {
	  int ok = 1;

	  if (!match(i->hostmask, uhost)) {
	       log2("I: line match found for user, class %i, max connections in class is %i, num connections is %i",
		    (i->class->class, i->class->max_clients,
		     i->class->num_connects));
	       fprintf(stderr, "i->class->num_connects = %i, i->class->max_connects = %i\n", i->class->num_connects, i->class->max_clients);
	       if (i->class->num_connects < i->class->max_clients) {
		    if (i->password[0] != '*') {
#ifdef CRYPTEDOPER
			 char salt[2];

			 salt[0] = i->password[0];
			 salt[1] = i->password[1];
			 if (strcmp(crypt(fd->conn_pass, salt), i->password)) {
#else
			 if (strcmp(fd->conn_pass, i->password)) {
#endif
			      ok = 0;
			 } else {
			      ok = 1;
			 }
		    }
		    if (ok) {
			 log("connection class not full, allowing connection.");
			 gcIncRefcnt(i->class);
			 i->class->num_connects++;
			 i->class->refcount++;
			 fd->class = i->class;
			 i = NULL;
		    } else {
			 irc_fprintf(fd->sok, ":%s 464 * :Password mismatch\r\n", server_name);
			 fd->is_killed = 1;
			 irc_fflush(fd->sok);
			 return;
		    }
	       } else {
		    log("connection class full, disconnecting user..");
		    irc_fprintf(fd->sok, ":%s NOTICE %s :*** Your connection class is full, try again later!\r\n",
				server_name, fd->nick);
		    fd->is_killed = 1;
		    irc_fflush(fd->sok);
		    return;
	       }
	  } else if (!match(i->ip_addr_mask, ip_mask)) {
	       log("I: match found for ip address");
	       log2("I: line match found for user, class %i, max connections in class is %i, num connections is %i",
		    (i->class->class, i->class->max_clients,
		     i->class->num_connects));
	       if (i->password[0] != '*') {
#ifdef CRYPTEDOPER
		    char salt[2];

		    salt[0] = i->password[0];
		    salt[1] = i->password[1];
		    if (!strcmp(i->password, crypt(fd->conn_pass, salt))) {
#else
		    if (!strcmp(i->password, fd->conn_pass)) {
#endif
			 ok = 1;
		    } else {
			 ok = 0;
		    }
	       }
	       if (!ok) {
		    irc_fprintf(fd->sok, ":%s 464 * :Password mismatch\r\n", server_name);
		    fd->is_killed = 1;
		    irc_fflush(fd->sok);
		    return;
	       }
	       if (i->class->num_connects < i->class->max_clients) {
		    log("connection class not full, allowing connection.");
		    i->class->num_connects++;
		    i->class->refcount++;
		    gcIncRefcnt(i->class);
		    fd->class = i->class;
		    i = NULL;
	       } else {
		    log("connection class full, disconnecting user..");
		    irc_fprintf(fd->sok, ":%s NOTICE %s :*** Your connection class is full, try again later!\r\n",
				server_name, fd->nick);
		    fd->is_killed = 1;
		    irc_fflush(fd->sok);
		    return;
	       }
	  } else {
	       i = i->next;
	  }
     }

     if (fd->class == NULL) {
	  log("I: match never found, not allowing connection!");
	  irc_fprintf(fd->sok, ":%s NOTICE %s :*** You are not authorized to use this server!", server_name, fd->nick);
	  fd->is_killed = 1;
	  irc_fflush(fd->sok);
	  return;
     } {
	  struct fd_list *u = __head->next;
	  int my_port, len;
	  struct sockaddr_in sa;

	  len = sizeof(sa);
	  if (!getsockname(fd->fd, (struct sockaddr *) &sa, &len)) {
	       my_port = ntohs(sa.sin_port);
	  } else {
	       my_port = port;
	  }

	  syslog(LOG_INFO, "Client connecting on port %i: %s (%s@%s)", my_port, fd->nick, fd->userid, fd->hostname);
	  while (u) {
	       if (u->show_connect && u->oper) {
		    irc_fprintf(u->sok, ":%s NOTICE %s :*** Notice -- Client connecting on port %i: %s (%s@%s)\r\n",
				server_name, u->nick, my_port, fd->nick, fd->userid, fd->hostname);
	       }
	       u = u->next;
	  }
     }

     log("executing register handlers");
     {
	int done = 0;
	Module *q = module_list;
	while (q) {
		if (q->register_handler) {
			done = q->register_handler(fd);
		}
		q = q->next;
	}
	if (done) return;
     }

     log("sending welcome message");
     irc_fprintf(fd->sok, RPL_WELCOME, NAME, fd->nick, fd->nick);
     irc_fprintf(fd->sok, RPL_HOST, NAME, fd->nick, NAME, buf->nodename, VERSION);
     irc_fprintf(fd->sok, RPL_CREATED, NAME, fd->nick, CREATE_DATE);
     irc_fprintf(fd->sok, RPL_LONGVERSION, NAME, fd->nick, NAME, VERSION, USERMODES, CHANMODES);
     {
	  struct fd_list *t;
	  int visible = 0, invisible = 0, opers = 0;

	  t = __head->next;
	  while (t != NULL) {
	       if (t->invisible)
		    invisible++;
	       else
		    visible++;
	       if (t->oper)
		    opers++;
	       t = t->next;
	  }
	  irc_fprintf(fd->sok, RPL_LUSERCLIENT, NAME, fd->nick, visible, invisible, 1);
	  irc_fprintf(fd->sok, RPL_LUSEROPER, NAME, fd->nick, opers);
	  {
	       struct chan_list *t;
	       int n = 0;

	       t = __clist_head->next;
	       while (t != NULL) {
		    n++;
		    t = t->next;
	       }
	       irc_fprintf(fd->sok, RPL_LUSERCHANNELS, NAME, fd->nick, n);
	  }
	  irc_fprintf(fd->sok, RPL_LUSERME, NAME, fd->nick, visible + invisible, 0);
	  if (highest_users < invisible + visible) {
	       highest_users = invisible + visible;
	  }
	  irc_fprintf(fd->sok, RPL_HIGHESTUSERS, NAME, fd->nick, highest_users, highest_users);
     }
     irc_fflush(fd->sok);
     __free(buf);
     log("completed welcome message, sending motd");
     fd->signon = fd->last_msg = __time(NULL);
     m_motd(fd, NULL, 0);	/* m_motd doesn't care what av and ac are. */
     log("returning");
     return;
}

/* void __version_reply(struct fd_list *fd)
   returns a version reply to the client
 */

void __version_reply(struct fd_list *fd)
{
     char server_defs[10];
     int i;

     bzero(server_defs, 10);

     i = 0;
#if defined(CRYPTEDOPER)
     server_defs[i++] = 'p';
#endif
#if defined(OPERONLYSTATS)
     server_defs[i++] = 'S';
#endif
#if defined(DEBUGMEM)
     server_defs[i++] = 'd';
#endif
#if defined(DEBUGSOCK)
     server_defs[i++] = 'D';
#endif
#if defined(DEBUGMAIN)
     server_defs[i++] = 'M';
#endif
#if defined(CHECKIDENT)
     server_defs[i++] = 'I';
#endif
#if defined(PTHREADIDENT)
     server_defs[i++] = 'P';
#endif
     irc_fprintf(fd->sok, RPL_VERSION, NAME, fd->nick, VERSION, NAME, server_defs);
     irc_fflush(fd->sok);
     return;
}

/* void __info_reply(struct fd_list *fd)
   returns the copyright message and other information to the client
 */

void __info_reply(struct fd_list *fd)
{
     struct l *q = p;

     while (q->line) {
	  irc_fprintf(fd->sok, RPL_INFO, NAME, fd->nick, q->line);
	  q++;
     }

     irc_fprintf(fd->sok, RPL_INFO2, NAME, fd->nick, "Birth date", CREATE_DATE);
     irc_fprintf(fd->sok, RPL_INFO, NAME, fd->nick, " ");

     {
	  Module *modules;

	  modules = module_list->next;

	  if (modules != NULL) {
	       irc_fprintf(fd->sok, RPL_INFOMODHDR, NAME, fd->nick);
	       while (modules != NULL) {
		    irc_fprintf(fd->sok, RPL_INFOMOD, NAME, fd->nick, modules->info->module_name);
		    irc_fprintf(fd->sok, RPL_INFOMOD2, NAME, fd->nick, modules->info->module_description);
		    irc_fprintf(fd->sok, RPL_INFOMOD3, NAME, fd->nick, modules->info->module_version);
		    modules = modules->next;
	       }
	  }
     }
     irc_fflush(fd->sok);
     return;
}

/* void __stats_*_reply(struct fd_list *fd)
   void __stats_reply(struct fd_list *fd, char query, char *server)

   returns the appropriate stats information to the user
 */

void __stats_o_reply(struct fd_list *fd)
{
     struct oline *lines;

     lines = olines;

     while (lines != NULL) {
	  irc_fprintf(fd->sok, RPL_STATSOLINE, NAME, fd->nick, 'O', lines->hostmask, lines->nick);
	  lines = lines->next;
     }
     irc_fflush(fd->sok);
}

void __stats_p_reply(struct fd_list *fd)
{
     struct pline *lines;

     lines = plines;

     while (lines != NULL) {
	  irc_fprintf(fd->sok, RPL_STATSPLINE, NAME, fd->nick, lines->port, lines->hostname);
	  lines = lines->next;
     }
     irc_fflush(fd->sok);
}

void stats_y_reply(struct fd_list *fd)
{
     char *yreply = ":%s 218 %s Y %i %i %i %i %i\r\n";
     struct yline *y;

     y = yline_list;
     while (y) {
	  irc_fprintf(fd->sok, yreply, server_name, fd->nick, y->class, y->ping_freq,
		      y->conn_freq, y->max_clients, y->sendq);
	  y = y->next;
     }
}

void stats_i_reply(struct fd_list *fd)
{
     char *ireply = ":%s 215 %s I %s * %s 0 %i\r\n";
     struct iline *i;

     i = iline_list;
     while (i) {
	  irc_fprintf(fd->sok, ireply, server_name, fd->nick, i->ip_addr_mask,
		      i->hostmask, i->class->class);
	  i = i->next;
     }
}

void __stats_u_reply(struct fd_list *fd)
{
     int d, h, m, s, now, t;

     now = __time(NULL);
     t = now - start_time;

     d = t / 60 / 60 / 24;
     t = t - (d * 60 * 60 * 24);

     h = t / 60 / 60;
     t = t - (h * 60 * 60);

     m = t / 60;
     t = t - (m * 60);

     s = t;

     irc_fprintf(fd->sok, RPL_STATSUREPLY, NAME, fd->nick, d, h, m, s);
     irc_fprintf(fd->sok, RPL_250, NAME, fd->nick, highest_users, highest_users);
}

void __stats_f_reply(struct fd_list *fd)
{
     struct motd *l = motd_head;

     while (l) {
	  irc_fprintf(fd->sok, ":%s 601 %s F %s %s\r\n", server_name, fd->nick, l->hostmask, l->motdfile);
	  l = l->next;
     }
     irc_fprintf(fd->sok, ":%s 601 %s F %s %s\r\n", server_name, fd->nick, "*", "ircd.motd");
}

void __stats_reply(struct fd_list *fd, char query, char *server)
{
#ifdef OPERONLYSTATS
     int ok = 0;

     if (fd->oper)
	  ok = 1;
#else
     int ok = 1;

#endif

     switch (query) {
     case 'c':
	  break;
     case 'f':
	  __stats_f_reply(fd);
	  break;
     case 'g':
	  {
	       char *argv[2];

	       argv[0] = "STATS";
	       argv[1] = "g";
	       m_gline(fd, argv, 2);
	  };
	  break;
     case 'h':
	  break;
     case 'i':
	  stats_i_reply(fd);
	  break;
     case 'k':
	  break;
     case 'l':
	  break;
     case 'm':
	  __stats_m_reply(fd);
	  break;
     case 'n':
	  break;
     case 'o':
	  if (ok) {
	       __stats_o_reply(fd);
	  }
	  break;
     case 'p':
	  if (ok) {
	       __stats_p_reply(fd);
	  }
	  break;
     case 'q':
	  break;
     case 'u':
	  __stats_u_reply(fd);
	  break;
     case 'y':
	  stats_y_reply(fd);
	  break;
     default:
     };

     irc_fprintf(fd->sok, ":%s 219 %s %c :End of /STATS report.\r\n", NAME, fd->nick, query);

     return;
}

/* m_version
   av[0] = "VERSION"
   sends a version reply
 */

void m_version(struct fd_list *fd, char **av, int ac)
{
     msgs.version.use++;
     msgs.version.bytes += strlen(tmp);
     __version_reply(fd);
     return;
}

/* m_info
   av[0] = "INFO"
   sends an info reply
 */

void m_info(struct fd_list *fd, char **av, int ac)
{
     msgs.info.use++;
     msgs.info.bytes += strlen(tmp);
     __info_reply(fd);
     return;
}

/* m_lusers
   av[0] = "LUSERS"
   sends an lusers reply
 */

void m_lusers(struct fd_list *fd, char **av, int ac)
{
     msgs.lusers.use++;
     msgs.lusers.bytes += strlen(tmp);
     __lusers_reply(fd);
     return;
}

/* m_stats
   av[0] = "STATS"
   av[1] = query
   av[2] = server
   sends a stats reply
 */

void m_stats(struct fd_list *fd, char **av, int ac)
{
     msgs.stats.use++;
     msgs.stats.bytes += strlen(tmp);
     if (ac < 2) {
	  __not_enough_parameters(fd, t);
     } else {
	  char *query = av[1];

	  __stats_reply(fd, *query, NULL);
     }

     return;
}

/* m_admin
   av[0] = "ADMIN"
   sends an admin reply
 */

void m_admin(struct fd_list *fd, char **av, int ac)
{
     msgs.admin.use++;
     msgs.admin.bytes += strlen(tmp);
     if (admin1[0] != 0) {
	  irc_fprintf(fd->sok, RPL_ADMINME, NAME, fd->nick, NAME);
	  irc_fprintf(fd->sok, RPL_ADMINLOC1, NAME, fd->nick, admin1);
	  irc_fprintf(fd->sok, RPL_ADMINLOC2, NAME, fd->nick, admin2);
	  irc_fprintf(fd->sok, RPL_ADMINEMAIL, NAME, fd->nick, admin3);
     } else {
	  irc_fprintf(fd->sok, ERR_NOADMININFO, NAME, fd->nick, NAME);
     }
     return;
}

extern void strtolower(char *);

/* int strcaseicmp(const char *, char *, int)
   case-insensitively compares 2 strings from indexes 0 to len, returns 0 if the same, otherwise returns nonzero
 */

int strcaseicmp(const char *str, char *cstr, int len)
{
     int i;
     char buf1[len + 1];
     char buf2[len + 1];

     bzero(buf1, len + 1);
     bzero(buf2, len + 1);
     strncpy(buf1, str, len);
     strncpy(buf2, cstr, len);
     strtolower(buf1);
     strtolower(buf2);
     for (i = 0; i < len; i++) {
	  if (str[i] != cstr[i]) {
	       return cstr[i] - str[i];
	  }
     }
     return 0;
}

/* m_motd
   av[0] = "MOTD"
   displays the server motd
 */

void m_motd(struct fd_list *fd, char **av, int ac)
{
     FILE2 *motd = NULL;
     struct motd *list = motd_head;
     char *motdfile;

#ifdef NODEFAULTMOTD
     if (av == NULL) {
	irc_fprintf(fd->sok, ":%s 375 %s :- %s Message of the day -\r\n", server_name, fd->nick, server_name);
	irc_fprintf(fd->sok, ":%s 372 %s :- Type /MOTD to read the AUP for this server.\r\n", server_name, fd->nick);
	irc_fprintf(fd->sok, ":%s 376 %s :- End of /MOTD command\r\n", server_name, fd->nick);
	return;
     }
#endif

     log("entering m_motd");
     if (list == NULL) {
	  log("no F: lines in conf, opening default motd");
	  motd = irc_fopen("ircd.motd", "r");
	  motdfile = "ircd.motd";
     } else {
	  while (list) {
	       if (!match(list->hostmask, fd->hostname)) {
		    log("F: line match found, opening motd");
		    motd = irc_fopen(list->motdfile, "r");
		    fprintf(stderr, "motdfile = %s\n", list->motdfile);
		    fflush(stderr);
		    motdfile = list->motdfile;
		    list = NULL;
	       }
	       if (list)
		    list = list->next;
	  }
	  if (!motd) {
	       log("no F: line found, opening default");
	       motd = irc_fopen("ircd.motd", "r");
	       motdfile = "ircd.motd";
	  }
     }

     if (!motd) {
	  log("no motd file found");
	  irc_fprintf(fd->sok, ":%s 422 %s :MOTD file is missing\r\n", server_name, fd->nick);
     } else {
	  struct stat sbuf;

	  log("displaying motd");
	  stat(motdfile, &sbuf);
	  irc_fprintf(fd->sok, ":%s 375 %s :- %s Message of the Day -\r\n", server_name, fd->nick, server_name);
	  irc_fprintf(fd->sok, ":%s 372 %s :- %s\r\n", server_name, fd->nick, ctime(&(sbuf.st_mtime)));
	  while (!irc_feof(motd)) {
	       char buf[255];
	       char buf2[512];

	       bzero(buf2, 512);
	       bzero(buf, 255);
	       irc_fgets(buf, 255, motd);
	       buf[strlen(buf) - 1] = 0;
	       if (!strcaseicmp(buf, "%{include", 9)) {
		    FILE2 *inc_file;
		    char filename[strlen(buf)];
		    int i;

		    bzero(filename, strlen(buf));
		    for (i = 10; i < strlen(buf); i++) {
			 if (buf[i] == '}') {
			      i = strlen(buf);
			 } else {
			      filename[i - 10] = buf[i];
			 }
		    }
		    inc_file = irc_fopen(filename, "r");
		    if (inc_file) {
			 char ibuf[255];

			 while (!irc_feof(inc_file)) {
			      bzero(ibuf, 255);
			      irc_fgets(ibuf, 255, inc_file);
			      irc_fprintf(fd->sok, ":%s 372 %s :- %s\r\n",
					  server_name, fd->nick, ibuf);
			 }
			 irc_fclose(inc_file);
		    } else {
			 irc_fprintf(fd->sok,
				     ":%s 372 %s :- Cannot open %{include} file %s!\r\n",
				     server_name, fd->nick, filename);
		    }
	       } else {
		    int i;
		    int m = 0;
		    int j = 0;

		    for (i = 0; i < strlen(buf); i++) {
			 if (buf[i] == '%' && buf[i + 1] == '{') {
			      if (!strcaseicmp(&(buf[i + 2]), "nick", 4)) {
				   int q;

				   for (q = 0; q < strlen(fd->nick); q++) {
					buf2[j++] = fd->nick[q];
				   }
				   i += strlen("%{nick}") - 1;
			      } else if (!strcaseicmp(&(buf[i + 2]), "userhost", 8)) {
				   int q;
				   char buf3[strlen(fd->hostname) +
					     strlen(fd->userid) + 2];

				   sprintf(buf3, "%s@%s", fd->userid,
					   fd->hostname);
				   for (q = 0; q < strlen(buf3); q++) {
					buf2[j++] = buf3[q];
				   }
				   i += strlen("%{userhost}") - 1;
			      } else {
				   int q;

				   q = i;
				   while (buf[q] != '}' && buf[q] != 0) {
					q++;
				   }
				   i = q - 1;
			      }
			 } else {
			      buf2[j++] = buf[i];
			 }
		    }
	       }

	       if (buf2[0] != 0) {
		    irc_fprintf(fd->sok, ":%s 372 %s :- %s\r\n", server_name, fd->nick,
				buf2);
	       }
	  }
	  irc_fprintf(fd->sok, ":%s 376 %s :End of MOTD.\r\n", server_name, fd->nick);
	  irc_fclose(motd);
     }
     return;
}

/* m_links
   av[0] = "LINKS"
   returns the list of links
 */

void m_links(struct fd_list *fd, char **av, int ac)
{
     irc_fprintf(fd->sok, RPL_LINKS, NAME, fd->nick, NAME, NAME, 0, server_desc);
     irc_fprintf(fd->sok, RPL_ENDOFLINKS, NAME, fd->nick, "*");
     irc_fflush(fd->sok);
     return;
}

const char *months[] =
{
     "January",
     "February",
     "March",
     "April",
     "May",
     "June",
     "July",
     "August",
     "September",
     "October",
     "November",
     "December"
};

const char *days[] =
{
     "Sunday",
     "Monday",
     "Tuesday",
     "Wednesday",
     "Thursday",
     "Friday",
     "Saturday"
};

extern char *tzname[2];
extern long int timezone;
extern int daylight;

static char *my_ctime(time_t q)
{
     struct tm *t = localtime(&q);
     char is_past_gmt;
     long int mytz;
     int tzhr;
     int tzmn;
     static char buf[255];

     bzero(buf, 255);
     is_past_gmt = (timezone >= 0 ? 1 : 0);
     mytz = (timezone >= 0 ? timezone : -timezone);
     tzhr = (mytz / 60) / 60;
     tzmn = (mytz - (tzhr * 60 * 60)) / 60;

     snprintf(buf, 255, "%s %s %i %i -- %02i:%02i %c%02i:%02i", days[t->tm_wday], months[t->tm_mon], t->tm_mday,
	      t->tm_year + 1900, t->tm_hour, t->tm_min, (is_past_gmt ? '-' : '+'), tzhr, tzmn);
     return buf;
}

/* m_time
   av[0] = "TIME"
   returns the current time (including any offsets set with settime)
 */

void m_time(struct fd_list *fd, char **av, int ac)
{
     time_t i = time(NULL) + my_time_offset;
     struct tm *t = localtime(&i);
     char *buf;
     int tzhr;
     int tzmn;
     long int mytz;
     char is_past_gmt;

     buf = (char *) malloc(255);
     bzero(buf, 255);

     is_past_gmt = (timezone >= 0 ? 1 : 0);
     mytz = (timezone >= 0 ? timezone : -timezone);
     tzhr = (mytz / 60) / 60;
     tzmn = (mytz - (tzhr * 60 * 60)) / 60;

     snprintf(buf, 255, "%i %i :%s %s %i %i -- %02i:%02i %c%02i:%02i", i, my_time_offset, days[t->tm_wday], months[t->tm_mon], t->tm_mday,
	      t->tm_year + 1900, t->tm_hour, t->tm_min, (is_past_gmt ? '-' : '+'), tzhr, tzmn);
     irc_fprintf(fd->sok, ":%s 391 %s %s %s\r\n", server_name, fd->nick, server_name, buf);
     free(buf);
     irc_fflush(fd->sok);
     return;
}

/* m_settime
   av[0] = "SETTIME"
   av[1] = change
   changes the server time by change seconds
 */

void m_settime(struct fd_list *fd, char **av, int ac)
{
     if (ac >= 2) {
	  int new_time = atoi(av[1]);
	  int sys_time = time(NULL);
	  int cur_time = sys_time + my_time_offset;

	  if (new_time - cur_time > 50000 || new_time - cur_time < -50000) {
	       irc_fprintf(fd->sok, ":%s NOTICE %s :SETTIME: Bad value\r\n", server_name, fd->nick);
	       return;
	  } else {
	       int new_time_offset = new_time - sys_time;
	       int time_difference = new_time - cur_time;
	       int q = (time_difference >= 0 ? time_difference : -time_difference);

	       irc_fprintf(fd->sok, ":%s NOTICE %s :clock is set %i seconds %s\r\n",
		     server_name, fd->nick, q, (q == time_difference ? "forwards" : "backwards"));
	       my_time_offset = new_time_offset;
	       return;
	  }
     } else {
	  __not_enough_parameters(fd, av[0]);
	  return;
     }
}
