#include <string.h>
#include <stdio.h>
#include <unistd.h>
#include <crypt.h>
#include <stdlib.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <arpa/inet.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include "listen.h"
#include <sys/time.h>
#include <fcntl.h>
#include <errno.h>
#include <signal.h>
#include <netdb.h>
#include <sys/utsname.h>
#include <syslog.h>
#include <signal.h>

#define APIEXPORT

#include "stats.h"
#include "log.h"
#include "mem.h"
#include "struct.h"
#include "init.h"
#include "dl.h"
#include "errors.h"
#include "defines.h"
#include "conf.h"
#include "channels.h"
#include "user.h"
#include "msg.h"
#include "ping.h"
#include "server.h"
#include "who.h"
#include "mode.h"
#include "info.h"
#include "gline.h"
#include "gc.h"
#include "irc_printf.h"
#include "msg_stats.h"

#define PORT port

struct msg_use msgs;

struct oline *olines = NULL;
struct pline *plines = NULL;

extern int h_errno;

typedef void (*sighandler_t) (int);

struct commands {
     char *name;
     void (*function) (struct fd_list * fd, char **av, int ac);
     int need_registered;
     int need_unregistered;
     int need_oper;
     char *helptext;
};

typedef struct commands Commands;

extern int flud;

/* m_dl
   av[0] = "LOAD"
   av[1] = module name
   loads a loadable module
 */

void m_dl(struct fd_list *fd, char **av, int ac)
{
     if (ac > 1) {
	  int rval;
	  struct fd_list *list = __head->next;

	  rval = load_module(av[1]);
	  if (!rval) {
	       while (list) {
		    if (list->globops) {
			 irc_fprintf(list->sok, ":%s NOTICE %s :Notice -- %s has just loaded module %s\r\n",
				     NAME, list->nick, fd->nick, av[1]);
		    }
		    list = list->next;
	       }
	  } else {
	       irc_fprintf(fd->sok, ":%s NOTICE %s :Notice -- Module %s cannot be loaded\r\n", NAME, fd->nick,
			   av[1]);
	  }
     }
}

/* m_unload
   av[0] = "UNLOAD"
   av[1] = module name
   unloads a loadable module
 */

void m_unload(struct fd_list *fd, char **av, int ac)
{
     if (ac > 1) {
	  int rval;
	  struct fd_list *list = __head->next;

	  rval = unload_module(av[1]);
	  if (!rval) {
	       while (list) {
		    if (list->globops) {
			 irc_fprintf(list->sok, ":%s NOTICE %s :Notice -- %s has just unloaded module %s\r\n",
				     NAME, list->nick, fd->nick, av[1]);
		    }
		    list = list->next;
	       }
	  } else {
	       irc_fprintf(fd->sok, ":%s NOTICE %s :Notice -- Module %s is not loaded\r\n", NAME, fd->nick, av[1]);
	  }
     }
}

/* m_setflud
   av[0] = "SETFLUD"
   av[1] = new flood rate
   sets the flood rate
 */

void m_setflud(struct fd_list *fd, char **av, int ac)
{
     if (ac > 1) {
	  struct fd_list *list = __head->next;

	  flud = atoi(av[1]);
	  while (list) {
	       if (list->globops) {
		    irc_fprintf(list->sok, ":%s NOTICE %s :Notice -- %s has just changed the flood setting to %i\r\n",
				NAME, list->nick, fd->nick, flud);
	       }
	       list = list->next;
	  }
     }
}

int *quit, *die, *restart;
char *_parse(char *, struct fd_list *, int *, int *, int *);

/* char *parse(char *a, struct fd_list *b, int *c, int *d)
   compatibility function - calls _parse(a, b, c, d, &dummy)
 */

char *parse(char *a, struct fd_list *b, int *c, int *d)
{
     int dummy;

     return _parse(a, b, c, d, &dummy);
}

/* m_kill
   av[0] = "KILL"
   av[1] = nick
   av[2] = reason
   kills a user
 */

void m_kill(struct fd_list *fd, char **av, int ac)
{
     struct fd_list *u = __head->next;
     char *kmsg;

     if (ac == 2) {		/* nick only, no msg */
	  kmsg = fd->nick;
     } else if (ac > 2) {
	  kmsg = av[2];
     } else if (ac < 2) {
	  __not_enough_parameters(fd, av[0]);
	  return;
     }
     while (u) {
	  if (!strcasecmp(u->nick, av[1])) {
	       int nuh;
	       int *my_quit, *my_die;
	       char tmp[255];
	       char path[255];
	       char *my_hostname = (fd->hide_hostname ? "hidden.hurricane.net" : fd->hostname);

		if (u->service) {
			irc_fprintf(fd->sok, ":%s 484 %s %s %s :Cannot kill, kick, or deop channel service\r\n",
				server_name, fd->nick, "KILL", u->nick);
			return;
		}

	       snprintf(path, 255, "%s!%s", my_hostname, fd->nick);
	       if (u == fd) {
		    irc_fprintf(fd->sok, ":%s 510 %s :You can't kill yourself!\r\n", server_name, fd->nick);
		    return;
	       } 

		{
		    struct fd_list *q;

		    q = __head->next;
		    while (q) {
			 if (q->oper && q->s_msg && (q != u)) {
			      irc_fprintf(q->sok, ":%s NOTICE %s :*** Notice -- Received KILL message for %s. From %s Path: %s (%s)\r\n",
					  server_name, q->nick, u->nick, fd->nick, path, kmsg);
			 }
			 q = q->next;
		    }
	       }

	       irc_fprintf(u->sok, ":%s!%s@%s KILL %s :%s (%s)\r\n", fd->nick, fd->userid, my_hostname, u->nick,
			   path, kmsg);
	       my_quit = quit;
	       my_die = die;
	       snprintf(tmp, 255, "QUIT :Killed (%s (%s))\r", fd->nick, kmsg);
	       parse(tmp, u, &nuh, &nuh);
	       quit = my_quit;
	       die = my_die;
	       u->is_killed = 1;
	       return;
	  }
	  u = u->next;
     }

     irc_fprintf(fd->sok, ERR_NOSUCHNICK, server_name, fd->nick, av[1]);
     return;
}

/* m_users
   av[0] = "USERS"
   does nothing
 */

void m_users(struct fd_list *fd, char **av, int ac)
{
     irc_fprintf(fd->sok, ":%s 446 %s :USERS has been disabled\r\n", server_name, fd->nick);
     return;
}

void m_help(struct fd_list *fd, char **av, int ac);
void m_die(struct fd_list *fd, char **av, int ac);
void m_restart(struct fd_list *fd, char **av, int ac);

/* m_error
   dummy function
 */

void m_error(struct fd_list *fd, char **av, int ac)
{
     return;
}

/* m_trace
   av[0] = "TRACE"
   av[1] = nick
   returns a trace to nick
 */

void m_trace(struct fd_list *fd, char **av, int ac)
{
     struct fd_list *z = __head->next;

     while (z) {
	  if (!strcasecmp(z->nick, av[1])) {
	       irc_fprintf(fd->sok, ":%s 204 %s %s %i %s[%s] %i\r\n", server_name, fd->nick,
			 (z->oper ? "Oper" : "User"), (z->oper ? 1 : 0), z->nick, z->hostname, 0);
	       return;
	  }
	  z = z->next;
     }
     irc_fprintf(fd->sok, ":%s 206 %s Serv 0 05 0C <No_match> *!*@%s 0 0\r\n", server_name, fd->nick, server_name);
     return;
}

/* m_summon
   dummy function
 */

void m_summon(struct fd_list *fd, char **av, int ac)
{
     irc_fprintf(fd->sok, ":%s 445 %s :SUMMON has been disabled\r\n", server_name, fd->nick);
}

/* int in_use(char *nick)
   returns 1 if nick is in use, else 0
 */

int in_use(char *nick)
{
     struct fd_list *users = __head;

     while (users != NULL) {
	  if (!strcasecmp(nick, users->nick)) {
	       return 1;
	  }
	  users = users->next;
     }
     return 0;
}

/* m_ison
   av[0] = "ISON"
   av[1...] = nicks
   returns a list of nicks online
 */

void m_ison(struct fd_list *fd, char **av, int ac)
{
     if (ac > 1) {
	  int i = 1, len = 0;
	  char *buf = NULL;
	  struct fd_list *users = __head;

	  for (i = 1; (i < ac && i < 5); i++) {
	       users = __head;
	       while (users != NULL) {
		    if (!strcasecmp(users->nick, av[i])) {
			 buf = (char *) realloc(buf, len + strlen(users->nick) + 1);
			 sprintf(&(buf[len]), "%s ", users->nick);
			 len += strlen(users->nick);
			 len++;
			 users = NULL;
		    }
		    if (users)
			 users = users->next;
	       }
	  }
	  irc_fprintf(fd->sok, ":%s 303 %s :%s\r\n", server_name, fd->nick, buf);
	  free(buf);
	  return;
     } else {
	  __not_enough_parameters(fd, av[0]);
	  return;
     }
}

/* m_userip/m_userhost
   av[0] = "USERIP" / "USERHOST"
   av[1...] = nicks
   returns the ip/hostname of the users
 */

void m_userip(struct fd_list *fd, char **av, int ac)
{
     fprintf(stderr, "entering m_userip - fd = %p, av = %p, ac = %i\n", fd, av, ac);
     if (ac > 1) {
	  int i = 1, len = 0;
	  char *buf = NULL, tmp[75];
	  struct fd_list *users = __head->next;

	  for (i = 1; (i < ac && i < 5); i++) {
	       if (av[i] != NULL) {
		    bzero(tmp, 75);
		    users = __head->next;
		    while (users) {
			 if (!strcasecmp(users->nick, av[i])) {
			      if (users->hide_hostname && !fd->oper) {
				   if (users->oper) {
					snprintf(tmp, 75, "%s*=%c%s@0.0.0.0 ", users->nick, (users->away_set ? '-' : '+'),
						 users->userid);
				   } else {
					snprintf(tmp, 75, "%s=%c%s@0.0.0.0 ", users->nick, (users->away_set ? '-' : '+'),
						 users->userid);
				   }
			      } else {
				   struct sockaddr_in sa;
				   int sa_len;

				   sa_len = sizeof(sa);
				   if (!getpeername(users->fd, &sa, &sa_len)) {
					unsigned char *t = (unsigned char *) &(sa.sin_addr.s_addr);

					if (users->oper) {
					     snprintf(tmp, 75, "%s*=%c%s@%i.%i.%i.%i ", users->nick, (users->away_set ? '-' : '+'),
						      users->userid, t[0], t[1], t[2], t[3]);
					} else {
					     snprintf(tmp, 75, "%s=%c%s@%i.%i.%i.%i ", users->nick, (users->away_set ? '-' : '+'),
						      users->userid, t[0], t[1], t[2], t[3]);
					}
				   }
			      }

			      if (!buf) {
				   buf = (char *) malloc(strlen(tmp));
				   len = strlen(tmp);
			      } else {
				   len += strlen(tmp);
				   buf = (char *) realloc((void *) buf, len);
			      }
			      sprintf(&(buf[len - strlen(tmp)]), "%s", tmp);
			      users = NULL;
			 }
			 if (users)
			      users = users->next;
		    }
	       }
	  }
	  if (len)
	       buf[len - 1] = 0;
	  irc_fprintf(fd->sok, ":%s 304 %s :%s\r\n", server_name, fd->nick, buf);
	  irc_fflush(fd->sok);
	  return;
     }
     return;
}

void m_userhost(struct fd_list *fd, char **av, int ac)
{
     if (ac > 1) {
	  int i = 1, len = 0;
	  char *buf = NULL, tmp[75];
	  struct fd_list *users = __head;

	  for (i = 1; (i < ac && i < 5); i++) {
/* reply ::= nick[*]=<+-><hostname> */
	       if (av[i] != NULL) {
		    bzero(tmp, 75);
		    users = __head;
		    while (users != NULL) {
			 if (!strcasecmp(users->nick, av[i])) {
			      if (!fd->oper && users->hide_hostname) {
				   if (users->oper) {
					snprintf(tmp, 75, "%s*=%c%s@hidden.hurricane.net ", users->nick,
						 (users->away_set ? '-' : '+'), users->userid);
				   } else {
					snprintf(tmp, 75, "%s=%c%s@hidden.hurricane.net ", users->nick,
						 (users->away_set ? '-' : '+'), users->userid);
				   }
			      } else {
				   if (users->oper) {
					snprintf(tmp, 75, "%s*=%c%s@%s ", users->nick, (users->away_set ? '-' : '+'),
						 users->userid, users->hostname);
				   } else {
					snprintf(tmp, 75, "%s=%c%s@%s ", users->nick, (users->away_set ? '-' : '+'),
						 users->userid, users->hostname);
				   }
			      }

			      if (buf == NULL) {
				   buf = (char *) malloc(strlen(tmp));
				   len = strlen(tmp);
			      } else {
				   len += strlen(tmp);
				   buf = (char *) realloc((void *) buf, len);
			      }
			      sprintf(&(buf[len - strlen(tmp)]), "%s", tmp);
			      users = NULL;
			 }
			 if (users != NULL)
			      users = users->next;
		    }
	       }
	  }
	  if (len > 0) {
	       buf[len - 1] = 0;
	  }
	  irc_fprintf(fd->sok, ":%s 302 %s :%s\r\n", server_name, fd->nick, (buf == NULL ? "" : buf));
	  irc_fflush(fd->sok);
	  return;
     } else {
	  __not_enough_parameters(fd, av[0]);
	  return;
     }
}

/* *INDENT-OFF* */

Commands cmd_list[] = {
	/* Command      Function        reg'd   unreg'd oper'd  help text */
{	"ADMIN", 	m_admin, 	1, 	0, 	0, 	"ADMIN - Get administrative info"},
{	"AWAY", 	m_away, 	1, 	0, 	0, 	"AWAY [:<message>] - Set/clear an away message"},
{	"CONNECT", 	m_server, 	1, 	0, 	0, 	"CONNECT <servername> [<port> [<remote server>]] - "
								"Connect a new server to the network"},
{	"DIE", 		m_die, 		1, 	0, 	1, 	"DIE - Shut the server down"},
{	"DNS", 		m_dns, 		1, 	0, 	0, 	"DNS <l> - Display resolver statistics"},
{	"ERROR", 	m_error, 	1, 	1, 	0, 	"ERROR <msg> - Report an error condition to the server (server-server only)"},
{	"GLINE", 	m_gline, 	1, 	0, 	0, 	"GLINE [+|-]<user@host> [length] :[reason] - "
      								"Adds, extends, searches for, or displays the list of glines."},
{	"GLOBOPS", 	m_globops, 	1, 	0, 	1, 	"GLOBOPS :<text> - Send a global message to all opers with +g"},
{	"HELP", 	m_help, 	1, 	0, 	0, 	"HELP <command> - Display information on a command, or show all supported commands"},
{	"INFO", 	m_info, 	1, 	0, 	0, 	"INFO - Get server information"},
{	"INVITE", 	m_invite, 	1, 	0, 	0, 	"INVITE <channel> <nick> - Invite a user to a channel"},
{	"INVITES",	m_invites, 	1, 	0, 	0, 	"INVITES <channel> - Shows the list of people invited to <channel>"},
{	"ISON", 	m_ison, 	1,	0, 	0, 	"ISON <nick> [<nicks..>] - Check to see if the users are online"},
{	"JOIN", 	m_join, 	1, 	0, 	0, 	"JOIN <channel> [<key>] - Join a channel"},
{	"KICK", 	m_kick, 	1, 	0, 	0, 	"KICK <channel> <nick> :<reason> - Kick a user out of a channel"},
{	"KILL", 	m_kill, 	1, 	0, 	1, 	"KILL <nick> :<kill msg> - Forcibly signs a user off of IRC"},
{	"LOAD", 	m_dl, 		1, 	0, 	1, 	"LOAD <module name> - Load a module"},
{	"LINKS", 	m_links, 	1, 	0, 	0, 	"LINKS - Get link information"},
{	"LIST", 	m_list, 	1, 	0, 	0, 	"LIST [<channel>] - Retrieve a listing of channels on the server"},
{	"LUSERS", 	m_lusers, 	1, 	0, 	0, 	"LUSERS - Get information on number of users on network"},
{	"MEMINFO", 	m_meminfo, 	1, 	0, 	1, 	"MEMINFO - Get information on currently allocated memory."},
{	"MODE", 	m_mode, 	1, 	0, 	0, 	"MODE <nick/channel> [<modes> [<parameters>]] - "
 								"Check/set channel/user modes"},
{	"MOTD", 	m_motd, 	1, 	0, 	0, 	"MOTD - Get the server MOTD"},
{	"NAMES", 	m_names, 	1, 	0, 	0, 	"NAMES <channel> - Get list of nicks on a channel"},
{	"NICK", 	m_nick, 	1, 	1, 	0, 	"NICK <newnick> - Change your nickname"},
{	"OPER", 	m_oper, 	1, 	0, 	0, 	"OPER <nick> <password> - Attempt to gain IRC Operator status"},
{	"PART", 	m_part, 	1, 	0, 	0, 	"PART <channel> - Part a channel"},
{	"PASS", 	m_pass, 	0, 	1, 	0, 	"PASS <pass> - Define a connection password"},
{	"PING", 	m_ping, 	1, 	1, 	0, 	"PING <destination> <ping> - Send a ping to a server"},
{	"PONG", 	m_pong, 	1, 	1, 	0, 	"PONG <ping> <destination> - Reply to a ping message"},
{	"PRIVMSG", 	m_sendmsg, 	1, 	0, 	0, 	"PRIVMSG <nick/channel> :<text> - Send a message"},
{	"NOTICE", 	m_sendmsg, 	1, 	0, 	0, 	"NOTICE <nick/channel> :<text> - Send a notice"},
{	"QUIT", 	m_quit, 	1, 	1, 	0, 	"QUIT :<quit msg> - Signs you off of IRC"},
{	"REHASH", 	m_rehash, 	1, 	0, 	1, 	"REHASH - Rehash the config file"},
{	"RESTART", 	m_restart, 	1, 	0, 	1, 	"RESTART - Restart the server"},
{	"SERVER", 	m_server, 	0, 	1, 	0, 	"SERVER <servername> <hops> :<server info> - "
 								"Register a new server"},
{	"SETFLUD", 	m_setflud, 	1, 	0, 	1, 	"SETFLUD <num msgs> - Set the flood rate, <num msgs> in 2 seconds"},
{	"SETTIME", 	m_settime, 	1, 	0, 	1, 	"SETTIME <time in seconds> - Set the current time"},
{	"SQUIT", 	m_server, 	1, 	0, 	0, 	"SQUIT <servername> [:<reason>] - Disconnect a server"},
{	"STATS", 	m_stats, 	1, 	0, 	0, 	"STATS <type> <server> - Get server conf information"},
{	"SUMMON", 	m_summon, 	1, 	0, 	0, 	"SUMMON <userid> - Summon a user to IRC"},
{	"TIME", 	m_time, 	1, 	0, 	0, 	"TIME - Get the current time"},
{	"TOPIC", 	m_topic, 	1, 	0, 	0, 	"TOPIC <channel> [:<text>] - Check/set channel topic"},
{	"TRACE", 	m_trace, 	1, 	0, 	0, 	"TRACE <nick> - Find the route to a user"},
{	"UNLOAD", 	m_unload, 	1, 	0, 	1, 	"UNLOAD <module name> - Unload a module"},
{	"USER", 	m_user, 	0, 	1, 	0, 	"USER <userid> <hostname> <servername> :<realname> - "
 								"Tells the server that a new user is signing on"},
{	"USERIP", 	m_userip, 	1, 	0, 	0, 	"USERIP <nick> [<nicks..>] - Get the IP address of the users"},
{	"USERHOST", 	m_userhost, 	1, 	0, 	0, 	"USERHOST <nick> [<nicks..>] - Get the hostname of the users"},
{	"USERS", 	m_users, 	1, 	0, 	0, 	"USERS [<server>] - Find out who is logged in on the machine the server is running on"},
{	"VERSION", 	m_version, 	1, 	0, 	0, 	"VERSION [<server>] - Get a server's version"},
{	"WALLOPS", 	m_wallops, 	1, 	0, 	1, 	"WALLOPS :<text> - Send a global message to all users with +w"},
{	"WHO", 		m_who, 		1, 	0, 	0, 	"WHO <nick/channel> - Get information on a user"},
{	"WHOIS", 	m_whois, 	1, 	0, 	0, 	"WHOIS [<server>] <nick> - Get information on a user"},
{	"WHOWAS", 	m_whowas, 	1, 	0, 	0, 	"WHOWAS <nick> [<num entries>] - Find out who was on earlier"},
{	NULL, 		NULL, 		0, 	0, 	0, 	NULL}
};

/* *IMDENT-ON* */

/* m_help
   av[0] = "HELP"
   av[1] = cmd
   if cmd: returns help for cmd
   otherwise returns list of valid commands
 */

void m_help(struct fd_list *fd, char **av, int ac)
{
     Commands *p = cmd_list;
     Module *m = module_list->next;

     if (ac > 1 && av[1][0] != 0) {
	  while (p->name != NULL) {
	       if (!strcasecmp(av[1], p->name)) {
		    irc_fprintf(fd->sok, ":%s NOTICE %s :%s\r\n", NAME, fd->nick, p->helptext);
		    return;
	       }
	       p++;
	  }

	  irc_fprintf(fd->sok, ":%s NOTICE %s :%s\r\n", NAME, fd->nick, "No help availiable for command.\r\n");
	  return;
     }
     while (p->name != NULL) {
	  irc_fprintf(fd->sok, ":%s NOTICE %s :%s\r\n", NAME, fd->nick, p->name);
	  p++;
     }
     while (m != NULL) {
	  Functions *fn_list;

	  fn_list = m->function_list;
	  while (fn_list->cmd_name != NULL) {
	       irc_fprintf(fd->sok, ":%s NOTICE %s :%s (from %s)\r\n", NAME, fd->nick, fn_list->cmd_name,
			   m->info->module_name);
	       fn_list++;
	  }
	  m = m->next;
     }
}

char *tmp, *t;

/* char *_parse(char *input, struct fd_list *fd, int _quit, int *_die, int *_restart)
   parses input
 */

char *
 _parse(char *input, struct fd_list *fd, int *_quit, int *_die, int *_restart)
{
     char **av, *lptr, *p;
     int ac, colon, i, done, src;
     Module *module_ptr;

     quit = _quit;
     die = _die;
     restart = _restart;

     *quit = 0;
     *die = 0;
     *restart = 0;

     if (input == NULL || (*input == '\r' || *input == '\n' || *input == 0)) {
	  return NULL;
     }
     tmp = (char *) alloca(strlen(input) + 1);
     p = tmp;
     strncpy(tmp, input, strlen(input));

     av = NULL;
     lptr = tmp;

#if 0
     if (lptr[0] == ':') {
	  lptr++;
	  tmp++;
	  src = 1;
     }
#endif

     ac = 0;
     colon = 0;
     while (*tmp != 0 && *tmp != '\r' && *tmp != '\n' && !colon) {
	  if (*tmp == ':') {
	       colon = 1;
	       *tmp = 0;
	       tmp++;
	       lptr = tmp;
	  } else if (*tmp == ' ') {
	       av = (char **) realloc(av, sizeof(char **) * ++ac);

	       av[ac - 1] = lptr;
	       *tmp = 0;
	       tmp++;
	       lptr = tmp;
	  } else {
	       tmp++;
	  }
     }

#if 0
     if (!src) {
	  av = (char **) realloc(av, sizeof(char **) * ++ac);
	  memmove(av + 1, av, sizeof(char **) * (ac - 1));

	  if (fd->is_server) {
	       av[0] = fd->server_name;
	  } else {
	       av[0] = fd->nick;
	  }
     }
#endif

     av = (char **) realloc(av, sizeof(char **) * ++ac);

     av[ac - 1] = lptr;
     av = (char **) realloc(av, sizeof(char **) * (ac + 1));

     av[ac] = NULL;

     for (i = 0; i < strlen(av[ac - 1]); i++) {
	  if (av[ac - 1][i] == '\n' || av[ac - 1][i] == '\r') {
	       av[ac - 1][i] = 0;
	  }
     }

     tmp = p;

     t = av[0];
     done = 0;
     module_ptr = module_list->next;

     while (module_ptr != NULL) {
	  Functions *fn_list;
	  int passthru;

	  fn_list = module_ptr->function_list;
	  while (fn_list->cmd_name != NULL) {
	       if (!strcasecmp(fn_list->cmd_name, t)) {
		    int ok = 0;

		    if (fn_list->need_registered && fn_list->need_unregistered) {
			 ok = 1;
		    } else if (fn_list->need_registered) {
			 if (fd->registered) {
			      if (fn_list->need_oper) {
				   if (fd->oper) {
					ok = 1;
				   } else {
					__no_priveliges(fd);
					ok = 0;
				   }
			      } else {
				   ok = 1;
			      }
			 } else {
			      __not_registered(fd);
			      ok = 0;
			 }
		    } else if (fn_list->need_unregistered) {
			 if (!fd->registered && fd->nick[0] == 0 || fd->userid[0] == 0 || fd->hostname[0] == 0) {
			      ok = 1;
			 } else {
			      __already_registered(fd);
			      ok = 0;
			 }
		    }
		    if (ok) {
			 char **p = av;

			 done = 1;
			 passthru = fn_list->function(fd, p, ac);
			 fprintf(stderr, "returned from function..\n");
		    }
		    if (!passthru) {
			 fprintf(stderr, "freeing arg list & tmp variable and returning..\n");
			 free(av);
			 fprintf(stderr, "freed tmp var\n");
			 goto done;
		    }
	       }
	       fn_list++;
	  }
	  module_ptr = module_ptr->next;
     }

     for (i = 0; i < ((sizeof(cmd_list) / sizeof(cmd_list[0])) - 1); i++) {
	  if (!strcasecmp(cmd_list[i].name, t)) {
	       int ok = 0;

	       if (cmd_list[i].need_registered && cmd_list[i].need_unregistered) {
		    ok = 1;
	       } else if (cmd_list[i].need_registered) {
		    if (fd->registered) {
			 if (cmd_list[i].need_oper) {
			      if (fd->oper) {
				   ok = 1;
			      } else {
				   __no_priveliges(fd);
				   ok = 0;
			      }
			 } else {
			      ok = 1;
			 }
		    } else {
			 __not_registered(fd);
			 ok = 0;
		    }
	       } else if (cmd_list[i].need_unregistered) {
		    if (!fd->registered) {
			 ok = 1;
		    } else {
			 __already_registered(fd);
			 ok = 0;
		    }
	       }
	       if (ok) {
		    cmd_list[i].function(fd, av, ac);
	       }
	       free(av);
	       goto done;
	  }
     }

     goto unknown;
   done:

     update_cmd(p, input);

     if (fd->msgs.time + 2 < __time(NULL)) {
	  fd->msgs.time = __time(NULL);
	  fd->msgs.nmsgs = 1;
     } else {
	  fd->msgs.nmsgs++;
	  if (fd->msgs.nmsgs > flud) {
	       parse("QUIT :Excess flood\r", fd, quit, die);
	  }
     }

     return NULL;

   unknown:
     __unknown_command(fd, t);
     free(av);
     return NULL;
}

/* m_die
   av[0] = "DIE"
   kills the server
 */

void m_die(struct fd_list *fd, char **av, int ac)
{
     *die = 1;
/*      syslog(LOG_CRIT, "%s (%s@%s) is shutting down ircd!!", fd->nick, fd->userid, fd->hostname); */
     return;
}

/* m_restart
   av[0] = "RESTART"
   restarts the server
 */

void m_restart(struct fd_list *fd, char **av, int ac)
{
     *die = 1;
     *restart = 1;
/*      syslog(LOG_CRIT, "%s (%s@%s) is restarting ircd!!", fd->nick, fd->userid, fd->hostname); */
     return;
}

/* int m_read(int fd, void *buf, int maxlen)
   reads from fd, stores in buf
 */

int m_read(int fd, void *buf, int maxlen)
{
     char *t;
     int m = 0, n = 0;
     int no_read = 0;

     errno = 0;
     t = (char *) buf;
     while (*t != '\n' && *t != '\r' && n < maxlen) {
	  if (m > 0) {
	       t++;
	       n++;
	  } else if (m == 0) {
	       no_read++;
	       if (no_read > 200) {
		    errno = ECONNRESET;
		    return -1;
	       }
	  }
	  m = read(fd, (void *) t, 1);
	  if (errno != 0 && errno != ESPIPE && errno != EAGAIN) {
	       return -1;
	  }
	  errno = 0;
     }
     return n;
}

struct server_list *sptr_head;

time_t last_ping;
time_t ping_check_len = 60;

time_t
check_pings()
{
     struct fd_list *u = __head->next;
     time_t now = __time(NULL);
     int nuh;


     while (u) {
	  irc_fprintf(u->sok, "PING :%s\r\n", server_name);
	  irc_fflush(u->sok);
	  u = u->next;
     }
     return now;
}

#include <signal.h>

int pipe_caught;

void pipe_handler(int p)
{
     pipe_caught = 1;
     return;
}

#ifdef PTHREADIDENT
void exit_thread()
{
     pthread_exit(NULL);
}

#endif

/*

   struct __pthread_mem_t {
   void *__ptr;
   void *__next;
   };

   typedef struct __pthread_mem_t pthread_mem_t;

   pthread_mem_t *pt_mem_list = NULL;

   void *pthread_free_mem_internal(void *ptr) {
   struct __pthread_mem_t *list = (struct __pthread_mem_t *) ptr;
   while (list) {
   struct fd_list *q = list->__ptr;
   while (q->nslookup_is_in_use || q->ident_is_in_use) {
   sleep(1);
   }
   free(q);
   list = list->__next;
   }
   list = (struct __pthread_mem_t *) ptr;
   while (list) {
   struct __pthread_mem_t *next = list->__next;
   free(list);
   list = next;
   }
   pthread_exit(NULL);
   }

   void pthread_do_free_mem_np(void *head) {
   pthread_t thread;
   pthread_create(&thread, NULL, pthread_free_mem_internal, head);
   pthread_detach(thread);
   pt_mem_list = NULL;
   return;
   }

   void pthread_free_mem_np(void *mem) {
   FILE2 *pt_log = irc_fopen("pthread.log", "a+");
   irc_fprintf(pt_log, "thread %i: pthread_free_mem_np(%p)\n", pthread_self, mem);
   irc_fclose(pt_log);
   if (!pt_mem_list) {
   pthread_mem_t *q = malloc(sizeof(pthread_mem_t));
   q->__ptr = mem;
   q->__next = NULL;
   } else {
   pthread_mem_t *l = pt_mem_list;
   pthread_mem_t *q = malloc(sizeof(pthread_mem_t));
   while (l->__next) l = l->__next;
   l->__next = q;
   q->__next = NULL;
   q->__ptr = mem;
   }
   return;
   }

 */

void sigsegv(int) __attribute__ ((__noreturn__));

void init_signals()
{
     struct sigaction *sa;
     sa = (struct sigaction *) malloc(sizeof(struct sigaction));

     sa->sa_flags = SA_ONESHOT;

     sigaction(SIGPIPE, sa, NULL);
#ifdef PTHREADIDENT
     signal(SIGTERM, exit_thread);
#endif
     signal(SIGSEGV, sigsegv);
     return;
}

void sigsegv(int i) {
   char buf[255];
   struct timeval tv;
   write(2, (void *) "oops! sigsegv occurred, attempting to dump core\n\0", 48);
   gettimeofday(&tv, NULL);
   snprintf(buf, 255, "%s.%i.%i", "core", tv.tv_sec, tv.tv_usec);
   rename("core", buf);
   raise(SIGTRAP);
   _exit(1);
}

int main(int argc, char **argv, char **envp)
{
     int fd, fd2, pid, bigfd;
     int restart;
     struct sockaddr_in *sa;
     unsigned char *t;
     FILE2 *sok;
     FILE2 q;
     fd_set fds;
     struct timeval tv;
     struct fd_list *head, *list;

     restart = 0;
     openlog("ircd", LOG_PID | LOG_CONS, LOG_DAEMON);
     last_ping = __time(NULL);
/* #if !defined(DEBUGMAIN) || !defined(DEBUGSOCK) || !defined(DEBUGMEM) */
     pid = fork();
     if (pid) {
	  printf("forked into background, pid %i\n", pid);
	  exit(0);
     }
/*      syslog(LOG_INFO, "ircd %s starting up", VERSION); */
      close(0);
     close(1);
     close(2);
     open("/dev/null", O_RDONLY);
     open("/dev/null", O_WRONLY);
#if defined(DEBUGMAIN) || defined(DEBUGSOCK) || defined(DEBUGMEM)
     unlink("ircd.out");
     open("ircd.out", O_WRONLY | O_CREAT, S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP);
#else
     open("/dev/null", O_WRONLY);
#endif

/* #endif */

#ifdef DEBUGMAIN
     log("defining signal handlers");
#endif
     init_signals();
     pipe_caught = 0;
     signal(SIGPROF, SIG_IGN);
     signal(SIGCHLD, SIG_IGN);	/* automatically reap any child processes.. */
#ifdef DEBUGMAIN
     log("finished defining signal handlers");
#endif

     log("initializing lists..");
     log("- garbage collector");
     __gcInit();
     log("- channels");
     __init_chan_list();
     log("- modules");
     __init_mod_list();
     log("- dns");
     __init_dns_list();
     __whowas_head = NULL;
     st_implicit_init();

     {
	  struct server_list *sp;
	  sp = (struct server_list *) gcAlloc(sizeof(struct server_list));

	  sp->magic = SERVER_MAGIC;
	  strncpy(sp->server_name, NAME, 100);
	  strncpy(sp->server_desc, server_desc, 150);
	  sp->hops = 0;
	  sptr_head = sp;
     }

     memset((void *) &msgs, 0, sizeof(struct msg_use));

     start_time = __time(NULL);

     __read_conf_file();

     /* check number of arguments */
     if (argc > 1) {		/* the only argument we accept right now is a port number, so if there's more than one argument
				   (the first being the name of the program), try to convert it into a number. */
	  port = strtol(argv[1], (char **) NULL, 10);
	  if (port == 0 || port > 65535) {	/* is the port invalid? */
	       error("argv[1] was not a valid port number", 2);		/* exit */
	  }
     } else {			/* there aren't any arguments, so use the compiled-in port number */
	  port = PORT;
     }

     if (server_hn[0] == 0 || server_hn[0] == '*') {
	  fd = open_listen_sok(port);	/* open a socket to listen on (see listen.c) */
     } else {
	  struct hostent *host_struct, he_buf;
	  char buf[100];

	  gethostbyname_r(server_hn, &he_buf, buf, 100, &host_struct, &h_errno);
	  if (host_struct != NULL) {
	       int *i = (int *) host_struct->h_addr_list[0];

	       server_ip = *i;
	  } else {
	       server_ip = 0;
	  }
	  fd = open_listen_sok_2(port, server_ip);
     }

     if (fd < 0) {		/* did the call fail? */
	  ERROR("open_listen_sok() failed", 1);		/* exit */
     }
/*      syslog(LOG_INFO, "Accepting connections on port %i", port); */
     {
	  struct pline *port_list;

	  port_list = plines;
	  while (port_list) {
	       port_list->fd = open_listen_sok_2(port_list->port, port_list->ipnum);
	       if (port_list->fd) {
/*                              syslog(LOG_INFO, "Accepting connections on port %i", port_list->port); */
	       }
	       port_list = port_list->next;
	  }
     }

     sa = (struct sockaddr_in *) malloc(sizeof(struct sockaddr_in));	/* allocate a sockaddr_in structure for the

									   accept_connect down a bit farther */
     if (sa == NULL) {		/* did we malloc() successfully? */
	  error("malloc() failed", 1);	/* no, exit */
     }
     head = list = (struct fd_list *) gcAlloc(sizeof(struct fd_list));	/* allocate an fd_list structure for us to store

									   file descriptors in (both FILE * and int
									   types) */
     if (head == NULL) {	/* did we malloc() successfully? */
	  error("malloc() failed", 1);	/* no, exit */
     }
     __head = head;
     head->magic = FD_MAGIC;

     fd2 = 0;			/* initialize this value for the next piece of code */

     while (fd2 < 1) {
	  FD_ZERO(&fds);
	  FD_SET(fd, &fds);
	  bigfd = fd;
	  {
	       struct pline *list;

	       list = plines;
	       while (list) {
		    if (list->fd != -1) {
			 FD_SET(list->fd, &fds);
			 if (bigfd < list->fd) {
			      bigfd = list->fd;
			 }
		    }
		    list = list->next;
	       }
	  }

	  select(bigfd + 1, &fds, NULL, NULL, NULL);

	  if (FD_ISSET(fd, &fds)) {	/* it's the primary port */
	       fd2 = accept_connect(fd, (struct sockaddr *) sa);
	  } else {
	       struct pline *list;

	       list = plines;
	       while (list) {
		    if (list->fd != -1) {
			 if (FD_ISSET(list->fd, &fds)) {
			      fd2 = accept_connect(list->fd, (struct sockaddr *) sa);
			      list = NULL;
			 }
		    }
		    if (list)
			 list = list->next;
	       }
	  }
     }

     fcntl(fd2, F_SETFL, O_NONBLOCK);	/* make this a non-blocking fd so that accept_connect() immediately returns
					   in any further calls */

     t = (char *) &(sa->sin_addr.s_addr);	/* set this to the value of the ip address so we can print the x.x.x.x
						   formatted value of it */
     /* print the value of sa to the client and the console */
     printf("fd for this connection: %i\n", fd2);
     printf("sa->sin_addr.s_addr = %u (%i.%i.%i.%i)\n", (unsigned int) ntohl(sa->sin_addr.s_addr), t[0], t[1], t[2], t[3]);
     printf("sa->sin_port = %i\n", ntohs(sa->sin_port));
     printf("sa->sin_family = %i\n", sa->sin_family);

     sok = irc_fdopen(fd2, "r+");	/* before we send sa to the client, turn it into a FILE * so it can be accessed using
					   the f* family of calls */
     irc_fflush(sok);		/* we have to fflush the socket after each call, otherwise it doesn't receive any data */

     /* addition of module connect handlers here */

     q.sok = stdin;
     pthread_mutex_init(&(q.mut), NULL);


     list->next = NULL;		/* we don't have a next record ... */
     list->prev = NULL;		/* this is the first record in the list */
     list->sok = &q;
     list->fd = 0;		/* this is the fd of stdin */
     fcntl(0, F_SETFL, O_NONBLOCK);
     sprintf(list->nick, "console");	/* this would be the console user */

     {
	  struct fd_list *t;

	  t = (struct fd_list *) gcAlloc(sizeof(struct fd_list));

	  if (t == NULL) {
	       error("malloc() failed", 1);
	  }
	  t->magic = FD_MAGIC;

	  list->next = t;

	  t->next = NULL;	/* we don't have a next record here either */
	  t->prev = list;	/* we DO have a previous record here, the head of the list */
	  t->sok = sok;		/* this is our FILE * from fdopen */
	  t->fd = fd2;		/* the fd that we got from accept_connect */
	  t->nick[0] = 0;	/* set it to this so that all of the routines know we haven't set a username */
	  t->sptr = sptr_head;
	  setvbuf(t->sok->sok, t->buf, _IOLBF, 1024);
     }

     {
	  Module *_list = module_list->next;

	  while (_list) {
	       if (_list->other_funcs) {
		    EventFunctions *p = _list->other_funcs->connect;

		    if (p) {
			 while (p->function != NULL) {
			      p->function(list->next);
			      p++;
			 }
		    }
	       }
	       _list = _list->next;
	  }
     }

     while (1) {
	  int rval, max_fd = 0;

	  FD_ZERO(&fds);
	  /* load the fd set with the fds that we have */
	  {
#ifdef DEBUGMAIN
	       log("traversing linked list");
#endif
	       list = head;
	       while (list != NULL) {
#ifdef DEBUGSOCK
		    {
			 char tmp[255];

			 snprintf(tmp, 255, "adding fd %i to fd set", list->fd);
			 log(tmp);
		    }
#endif
		    if (list->fd != 0) {
			 FD_SET(list->fd, &fds);
			 if (list->fd > max_fd) {	/* is the fd we're adding to the set greater than
							   the largest fd that we've added to the set? */
			      max_fd = list->fd;
			 }
		    }
		    list = list->next;
	       }
	  }
	  FD_SET(fd, &fds);	/* add the fd that we're listening on */
	  if (fd > max_fd) {	/* is the number of the fd that we're listening on greater than any other fd? */
	       max_fd = fd;
	  } {
	       struct pline *list;

	       list = plines;
	       while (list) {
		    if (list->fd != -1) {
			 FD_SET(list->fd, &fds);
			 if (list->fd > max_fd) {
			      max_fd = list->fd;
			 }
		    }
		    list = list->next;
	       }
	  }

	  /* timeout of 10 seconds on select() */
	  tv.tv_sec = 10;
	  tv.tv_usec = 0;
#ifdef DEBUGSOCK
	  log("calling select()");
#endif
	  rval = select(max_fd + 1, &fds, NULL, NULL, &tv);
#ifdef DEBUGSOCK
	  {
	       char tmp[255];

	       snprintf(tmp, 255, "select() returned %i", rval);
	       log(tmp);
	  }
#endif
	  if (rval < 0) {	/* error in select, client probably closed connection */
	       list = head;
	       while (list != NULL) {
		    FD_ZERO(&fds);
		    FD_SET(list->fd, &fds);
		    tv.tv_sec = 0;
		    tv.tv_usec = 1;
		    rval = select(list->fd + 1, &fds, NULL, NULL, &tv);
		    if (rval < 0) {	/* here's the client that disconnected */
			 irc_fclose(list->sok);		/* close their socket */
			 (list->prev)->next = list->next;	/* unlink them from the list */
			 (list->next)->prev = list->prev;
			 {
			      struct fd_list *t = list->prev;

			      gcDecRefcnt(list);
			      list = t;
			 }
		    }
		    list = list->next;
	       }
	  } else if (rval > 0) {	/* someone has some data on their socket */
	       int qval = 0;

#ifdef DEBUGMAIN
	       log("traversing linked list");
#endif
	       list = head;
	       while (list != NULL) {

#ifdef DEBUGSOCK
		    {
			 char tmp[255];

			 snprintf(tmp, 255, "checking to see if fd %i is in set", list->fd);
			 log(tmp);
		    }
#endif

		    if (FD_ISSET(list->fd, &fds)) {
			 char buf[1024];
			 int nbytes;

			 memset((void *) buf, 0, 1024);
#ifdef DEBUGSOCK
			 {
			      char tmp[255];

			      snprintf(tmp, 255, "fd %i is in list, reading from fd", list->fd);
			      log(tmp);
			 }
#endif
			 /* fgets(buf, 255, list->sok); */
			 if (!qval) {
			      nbytes = m_read(list->fd, (void *) buf, 1024);
			 } else {
			      nbytes = 1;
			      snprintf(buf, 1024, "QUIT :Ping timeout\r");
			 }

			 if (nbytes == 0) {
#ifdef DEBUGSOCK
			      log("socket was reported as ready for read but no data was availiable");
#endif
			 } else if (nbytes == -1) {
			      snprintf(buf, 1024, "QUIT :Read error for %s[%s]: errno %i: %s\r", list->nick,
				       list->hostname, errno, strerror(errno));
			 }
#ifdef DEBUGSOCK
			 {
			      char tmp[1050];

			      snprintf(tmp, 1050, "data read: %s", buf);
			      log(tmp);
			      snprintf(tmp, 255, "number of bytes: %i", strlen(buf));
			      log(tmp);
			 }
#endif
			 {
			      struct fd_list *t;
			      char *tmp;
			      int quit = 0, die = 0;

			      t = head;

			      tmp = _parse(buf, list, &quit, &die, &restart);
			      if (tmp != NULL) {
				   while (t != NULL) {
					if (t->fd == 0) {
					     fprintf(stdout, "%s\n", tmp);
					} else {
					     irc_fprintf(t->sok, "%s\n", tmp);
					}
					irc_fflush(t->sok);
					t = t->next;
				   }
				   __free(tmp);
			      }
			      if (pipe_caught && !quit) {
				   snprintf(buf, 1024, "QUIT :Write error for %s[%s]: %s\r", list->nick, list->hostname,
					    strerror(errno));
				   pipe_caught = 0;
				   parse(buf, list, &quit, &die);
			      }
			      if ((quit || list->is_killed)) {
				   struct fd_list *t2;

				   {
					Module *list = module_list;

					while (list) {
					     if (list->other_funcs) {
						  EventFunctions *p = list->other_funcs->signoff;

						  if (p) {
						       while (p->function != NULL) {
							    p->function(t);
							    p++;
						       }
						  }
					     }
					     list = list->next;
					}
				   }
#ifdef DEBUGMAIN
				   log("delinking fd_list structure");
#endif

				   if (list->prev == NULL) {
					t2 = list->next;
					(list->next)->prev = NULL;
				   } else if (list->next == NULL) {
					t2 = list->prev;
					(list->prev)->next = NULL;
				   } else if (list->next == NULL && list->prev == NULL) {
					t2 = NULL;
				   } else {
					t2 = list->prev;
					(list->prev)->next = list->next;
					(list->next)->prev = list->prev;
				   }
				   irc_fclose(list->sok);
				   gcDecRefcnt(list);
				   list = t2;
			      }
			      if (die) {
				   struct fd_list *t;

				   close(fd);
				   t = __head;
				   while (t != NULL) {
					close(t->fd);
					t = t->next;
				   }
				   {
					struct pline *q = plines;

					while (q) {
					     close(q->fd);
					     q = q->next;
					}
				   }
				   if (restart) {
					int pid;

					gcCancel();
					pthread_kill_other_threads_np();
					pid = fork();
					if (pid < 0) {
/*                                                                      syslog(LOG_CRIT, "fork() failed: %s", strerror(errno)); */
					     exit(0);
					} else if (!pid) {
					     char **argv;
					     int i;

					     for (i = 0; i < NFDBITS; i++) {
						  close(i);
					     }
					     argv = (char **) malloc(sizeof(char *) * 2);

					     argv[0] = "ircd";
					     argv[1] = NULL;
					     execve("ircd", argv, envp);
					}
				   }
				   exit(0);
			      }
			 }
		    }
		    if (list->is_killed) {
			 struct fd_list *t2;

			 {
			      Module *list = module_list;

			      while (list) {
				   if (list->other_funcs) {
					EventFunctions *p = list->other_funcs->signoff;

					if (p) {
					     while (p->function != NULL) {
						  p->function(head);
						  p++;
					     }
					}
				   }
				   list = list->next;
			      }
			 }
#ifdef DEBUGMAIN
			 log("delinking fd_list structure");
#endif
			 if (list->prev == NULL) {
			      t2 = list->next;
			      (list->next)->prev = NULL;
			 } else if (list->next == NULL) {
			      t2 = list->prev;
			      (list->prev)->next = NULL;
			 } else if (list->next == NULL && list->prev == NULL) {
			      t2 = NULL;
			 } else {
			      t2 = list->prev;
			      (list->prev)->next = list->next;
			      (list->next)->prev = list->prev;
			 }
			 irc_fclose(list->sok);
			 gcDecRefcnt(list);
			 list = t2;

		    }
		    list = list->next;
	       }
	  }
#ifdef DEBUGSOCK
	  log("finished checking open sockets, checking for new connections");
#endif

	  if (FD_ISSET(fd, &fds)) {
	       fd2 = accept_connect(fd, (struct sockaddr *) sa);
	       if (fd2 > 0) {
		    struct fd_list *t;
		    char *r;

#ifdef DEBUGSOCK
		    {
			 char tmp[255];

			 snprintf(tmp, 255, "new connection: fd %i", fd2);
			 log(tmp);
		    }
#endif
		    r = (char *) &(sa->sin_addr.s_addr);
		    fcntl(fd2, F_SETFL, O_NONBLOCK);
		    t = head;
		    while (t != NULL)
			 t = t->next;

		    {
			 struct fd_list *new_ent;
			 new_ent = (struct fd_list *) gcAlloc(sizeof(struct fd_list));

			 if (new_ent == NULL) {
			      error("malloc() failed", 1);
			 }
			 new_ent->magic = FD_MAGIC;
#ifdef DEBUGMAIN
			 {
			      char tmp[255];

			      snprintf(tmp, 255, "head of linked list = %p", (void *) head);
			      log(tmp);
			      snprintf(tmp, 255, "next item in list = %p", (void *) (head->next));
			      log(tmp);
			 }
#endif
			 t = head;
			 while (t->next != NULL) {
			      t = t->next;
#ifdef DEBUGMAIN
			      {
				   char tmp[255];

				   snprintf(tmp, 255, "current entry in list = %p", (void *) t);
				   log(tmp);
				   snprintf(tmp, 255, "next entry in list = %p", (void *) (t->next));
				   log(tmp);
			      }
#endif
			 }
			 t->next = new_ent;

			 new_ent->prev = t;	/* it's linked in after the last record */
			 new_ent->next = NULL;	/* it's the last record */
			 {
			      FILE2 *new_sok;

			      new_sok = irc_fdopen(fd2, "r+");
			      setvbuf(new_sok->sok, new_ent->buf, _IOLBF, 1024);
			      new_ent->sok = new_sok;
			 }
			 new_ent->fd = fd2;
			 new_ent->nick[0] = 0;
			 new_ent->sptr = sptr_head;

			 {
			      Module *list = module_list->next;

			      while (list) {
				   if (list->other_funcs) {
					EventFunctions *p = list->other_funcs->connect;

					if (p) {
					     while (p->function != NULL) {
						  p->function(new_ent);
						  p++;
					     }
					}
				   }
				   list = list->next;
			      }
			 }
		    }
	       }
	  } {
	       struct pline *plist;

	       plist = plines;
	       while (plist) {
		    if (plist->fd != -1) {
			 if (FD_ISSET(plist->fd, &fds)) {
			      fd2 = accept_connect(plist->fd, (struct sockaddr *) sa);
			      if (fd2 > 0) {
				   struct fd_list *t;
				   char *r;

#ifdef DEBUGSOCK
				   {
					char tmp[255];

					snprintf(tmp, 255, "new connection: fd %i", fd2);
					log(tmp);
				   }
#endif
				   r = (char *) &(sa->sin_addr.s_addr);
				   fcntl(fd2, F_SETFL, O_NONBLOCK);
				   t = head;
				   while (t != NULL)
					t = t->next;

				   {
					struct fd_list *new_ent;
					new_ent = (struct fd_list *) gcAlloc(sizeof(struct fd_list));

					if (new_ent == NULL) {
					     error("malloc() failed", 1);
					}
					new_ent->magic = FD_MAGIC;
#ifdef DEBUGMAIN
					{
					     char tmp[255];

					     snprintf(tmp, 255, "head of linked list = %p", (void *) head);
					     log(tmp);
					     snprintf(tmp, 255, "next item in list = %p", (void *) (head->next));
					     log(tmp);
					}
#endif
					t = head;
					while (t->next != NULL) {
					     t = t->next;
#ifdef DEBUGMAIN
					     {
						  char tmp[255];

						  snprintf(tmp, 255, "current entry in list = %p", (void *) t);
						  log(tmp);
						  snprintf(tmp, 255, "next entry in list = %p", (void *) (t->next));
						  log(tmp);
					     }
#endif
					}
					t->next = new_ent;

					new_ent->prev = t;	/* it's linked in after the last record */
					new_ent->next = NULL;	/* it's the last record */
					{
					     FILE2 *new_sok;

					     new_sok = irc_fdopen(fd2, "r+");
					     setvbuf(new_sok->sok, new_ent->buf, _IOLBF, 1024);
					     new_ent->sok = new_sok;
					}
					new_ent->fd = fd2;
					new_ent->nick[0] = 0;
					new_ent->sptr = sptr_head;

					{
					     Module *list = module_list->next;

					     while (list) {
						  if (list->other_funcs) {
						       EventFunctions *p = list->other_funcs->connect;

						       if (p) {
							    while (p->function != NULL) {
								 p->function(new_ent);
								 p++;
							    }
						       }
						  }
						  list = list->next;
					     }
					}
				   }
			      }
			 }
		    }
		    plist = plist->next;
	       }
	  }

#ifdef DEBUGMAIN
	  log("flushing all streams");
#endif
	  fflush(NULL);
	  if (__time(NULL) > ping_check_len + last_ping) {
	       last_ping = check_pings();
	  }
	  expire_glines();
#ifdef DEBUGMAIN
	  log("finished loop, returning to beginning.");
#endif
     }

     irc_fclose(sok);

     close(fd);
     exit(0);
}
