#include <stdio.h>
#include "struct.h"
#include "defines.h"
#include "mem.h"
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include "log.h"
#include "channels.h"
#include "stats.h"
#include "errors.h"
#include <netinet/in.h>
#include <netdb.h>
#include "info.h"
#include "connect.h"
#include <sys/time.h>
#include <stdarg.h>
#include "gc.h"

#ifdef PTHREADIDENT
#include <pthread.h>
#endif

#ifdef CRYPTEDOPER
#include <crypt.h>
#endif

extern struct oline *olines;
extern struct msg_use msgs;
extern char *tmp, *t;
extern int *quit;
extern int h_errno;

struct modes {
     char mode;
     int mval;
};

struct modes mlist[] =
{
     {'b', 0x00000001},
     {'c', 0x00000100},
     {'d', 0x00000002},
     {'g', 0x00000004},
     {'k', 0x00000008},
     {'i', 0x00000010},
     {'s', 0x00000020},
     {'w', 0x00000040},
     {'o', 0x00000080},
     {'x', 0x00000200},
     {0, 0},
};

static const int nmodes = 10;

char *
  get_umode(struct fd_list *fd)
{
     int mode = 0, n, k = 0;
     static char mstr[12];

     if (fd->show_whois)
	  mode |= 0x00000001;
     if (fd->deaf)
	  mode |= 0x00000002;
     if (fd->globops)
	  mode |= 0x00000004;
     if (fd->service)
	  mode |= 0x00000008;
     if (fd->invisible)
	  mode |= 0x0000010;
     if (fd->s_msg)
	  mode |= 0x00000020;
     if (fd->wallops)
	  mode |= 0x00000040;
     if (fd->oper)
	  mode |= 0x00000080;
     if (fd->show_connect)
	  mode |= 0x00000100;
     if (fd->hide_hostname)
	  mode |= 0x00000200;
     mstr[0] = '+';
     k = 1;
     for (n = 0; n < nmodes; n++) {
	  if (mode & mlist[n].mval) {
	       mstr[k++] = mlist[n].mode;
	  }
     }
     mstr[k] = 0;
     return mstr;
}

void __whois_reply(struct fd_list *fd, char *nick)
{
     struct fd_list *user;
     char *w_hostname;

     user = __head;
     while (user != NULL) {
	  if (!strcasecmp(user->nick, nick)) {
	       if (user->show_whois) {
		    irc_fprintf(user->sok, ":%s NOTICE %s :%s!%s@%s is performing a /WHOIS on you.\r\n", NAME,
				user->nick, fd->nick, fd->userid, fd->hostname);
	       }
	       w_hostname = (user->hide_hostname ? "hidden.hurricane.net" : user->hostname);
	       irc_fprintf(fd->sok, RPL_WHOISUSER, NAME, fd->nick, user->nick, user->userid, w_hostname, user->realname);
	       if (fd->show_whois) {
		    irc_fprintf(fd->sok, ":%s 608 %s %s is using modes: [%s]\r\n", NAME, fd->nick,
				user->nick, get_umode(user));
	       }
	       if (fd->oper && user->hide_hostname) {
		    irc_fprintf(fd->sok, ":%s 609 %s %s is from host: [%s]\r\n", server_name, fd->nick,
				user->nick, user->hostname);
	       }
	       if (user->chan_list != NULL && !user->service) {
		    char *buf = NULL;
		    struct user_chan_list *t;

		    t = (struct user_chan_list *) (user->chan_list);
		    {
			 struct user_list *u;
			 int in_chan = 0;

			 u = t->chan->users;
			 if (u->oper)
			      in_chan = 1;
			 else {
			      while (u != NULL) {
				   if (u->user == fd || u->oper) {
					in_chan = 1;
				   }
				   u = u->next;
			      }
			 }
			 u = t->chan->users;
			 while (u != NULL) {
			      if (u->user == user) {
				   if (u->oper || u->voice) {
					if (t->chan->private || t->chan->secret) {
					     if (in_chan) {
						  buf = (char *) malloc(strlen(t->chan->name) + 3);
						  if (u->oper) {
						       sprintf(buf, "@%s", t->chan->name);
						  } else {
						       sprintf(buf, "+%s", t->chan->name);
						  }
					     }
					} else {
					     buf = (char *) malloc(strlen(t->chan->name) + 3);
					     if (u->oper) {
						  sprintf(buf, "@%s", t->chan->name);
					     } else {
						  sprintf(buf, "+%s", t->chan->name);
					     }
					}
				   } else {
					if (t->chan->private || t->chan->secret) {
					     if (in_chan) {
						  buf = (char *) malloc(strlen(t->chan->name) + 2);
						  sprintf(buf, "%s", t->chan->name);
					     }
					} else {
					     buf = (char *) malloc(strlen(t->chan->name) + 2);
					     sprintf(buf, "%s", t->chan->name);
					}
				   }
				   u = NULL;
			      }
			      if (u != NULL)
				   u = u->next;
			 }
		    }
		    t = t->next;
		    while (t != NULL) {
			 struct user_list *u;
			 int in_chan = 0;

			 u = t->chan->users;
			 while (u != NULL) {
			      if (u->user == fd)
				   in_chan = 1;
			      if (u->user == user) {
				   if (u->oper || u->voice) {
					if (t->chan->private || t->chan->secret) {
					     if (in_chan) {
						  buf = (char *) realloc(buf, strlen(buf) + strlen(t->chan->name) + 3);
						  if (u->oper) {
						       sprintf(buf, "%s @%s", buf, t->chan->name);
						  } else {
						       sprintf(buf, "%s +%s", buf, t->chan->name);
						  }
					     }
					} else {
					     buf = (char *) realloc(buf, strlen(buf) + strlen(t->chan->name) + 3);
					     if (u->oper) {
						  sprintf(buf, "%s @%s", buf, t->chan->name);
					     } else {
						  sprintf(buf, "%s +%s", buf, t->chan->name);
					     }
					}
				   } else {
					if (t->chan->private || t->chan->secret) {
					     if (in_chan) {
						  buf = (char *) realloc(buf, strlen(buf) + strlen(t->chan->name) + 2);
						  sprintf(buf, "%s %s", buf, t->chan->name);
					     }
					} else {
					     buf = (char *) realloc(buf, strlen(buf) + strlen(t->chan->name) + 2);
					     sprintf(buf, "%s %s", buf, t->chan->name);
					}
				   }
				   u = NULL;
			      }
			      if (u != NULL)
				   u = u->next;
			 }
			 /* sprintf(buf, "%s %s", buf, t->chan->name); */
			 t = t->next;
		    }
		    if (buf != NULL) {
			 irc_fprintf(fd->sok, RPL_WHOISCHANNELS, NAME, fd->nick, user->nick, buf);
			 free(buf);
		    }
	       }
	       irc_fprintf(fd->sok, RPL_WHOISSERVER, NAME, fd->nick, user->nick, NAME, server_desc);
	       if (user->oper) {
		    irc_fprintf(fd->sok, RPL_WHOISOPERATOR, NAME, fd->nick, user->nick);
	       }
	       irc_fprintf(fd->sok, RPL_WHOISIDLE, NAME, fd->nick, user->nick, (int) (__time(NULL) - user->last_msg),
			   user->signon);
	       if (user->away_set) {
		    irc_fprintf(fd->sok, RPL_AWAY, NAME, fd->nick, user->nick, user->away_msg);
	       }
	       irc_fprintf(fd->sok, RPL_ENDOFWHOIS, NAME, fd->nick, user->nick);
	       irc_fflush(fd->sok);
	       return;
	  }
	  user = user->next;
     }

     irc_fprintf(fd->sok, ERR_NOSUCHNICK, NAME, fd->nick, nick);
     irc_fflush(fd->sok);
     return;
}

void sendto_opers(char *fmt,...)
{
     va_list ap;
     char buf[255];
     struct fd_list *fd = __head->next;

     va_start(ap, fmt);

     while (fd) {
	  if (fd->oper) {
	       snprintf(buf, 255, ":%s NOTICE %s :*** Notice -- %s\r\n", server_name, fd->nick, fmt);
	       irc_vfprintf(fd->sok, buf, ap);
	  }
	  fd = fd->next;
     }

     va_end(ap);
     return;
}

extern int match(char *, char *);

void __oper_reply(struct fd_list *fd, char *op_nick, char *op_pass)
{
     struct oline *lines;
     char userhost[200];

     snprintf(userhost, 200, "%s@%s", fd->userid, fd->hostname);
     lines = olines;

     while (lines != NULL) {
	  if (!match(lines->hostmask, userhost)) {
	       if (!strcasecmp(op_nick, lines->nick)) {
#ifdef CRYPTEDOPER
		    char salt[2];

		    salt[0] = lines->pass[0];
		    salt[1] = lines->pass[1];
		    if (!strcasecmp(crypt(op_pass, salt), lines->pass)) {
#else
		    if (!strcasecmp(op_pass, lines->pass)) {
#endif
			 sendto_opers("%s (%s@%s) is now operator (O)", fd->nick, fd->userid, fd->hostname);
			 irc_fprintf(fd->sok, RPL_YOUREOPER, NAME, fd->nick);
			 irc_fprintf(fd->sok, RPL_OPERMODE, fd->nick, fd->userid, fd->hostname, fd->nick);
			 irc_fflush(fd->sok);
			 fd->oper = 1;
			 return;
		    } else {
			 irc_fprintf(fd->sok, ERR_PASSWDMISMATCH, NAME, fd->nick);
			 irc_fflush(fd->sok);
			 return;
		    }
	       }
	  }
	  lines = lines->next;
     }
     irc_fprintf(fd->sok, ERR_NOOPERHOST, NAME, fd->nick);
     irc_fflush(fd->sok);
     return;
}

void __send_hprivmsg(struct fd_list *fd, char *dest, char *msg)
{
     struct fd_list *list;
     char *my_hostname = (fd->hide_hostname ? "hidden.hurricane.net" : fd->hostname);
     char *hmask = dest;

     hmask++;

     list = __head;

     fd->last_msg = __time(NULL);

     if (!fd->oper) {
	  irc_fprintf(fd->sok, ERR_NOSUCHNICK, server_name, fd->nick, dest);
	  return;
     }
     while (list != NULL) {
	  if (!match(hmask, list->hostname)) {
	       irc_fprintf(list->sok, PRIVMSG, fd->nick, fd->userid, my_hostname,
			   dest, msg);
	  }
	  list = list->next;
     }

     return;
}

void __send_hnotice(struct fd_list *fd, char *dest, char *msg)
{
     struct fd_list *list;
     char *my_hostname = (fd->hide_hostname ? "hidden.hurricane.net" : fd->hostname);
     char *hmask = dest;

     hmask++;

     list = __head;

     fd->last_msg = __time(NULL);

     if (!fd->oper) {
	  irc_fprintf(fd->sok, ERR_NOSUCHNICK, server_name, fd->nick, dest);
	  return;
     }
     while (list != NULL) {
	  if (!match(hmask, list->hostname)) {
	       irc_fprintf(list->sok, NOTICE, fd->nick, fd->userid, my_hostname,
			   dest, msg);
	  }
	  list = list->next;
     }

     return;
}

void __send_privmsg(struct fd_list *fd, char *dest, char *msg)
{
     struct fd_list *list;
     char *my_hostname = (fd->hide_hostname ? "hidden.hurrciane.net" : fd->hostname);

     list = __head;


     if (dest[0] == '#') {
	  __send_cprivmsg(fd, dest, msg);
	  return;
     } else if (dest[0] == '$') {
	  __send_hprivmsg(fd, dest, msg);
	  return;
     } else if (dest[0] == '@' || dest[0] == '+') {
	  __send_tprivmsg(fd, dest, msg);
	  return;
     }
     fd->last_msg = __time(NULL);

     while (list != NULL) {
	  if (!strcasecmp(list->nick, dest)) {
	       if (list->away_set) {
		    irc_fprintf(fd->sok, RPL_AWAY, NAME, fd->nick, list->nick, list->away_msg);
	       }
	       irc_fprintf(list->sok, PRIVMSG, fd->nick, fd->userid, my_hostname, dest, msg);
	       irc_fflush(list->sok);
	       return;
	  }
	  list = list->next;
     }

     irc_fprintf(fd->sok, ERR_NOSUCHNICK, NAME, fd->nick, dest);
     irc_fflush(fd->sok);
     return;
}

void __send_notice(struct fd_list *fd, char *dest, char *msg)
{
     struct fd_list *list;
     char *my_hostname = (fd->hide_hostname ? "hidden.hurricane.net" : fd->hostname);

     list = __head;

     if (dest[0] == '#') {
	  __send_cnotice(fd, dest, msg);
	  return;
     } else if (dest[0] == '$') {
	  __send_hnotice(fd, dest, msg);
	  return;
     } else if (dest[0] == '+' || dest[0] == '@') {
	  __send_tnotice(fd, dest, msg);
	  return;
     }
     while (list != NULL) {
	  if (!strcasecmp(list->nick, dest)) {
	       irc_fprintf(list->sok, NOTICE, fd->nick, fd->userid, my_hostname, dest, msg);
	       irc_fflush(list->sok);
	       return;
	  }
	  list = list->next;
     }

     irc_fprintf(fd->sok, ERR_NOSUCHNICK, NAME, fd->nick, dest);
     irc_fflush(fd->sok);
     return;
}

void __who_user(struct fd_list *fd, char *nick)
{
     struct fd_list *ulist;
     char *my_hostname;

     ulist = __head;

     while (ulist != NULL) {
	  if (!strcasecmp(ulist->nick, nick)) {
	       my_hostname = (ulist->hide_hostname ? "hidden.hurricane.net" : ulist->hostname);
	       irc_fprintf(fd->sok, RPL_WHOREPLY, NAME, fd->nick, "*", ulist->userid, my_hostname, NAME, ulist->nick,
			   'H', "", 1, ulist->realname);
	       irc_fprintf(fd->sok, RPL_ENDOFWHO, NAME, fd->nick, ulist->nick);
	       irc_fflush(fd->sok);
	       return;
	  }
	  ulist = ulist->next;
     }

     irc_fprintf(fd->sok, ERR_NOSUCHNICK, NAME, fd->nick, nick);
     irc_fflush(fd->sok);
     return;
}


#define ERR_UMODEUNKNOWNFLAG ":%s 501 %s :Unknown mode flag %c\r\n"

struct user_mode_change *
  __parse_umode(struct fd_list *fd, char *mode_str)
{
     struct user_mode_change *mode;
     int flag = 1;

     mode = (struct user_mode_change *) calloc(sizeof(struct user_mode_change));

     mode->invisible.orig = mode->invisible.new = fd->invisible;
     mode->s_msg.orig = mode->s_msg.new = fd->s_msg;
     mode->wallops.orig = mode->wallops.new = fd->wallops;
     mode->oper.orig = mode->oper.new = fd->oper;
     mode->deaf.orig = mode->deaf.new = fd->deaf;
     mode->service.orig = mode->service.new = fd->service;
     mode->globops.orig = mode->globops.new = fd->globops;
     mode->show_whois.orig = mode->show_whois.new = fd->show_whois;
     mode->show_connect.orig = mode->show_connect.new = fd->show_connect;
     mode->hide_hostname.orig = mode->hide_hostname.new = fd->hide_hostname;

     while (*mode_str != 0) {
	  switch (*mode_str) {
	  case '+':
	       flag = 1;
	       break;
	  case '-':
	       flag = 0;
	       break;
	  case 'b':
	       if (fd->oper) {
		    mode->show_whois.new = flag;
		    if (flag)
			 mode->num_add++;
		    else
			 mode->num_sub++;
	       } else {
		    irc_fprintf(fd->sok, ":%s 502 %s :Only Operators can set user mode +b\r\n", NAME, fd->nick);
	       }
	       break;
	  case 'c':
	       if (fd->oper) {
		    mode->show_connect.new = flag;
		    if (flag)
			 mode->num_add++;
		    else
			 mode->num_sub++;
	       }
	       break;
	  case 'i':
	       mode->invisible.new = flag;
	       if (flag)
		    mode->num_add++;
	       else
		    mode->num_sub++;
	       break;
	  case 's':
	       mode->s_msg.new = flag;
	       if (flag)
		    mode->num_add++;
	       else
		    mode->num_sub++;
	       break;
	  case 'w':
	       mode->wallops.new = flag;
	       if (flag)
		    mode->num_add++;
	       else
		    mode->num_sub++;
	       break;
	  case 'o':
	       mode->oper.new = flag;
	       if (flag)
		    mode->num_add++;
	       else
		    mode->num_sub++;
	       break;
	  case 'd':
	       mode->deaf.new = flag;
	       if (flag)
		    mode->num_add++;
	       else
		    mode->num_sub++;
	       break;
	  case 'k':
	       mode->service.new = flag;
	       if (flag)
		    mode->num_add++;
	       else
		    mode->num_sub++;
	       break;
	  case 'g':
	       mode->globops.new = flag;
	       if (flag)
		    mode->num_add++;
	       else
		    mode->num_sub++;
	       break;
	  case 'x':
	       mode->hide_hostname.new = flag;
	       if (flag)
		    mode->num_add++;
	       else
		    mode->num_sub++;
	       break;
	  default:
	       irc_fprintf(fd->sok, ERR_UMODEUNKNOWNFLAG, NAME, fd->nick, *mode_str);
	       irc_fflush(fd->sok);
	  }
	  mode_str++;
     }
     return mode;
}

void __set_umode(struct fd_list *fd, struct user_mode_change *mode)
{
     char mode_str[20];
     int i, nadd, ndel;

     memset((void *) mode_str, 0, 20);

     i = 0;
     nadd = 0;
     ndel = 0;

     if (mode->show_whois.new != mode->show_whois.orig) {
	  if (mode->show_whois.new) {
	       nadd++;
	  } else {
	       ndel++;
	  }
     }
     if (mode->invisible.new != mode->invisible.orig) {
	  if (mode->invisible.new) {
	       nadd++;
	  } else {
	       ndel++;
	  }
     }
     if (mode->s_msg.new != mode->s_msg.orig) {
	  if (mode->s_msg.new) {
	       nadd++;
	  } else {
	       ndel++;
	  }
     }
     if (mode->wallops.new != mode->wallops.orig) {
	  if (mode->wallops.new) {
	       nadd++;
	  } else {
	       ndel++;
	  }
     }
     if (mode->deaf.new != mode->deaf.orig) {
	  if (mode->deaf.new) {
	       nadd++;
	  } else {
	       ndel++;
	  }
     }
     if (mode->globops.new != mode->globops.orig) {
	  if (mode->globops.new) {
	       nadd++;
	  } else {
	       ndel++;
	  }
     }
     if (mode->show_connect.new != mode->show_connect.orig) {
	  if (mode->show_connect.new) {
	       nadd++;
	  } else {
	       ndel++;
	  }
     }
     if (mode->oper.new != mode->oper.orig && mode->oper.new == 0) {
	  ndel++;
     }
#if defined(NO_OPER_SERVICE)
     if (mode->service.new != mode->service.orig && mode->service.new == 0) {
	  ndel++;
     }
#else
     if (mode->service.new != mode->service.orig) {
	if (mode->service.new == 1) {
		if (fd->oper) {
			nadd++;
		} else {
			mode->service.new = 0;
		}
	} else {
		ndel++;
	}
     }
#endif
     if (mode->hide_hostname.new != mode->hide_hostname.orig) {
	  if (mode->hide_hostname.new) {
	       nadd++;
	  } else {
	       ndel++;
	  }
     }
     if (nadd > 0) {
	  mode_str[0] = '+';
	  i++;
	  if (mode->show_whois.new == 1 && mode->show_whois.new != mode->show_whois.orig) {
	       mode_str[i] = 'b';
	       i++;
	  }
	  if (mode->invisible.new == 1 && mode->invisible.new != mode->invisible.orig) {
	       mode_str[i] = 'i';
	       i++;
	  }
	  if (mode->s_msg.new == 1 && mode->s_msg.new != mode->s_msg.orig) {
	       mode_str[i] = 's';
	       i++;
	  }
	  if (mode->wallops.new == 1 && mode->wallops.new != mode->wallops.orig) {
	       mode_str[i] = 'w';
	       i++;
	  }
	  if (mode->deaf.new == 1 && mode->deaf.new != mode->deaf.orig) {
	       mode_str[i] = 'd';
	       i++;
	  }
	  if (mode->globops.new == 1 && mode->globops.new != mode->globops.orig) {
	       mode_str[i] = 'g';
	       i++;
	  }
	  if (mode->show_connect.new == 1 && mode->show_connect.new != mode->show_connect.orig) {
	       mode_str[i] = 'c';
	       i++;
	  }
	  if (mode->oper.new == 1 && mode->oper.new != mode->oper.orig) {
	       mode->oper.new = 0;
	  }
	  if (mode->service.new == 1 && mode->service.new != mode->service.orig) {
	       if (fd->oper) {
		mode_str[i] = 'k';
		i++;
	       } else {
                mode->service.new = 0;
               }
	  }
	  if (mode->hide_hostname.new == 1 && mode->hide_hostname.new != mode->hide_hostname.orig) {
	       mode_str[i] = 'x';
	       i++;
	  }
     }
     if (ndel > 0) {
	  mode_str[i] = '-';
	  i++;
	  if (mode->show_whois.new == 0 && mode->show_whois.new != mode->show_whois.orig) {
	       mode_str[i] = 'b';
	       i++;
	  }
	  if (mode->invisible.new == 0 && mode->invisible.new != mode->invisible.orig) {
	       mode_str[i] = 'i';
	       i++;
	  }
	  if (mode->s_msg.new == 0 && mode->s_msg.new != mode->s_msg.orig) {
	       mode_str[i] = 's';
	       i++;
	  }
	  if (mode->wallops.new == 0 && mode->wallops.new != mode->wallops.orig) {
	       mode_str[i] = 'w';
	       i++;
	  }
	  if (mode->oper.new == 0 && mode->oper.new != mode->oper.orig) {
	       mode_str[i] = 'o';
	       i++;
	  }
	  if (mode->deaf.new == 0 && mode->deaf.new != mode->deaf.orig) {
	       mode_str[i] = 'd';
	       i++;
	  }
	  if (mode->service.new == 0 && mode->service.new != mode->service.orig) {
	       mode_str[i] = 'k';
	       i++;
	  }
	  if (mode->globops.new == 0 && mode->globops.new != mode->globops.orig) {
	       mode_str[i] = 'g';
	       i++;
	  }
	  if (mode->show_connect.new == 0 && mode->show_connect.new != mode->show_connect.orig) {
	       mode_str[i] = 'c';
	       i++;
	  }
	  if (mode->hide_hostname.new == 0 && mode->hide_hostname.new != mode->hide_hostname.orig) {
	       mode_str[i] = 'x';
	       i++;
	  }
     }
     if (nadd == 0 && ndel == 0) {
	  return;
     }
     irc_fprintf(fd->sok, ":%s!%s@%s MODE %s %s\r\n", fd->nick, fd->userid, fd->hostname, fd->nick, mode_str);
     irc_fflush(fd->sok);

     fd->show_whois = mode->show_whois.new;
     fd->invisible = mode->invisible.new;
     fd->s_msg = mode->s_msg.new;
     fd->wallops = mode->wallops.new;
     fd->oper = mode->oper.new;
     fd->deaf = mode->deaf.new;
     fd->service = mode->service.new;
     fd->globops = mode->globops.new;
     fd->show_connect = mode->show_connect.new;
     fd->hide_hostname = mode->hide_hostname.new;
     return;
}

void m_quit(struct fd_list *fd, char **av, int ac)
{
     char *t2, *quit_msg;

     msgs.quit.use++;
     msgs.quit.bytes += strlen(tmp);

     if (ac == 2) {
	  t2 = av[1];
     } else {
	  t2 = NULL;
     }

     if (t2 == NULL) {
	  char *t = "Client exited";

	  quit_msg = t;
     } else {
	  quit_msg = t2;
	  if (quit_msg[strlen(quit_msg) - 1] == '\r' || quit_msg[strlen(quit_msg) - 1] == '\n') {
	       quit_msg[strlen(quit_msg) - 1] = 0;
	  }
	  if (quit_msg[strlen(quit_msg) - 1] == '\r' || quit_msg[strlen(quit_msg) - 1] == '\n') {
	       quit_msg[strlen(quit_msg) - 1] = 0;
	  }
     }
     {
	  struct whowas *t, *l;
	  t = (struct whowas *) gcAlloc(sizeof(struct whowas));

	  t->magic = WHOWAS_MAGIC;
	  strcpy(t->nick, fd->nick);
	  strcpy(t->userid, fd->userid);
	  strcpy(t->hostname, fd->hostname);
	  strcpy(t->realname, fd->realname);
	  t->signoff = (int) __time(NULL);
	  l = __whowas_head;
	  if (l == NULL) {
	       __whowas_head = t;
	  } else {
	       while (l->next != NULL)
		    l = l->next;
	       l->next = t;
	       t->prev = l;
	  }
     }
     *quit = 1;

     fd->class->refcount--;
     fd->class->num_connects--;
     gcDecRefcnt(fd->class);

     {
	  struct fd_list *u = __head->next;

	  while (u) {
	       if (u->show_connect && u->oper) {
		    irc_fprintf(u->sok, ":%s NOTICE %s :*** Notice -- Client exiting: %s (%s@%s) [%s]\r\n",
			      server_name, u->nick, fd->nick, fd->userid, fd->hostname, quit_msg);
	       }
	       u = u->next;
	  }
     }

     {
	  struct user_chan_list *t, *u;

	  t = (struct user_chan_list *) (fd->chan_list);
	  while (t != NULL) {
	       {
		    char tmp[255];

		    snprintf(tmp, 255, "list pointer is %p", (void *) t);
		    log(tmp);
	       }
	       u = t->next;
	       __part_channel(fd, t->chan->name, 1, quit_msg, 0, NULL);
	       t = u;
	  }
     }
     {
	  struct user_invite_list *t, *u;

	  t = (struct user_invite_list *) (fd->invite_list);
	  while (t != NULL) {
	       u = t->next;
	       {
		    struct invite_list *p;

		    p = t->chan->invites;
		    while (p != NULL) {
			 if (p->user == fd) {
			      if (p->next == NULL && p->prev == NULL) {
				   gcDecRefcnt(p);
				   p = NULL;
				   t->chan->invites = NULL;
			      } else if (p->next == NULL && p->prev != NULL) {
				   (p->prev)->next = NULL;
				   gcDecRefcnt(p);
				   p = NULL;
			      } else if (p->next != NULL && p->prev == NULL) {
				   (p->next)->prev = NULL;
				   t->chan->invites = p;
				   gcDecRefcnt(p);
				   p = NULL;
			      } else {
				   (p->next)->prev = p->prev;
				   (p->prev)->next = p->next;
				   gcDecRefcnt(p);
				   p = NULL;
			      }
			 }
			 if (p != NULL)
			      p = p->next;
		    }
	       }
	       gcDecRefcnt(t);
	       t = u;
	  }
     }

     return;
}


struct dns_list {
     char hostname[50];
     unsigned int addr;
     int time;
};

struct dns_hash {
     struct dns_list list[50];
     int hash_ptr;
};

static struct dns_hash d_hash[256];	/* hash table based on first byte of ip address */

static int dns_ptr = 1;
static int dns_stale_time = 86400;	/* 1 day */

static struct {
     int ca_adds;
     int ca_dels;
     int ca_expires;
     int ca_lookups;
     int ca_na_hits;
     int ca_nu_hits;
     int ca_updates;
} cainfo;

static struct {
     int re_errors;
     int re_nu_look;
     int re_na_look;
     int re_replies;
     int re_requests;
     int re_resends;
     int re_sent;
     int re_timeouts;
     int re_shortttl;
     int re_unkrep;
} reinfo;

void __init_dns_hash()
{
     int i;

     for (i = 0; i < 256; i++) {
	  bzero(&(d_hash[i]), sizeof(struct dns_hash));
     }
     bzero(&cainfo, sizeof(int) * 7);
     bzero(&reinfo, sizeof(int) * 10);
}

void __init_dns_list()
{
     __init_dns_hash();
}

struct dns_list *
  lookup_dns_hash(unsigned int addr)
{
     unsigned char *p = (unsigned char *) &addr;

     if (d_hash[p[0]].hash_ptr = 0) {
	  return NULL;
     } else {
	  int h_ptr;

	  cainfo.ca_lookups++;
	  if (!addr || (signed int) addr == -1) {
	       errno = EINVAL;
	       reinfo.re_errors++;
	       return NULL;	/* 0 or (signed) -1 is not a valid address, give me a break. */
	  }
	  for (h_ptr = 0; h_ptr < 50; h_ptr++) {
	       if (d_hash[p[0]].list[h_ptr].addr == addr) {
		    int now = __time(NULL);

		    if (d_hash[p[0]].list[h_ptr].time + dns_stale_time > now) {
			 struct dns_list *q = &(d_hash[p[0]].list[h_ptr]);

			 cainfo.ca_nu_hits++;
			 return q;	/* valid entry */
		    } else {
			 cainfo.ca_expires++;
			 return NULL;	/* stale entry */
		    }
	       }
	  }
	  return NULL;		/* no entry found */
     }
}

/* add_dns_hash(struct hostent)
   returns -1 on error, sets errno appropriately -
   EFAULT if hp isn't a valid pointer (i.e. NULL)
   EINVAL if hp isn't a pointer to an AF_INET address

   returns 0 on success
 */

int add_dns_hash(struct hostent *hp)
{
     unsigned int addr;
     char hostname[50];
     unsigned char *p;
     int d_hash_ptr;

     if (!hp) {
	  errno = EFAULT;	/* null isn't a valid hostent structure! */
	  return -1;
     }
     if (hp->h_addrtype != AF_INET) {
	  errno = EINVAL;	/* what am i supposed to do with a non-inet address? */
	  return -1;
     }
     cainfo.ca_adds++;

     /* addr = hp->h_addr_list[0][0] << 24 + hp->h_addr_list[0][1] << 16 + hp->h_addr_list[0][2] << 8 + hp->h_addr_list[0][3]; */
     bcopy(hp->h_addr_list[0], &addr, hp->h_length);
     snprintf(hostname, 50, "%s", hp->h_name);
     p = (unsigned char *) &addr;

     d_hash_ptr = d_hash[p[0]].hash_ptr;
     if (d_hash_ptr > 50) {
	  d_hash_ptr = 0;
	  d_hash[p[0]].hash_ptr = 0;
     } else {
	  d_hash_ptr++;
	  d_hash[p[0]].hash_ptr = d_hash_ptr;
     }

     d_hash[p[0]].list[d_hash_ptr].time = __time(NULL);
     d_hash[p[0]].list[d_hash_ptr].addr = addr;
     snprintf(d_hash[p[0]].list[d_hash_ptr].hostname, 50, "%s", hostname);

     return 0;
}

char *
  _inetntoa(unsigned int addr)
{
     static char buf[16];
     unsigned char *p = (unsigned char *) &addr;

     bzero(buf, 16);
     snprintf(buf, 16, "%i.%i.%i.%i", p[0], p[1], p[2], p[3]);
     return buf;
}

void m_dns(struct fd_list *fd, char **av, int ac)
{
     int dptr, hptr, now;

     dptr = hptr = 0;

     now = __time(NULL);

     if (ac == 1 || ac == 2 && (!av[1] || av[1][0] == 0)) {
	  irc_fprintf(fd->sok, ":%s NOTICE %s :Ca %d Cd %d Ce %d Cl %d Ch %d:%d Cu %d\r\n", server_name, fd->nick,
		      cainfo.ca_adds, cainfo.ca_dels, cainfo.ca_expires, cainfo.ca_lookups, cainfo.ca_na_hits, cainfo.ca_nu_hits,
		      cainfo.ca_updates);
	  irc_fprintf(fd->sok, ":%s NOTICE %s :Re %d Rl %d/%d Rp %d Rq %d\r\n", server_name, fd->nick, reinfo.re_errors,
		      reinfo.re_nu_look, reinfo.re_na_look, reinfo.re_replies, reinfo.re_requests);
	  irc_fprintf(fd->sok, ":%s NOTICE %s :Ru %d Rsh %d Rs %d(%d) Rt %d\r\n", server_name, fd->nick,
		      reinfo.re_unkrep, reinfo.re_shortttl, reinfo.re_sent, reinfo.re_resends, reinfo.re_timeouts);
	  return;
     } else {
	  if (!fd->oper)
	       return;
	  for (dptr = 0; dptr < 256; dptr++) {
	       for (hptr = 0; hptr < 50; hptr++) {
		    if (d_hash[dptr].list[hptr].addr) {
			 if (d_hash[dptr].list[hptr].time + dns_stale_time > now) {
			      irc_fprintf(fd->sok, ":%s NOTICE %s :Ex %d ttl %d host %s(%s)\r\n",
					  server_name, fd->nick, d_hash[dptr].list[hptr].time + dns_stale_time - now,
					  dns_stale_time, d_hash[dptr].list[hptr].hostname, _inetntoa(d_hash[dptr].list[hptr].addr));
			 }
		    }
	       }
	  }
     }
}

void *
  check_hostname(void *arg)
{
     struct fd_list *fd = (struct fd_list *) arg;
     struct sockaddr_in *sa = alloca(sizeof(struct sockaddr_in));
     int salen = sizeof(struct sockaddr_in);
     struct dns_list *d_ptr;
     struct hostent *host_struct, he_buf2;
     char he_buf[100];
     int rval;

     log2("checking user hostname: arg = %p", arg);
     gcIncRefcnt(fd);
     gcIncRefcnt(fd->sok);

     rval = getpeername(fd->fd, (struct sockaddr *) sa, &salen);

     if (rval) {
	  log("client disconnected, exiting check_hostname");
	  fd->nslookup_is_in_use = 0;
	  gcDecRefcnt(fd);
	  gcDecRefcnt(fd->sok);
	  pthread_exit(NULL);
     }
     log("checking for cached entry..");
     d_ptr = lookup_dns_hash(sa->sin_addr.s_addr);
     if (d_ptr) {
	  log("found cached entry, using..");
	  log("copying hostname to user structure..");
	  snprintf(fd->hostname, 50, "%s", d_ptr->hostname);
	  log("notifying user..");
	  irc_fprintf(fd->sok, ":%s NOTICE %s :*** Notice -- Hostname found (cached)..\r\n", NAME,
		      (fd->nick[0] == 0 ? "*" : fd->nick));
     } else {
	  log("no cached entry, finding entry..");
	  reinfo.re_na_look++;
	  reinfo.re_requests++;
	  reinfo.re_sent++;
	  gethostbyaddr_r((unsigned char *) &(sa->sin_addr.s_addr), sizeof(sa->sin_addr.s_addr), AF_INET, &he_buf2, he_buf, 100, &host_struct,
			  &h_errno);
	  if (host_struct != NULL) {
	       log("hostname found..");
	       reinfo.re_replies++;
	       snprintf(fd->hostname, 50, "%s", host_struct->h_name);
	       irc_fprintf(fd->sok, ":%s NOTICE %s :*** Notice -- Hostname found..\r\n", NAME,
			   (fd->nick[0] == 0 ? "*" : fd->nick));
	  } else {
	       log2("no hostname for host: error = %s",
		    h_errno < 0 ? strerror(errno) : hstrerror(h_errno));
	       reinfo.re_errors++;
	       irc_fprintf(fd->sok, ":%s NOTICE %s :*** Notice -- Couldn't find your hostname: error %c%i (%s)\r\n",
			   NAME, (fd->nick[0] == 0 ? "*" : fd->nick), (h_errno < 0 ? 's' : 'd'), 
				(h_errno < 0 ? errno : h_errno), (h_errno < 0 ? strerror(errno) : hstrerror(h_errno)));
	  }
	  if (host_struct) {
	       add_dns_hash(host_struct);
	  }
     }
     log("returning..");
     gcDecRefcnt(fd);
     gcDecRefcnt(fd->sok);
#ifdef PTHREADIDENT
     fd->nslookup_is_in_use = 0;
     pthread_exit(NULL);
#endif
     return NULL;
}

void *
  check_ident(void *arg)
{
     struct fd_list *fd = (struct fd_list *) arg;
     struct sockaddr_in *sa = (struct sockaddr_in *) alloca(sizeof(struct sockaddr_in));
     int size = sizeof(struct sockaddr_in);

     bzero(sa, size);
     getpeername(fd->fd, sa, &size);

     log2("checking ident: arg = %p", arg);

     gcIncRefcnt(fd);
     gcIncRefcnt(fd->sok);

     irc_fprintf(fd->sok, ":%s NOTICE %s :*** Notice -- Checking ident...\r\n", NAME,
		 (fd->nick[0] == 0 ? "*" : fd->nick));
     {
	  int ifd = get_connection(fd->hostname, 113);

	  if (ifd == -1) {
	       irc_fprintf(fd->sok, ":%s NOTICE %s :*** Notice -- No ident response (%s)\r\n", NAME,
			   (fd->nick[0] == 0 ? "*" : fd->nick), strerror(errno));
	       gcDecRefcnt(fd);
#ifdef PTHREADIDENT
	       fd->ident_is_in_use = 0;
	       pthread_exit(NULL);
#else
	       return NULL;
#endif
	  } else {
	       FILE2 *ident = irc_fdopen(ifd, "r+");
	       char tmp[255];
	       int rval;

	       setvbuf(ident->sok, NULL, _IONBF, 0);
	       bzero(tmp, 255);
	       irc_fprintf(ident, "%i, %i\r\n", htons(sa->sin_port), port);
	       {
		    struct timeval tv;
		    fd_set rfd;

		    tv.tv_sec = 0;
		    tv.tv_usec = 10000;
		    FD_ZERO(&rfd);
		    FD_SET(ifd, &rfd);
		    select(ifd + 1, &rfd, NULL, NULL, &tv);
	       }
	       rval = m_read(ifd, (void *) tmp, 255);
	       if (rval) {
		    int remote_port, local_port;
		    char *r, *userid;
		    int p;

		    r = strtok(tmp, ":");
		    r = strtok(NULL, ":");
		    if (!strcasecmp(r, " USERID ")) {
			 p = 5;
			 r = strtok(NULL, ":");
			 r = strtok(NULL, ":");
			 userid = strtok(r, "\r\n");
			 if (*userid == ' ')
			      userid++;
		    } else {
			 p = 0;
		    }
		    if (p == 5) {
			 irc_fprintf(fd->sok, ":%s NOTICE %s :*** Notice -- Got ident response\r\n", NAME,
				     (fd->nick[0] == 0 ? "*" : fd->nick));
			 snprintf(fd->userid, 15, "%s", userid);
		    } else {
			 irc_fprintf(fd->sok, ":%s NOTICE %s :*** Notice -- No ident response\r\n", NAME,
				     (fd->nick[0] == 0 ? "*" : fd->nick));
		    }
	       } else {
		    irc_fprintf(fd->sok, ":%s NOTICE %s :*** Notice -- No ident response\r\n", NAME,
				(fd->nick[0] == 0 ? "*" : fd->nick));
	       }
	       irc_fclose(ident);
	       close(ifd);
	       gcDecRefcnt(fd);
	       gcDecRefcnt(fd->sok);
#ifdef PTHREADIDENT
	       fd->ident_is_in_use = 0;
	       pthread_exit(NULL);
#else
	       return NULL;
#endif
	  }
     }
}

void *watch_for_reg(void *arg) {
	struct fd_list *fd = (struct fd_list *) arg;
	while (fd->ident_is_in_use || !fd->nick[0] || !fd->userid[0]) {
		usleep(10000);
	}
	fd->waiting_for_reg = 0;
	fd->registered = 1;
        __registered_reply(fd);
	pthread_exit(NULL);
}

void check_reg(struct fd_list *fd) {
	if (!fd->registered) {
		if (!fd->ident_is_in_use && fd->nick[0] && fd->userid[0]) {
			fd->registered = 1;
		} else {
			if (!fd->waiting_for_reg) {
				int q;
				fd->waiting_for_reg = 1;
				q = pthread_create(&(fd->reg_id), NULL, watch_for_reg, (void *) fd);
			}
		}
	}
}

void m_user(struct fd_list *fd, char **av, int ac)
{
     char *t2, *msg;
     struct sockaddr_in *sa;
     int size;
     struct hostent *host_struct;
     struct dns_list *d_ptr;

#ifdef PTHREADIDENT
     pthread_t thread_id;

#endif

     msgs.user.use++;
     msgs.user.bytes += strlen(tmp);
     if (ac < 5) {
	  __not_enough_parameters(fd, t);
	  return;
     }
     t2 = av[1];		/* first parm is userid */

#ifdef CHECKIDENT
     snprintf(fd->userid, 15, "~%s", t2);
#else
     snprintf(fd->userid, 15, "%s", t2);
#endif /* CHECKIDENT */

     t2 = av[2];		/* second parm is hostname */
     irc_fprintf(fd->sok, ":%s NOTICE %s :*** Notice -- Looking up your hostname...\r\n", NAME,
		 (fd->nick[0] == 0 ? "*" : fd->nick));
     sa = (struct sockaddr_in *) calloc(sizeof(struct sockaddr_in));
     size = sizeof(struct sockaddr_in);

     getpeername(fd->fd, sa, &size);
     {
	  unsigned char *s = (char *) &(sa->sin_addr.s_addr);

	  sprintf(fd->hostname, "%u.%u.%u.%u", s[0], s[1], s[2], s[3]);
     }
#ifdef CHECKIDENT
#ifdef PTHREADIDENT
     {
	  int thread;

	  thread = pthread_create(&thread_id, NULL, check_ident, (void *) fd);
	  fd->ident_id = thread_id;
	  fd->ident_is_in_use = 1;
	  pthread_detach(thread_id);
     }
#else
     check_ident(fd);
#endif /* PTHREADIDENT */
#endif /* CHECKIDENT */
     /* This code here is the DNS lookup .. going into check_hostname() */
#ifdef PTHREADIDENT
     {
	  int thread;

	  thread = pthread_create(&thread_id, NULL, check_hostname, (void *) fd);
	  fd->nslookup_id = thread_id;
	  fd->nslookup_is_in_use = 1;
	  pthread_detach(thread_id);
     }
#else
     check_hostname(fd);
#endif
     __free(sa);
     t2 = av[3];
     msg = av[4];
     if (msg[strlen(msg) - 1] == '\r' || msg[strlen(msg) - 1] == '\n') {
	  msg[strlen(msg) - 1] = 0;
     }
     if (msg[strlen(msg) - 1] == '\r' || msg[strlen(msg) - 1] == '\n') {
	  msg[strlen(msg) - 1] = 0;
     }
     snprintf(fd->realname, 100, "%s", msg);
#if 0
     if (fd->nick[0] != 0 && !fd->ident_is_in_use) {
	  fd->registered = 1;
	  __registered_reply(fd);
     } else if (fd->ident_is_in_use && fd->nick[0] != 0) {
	  int thread;
	  pthread_create(&thread_id, NULL, watch_for_reg, (void *) fd);
     }
#else
     check_reg(fd);
#endif
     return;
}

int isValidNick(char *str) {
	char *q = str;
	for (; *q; q++) {
		if (*q < 33) {
			return 0;
		}
	}
	return 1;
}

void m_nick(struct fd_list *fd, char **av, int ac)
{
     char *t2;
     struct fd_list *list;

     msgs.nick.use++;
     msgs.nick.bytes += strlen(tmp);
     if (ac < 2) {
	  __not_enough_parameters(fd, t);
	  return;
     }
     t2 = av[1];
     if (!isValidNick(av[1])) {
	  irc_fprintf(fd->sok, ":%s 432 %s %s :Erroneous nickname\r\n", server_name, fd->nick, "(null)");
	  return;
     }
     if (!strcmp(fd->nick, av[1])) {
	return; /* ignore NICK command */
     }
     /* is the nick in use?? */
     list = __head;
     while (list != NULL) {
	  if (!strcasecmp(list->nick, t2)) {
	       __nickname_in_use(fd, t2);
	       return;
	  }
	  list = list->next;
     }
     if (fd->registered) {
	  char *my_hostname = (fd->hide_hostname ? "hidden.hurricane.net" : fd->hostname);

	  if (fd->chan_list != NULL) {
	       struct user_chan_list *t = (struct user_chan_list *) (fd->chan_list);

	       while (t != NULL) {
		    struct user_list *u;

		    u = t->chan->users;
		    while (u != NULL) {
			 irc_fprintf(u->user->sok, ":%s!%s@%s NICK %s\r\n", fd->nick, fd->userid,
				     my_hostname, t2);
			 irc_fflush(u->user->sok);
			 u = u->next;
		    }
		    t = t->next;
	       }
	  } else {
	       irc_fprintf(fd->sok, ":%s!%s@%s NICK %s\r\n", fd->nick, fd->userid, my_hostname, t2);
	       irc_fflush(fd->sok);
	  }
	  {
	       struct whowas *t, *u;
	       t = (struct whowas *) gcAlloc(sizeof(struct whowas));

	       t->magic = WHOWAS_MAGIC;
	       strcpy(t->nick, fd->nick);
	       strcpy(t->userid, fd->userid);
	       strcpy(t->hostname, my_hostname);
	       strcpy(t->realname, fd->realname);
	       t->signoff = (int) __time(NULL);
	       u = __whowas_head;
	       if (u == NULL) {
		    __whowas_head = t;
	       } else {
		    while (u->next != NULL)
			 u = u->next;
		    u->next = t;
		    t->prev = u;
	       }
	  }
     }
     snprintf(fd->nick, 15, "%s", t2);
#if 0
     if (!fd->registered && fd->userid[0] != 0) {
	  fd->registered = 1;
	  __registered_reply(fd);
     }
#else
     check_reg(fd);
#endif
     return;
}

void m_whois(struct fd_list *fd, char **av, int ac)
{
     char *t2;

     msgs.whois.use++;
     msgs.whois.bytes += strlen(tmp);
     if (ac < 2) {
	  __not_enough_parameters(fd, t);
     } else {
	  if (ac == 2) {
	       t2 = av[1];
	  } else {
	       t2 = av[2];
	  }
	  __whois_reply(fd, t2);
     }
     return;
}

void m_oper(struct fd_list *fd, char **av, int ac)
{
     msgs.oper.use++;
     msgs.oper.bytes += strlen(tmp);
     if (ac < 3) {
	  __not_enough_parameters(fd, t);
     } else {
	  char *op_nick = av[1];
	  char *op_pass = av[2];

	  __oper_reply(fd, op_nick, op_pass);
     }
     return;
}


void m_pass(struct fd_list *fd, char **av, int ac)
{
     char *p;

     msgs.pass.use++;
     msgs.pass.bytes += strlen(tmp);
     if (ac >= 2) {
	  p = av[1];
	  strncpy(fd->conn_pass, p, (strlen(p) > 20 ? 20 : strlen(p)));
     } else {
	  __not_enough_parameters(fd, t);
     }
     return;
}

void m_away(struct fd_list *fd, char **av, int ac)
{
     msgs.away.use++;
     msgs.away.bytes += strlen(tmp);
     if (ac == 1 || (ac == 2 && (av[1][0] == '\r' || av[1][0] == '\n' || av[1][0] == '\0'))) {
	  fd->away_set = 0;
	  memset((void *) (fd->away_msg), 0, 150);
	  irc_fprintf(fd->sok, RPL_UNAWAY, NAME, fd->nick);
     } else {
	  char *away_msg = av[1];

	  fd->away_set = 1;
	  if (away_msg[strlen(away_msg) - 1] == '\r' || away_msg[strlen(away_msg) - 1] == '\n') {
	       away_msg[strlen(away_msg) - 1] = 0;
	  }
	  if (away_msg[strlen(away_msg) - 1] == '\r' || away_msg[strlen(away_msg) - 1] == '\n') {
	       away_msg[strlen(away_msg) - 1] = 0;
	  }
	  strncpy(fd->away_msg, away_msg, (strlen(away_msg) > 150 ? 150 : strlen(away_msg)));
	  irc_fprintf(fd->sok, RPL_NOWAWAY, NAME, fd->nick);
     }
     return;
}

void m_whowas(struct fd_list *fd, char **av, int ac)
{
     struct whowas *u;
     char *nick, *tmp;
     int count, n;

     u = __whowas_head;
     n = 0;
     if (ac < 2) {
	  __not_enough_parameters(fd, t);
     } else {
	  nick = av[1];
	  tmp = av[2];
	  if (tmp == NULL) {
	       count = -1;
	  } else {
	       count = strtol(tmp, (char **) NULL, 10);
	       if (count < 1)
		    count = -1;
	  }
	  if (u != NULL) {
	       while (u->next != NULL)
		    u = u->next;
	       while (u != NULL) {
		    if (!strcasecmp(u->nick, nick)) {
			 irc_fprintf(fd->sok, RPL_WHOISUSER, NAME, fd->nick, u->nick, u->userid,
				     u->hostname, u->realname);
			 irc_fprintf(fd->sok, RPL_WHOISSERVER, NAME, fd->nick, u->nick, NAME,
				     ctime((time_t *) & (u->signoff)));
		    }
		    if (count != -1) {
			 n++;
			 if (n >= count) {
			      u = NULL;
			 }
		    }
		    if (u != NULL)
			 u = u->prev;
	       }
	  }
	  irc_fprintf(fd->sok, ":%s 369 %s %s :End of WHOWAS\r\n", NAME, fd->nick, nick);
     }
     return;
}
